#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 2000 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmdict.cpp - dictionary metaclass
Function
  
Notes
  
Modified
  01/24/00 MJRoberts  - Creation
*/

#include <stdlib.h>
#include <string.h>

#include "t3std.h"
#include "vmdict.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmtype.h"
#include "vmglob.h"
#include "vmobj.h"
#include "vmundo.h"
#include "vmmcreg.h"
#include "vmfile.h"
#include "vmhash.h"
#include "vmbif.h"
#include "vmmeta.h"
#include "vmstack.h"


/* ------------------------------------------------------------------------ */
/*
 *   Dictionary undo record.  Each time we change the dictionary, we track
 *   the change with one of these records - we attach this record to the
 *   standard undo record via the 'ptrval' field. 
 */
struct dict_undo_rec
{
    /* action type */
    enum dict_undo_action action;

    /* old truncation length (when action == DICT_UNDO_TRUNCLEN) */
    size_t trunc_len;
    
    /* object value stored in dictionary record */
    vm_obj_id_t obj;

    /* vocabulary property for word association */
    vm_prop_id_t prop;

    /* length of the word */
    size_t len;

    /* text of the word */
    char txt[1];
};

/* undo action codes */
enum dict_undo_action
{
    /* we added this word to the dictionary (undo by deleting it) */
    DICT_UNDO_ADD,

    /* we deleted this word from the dictionary (undo by adding it back) */
    DICT_UNDO_DEL,

    /* we changed the truncation length for the dictionary */
    DICT_UNDO_TRUNCLEN
};

/* ------------------------------------------------------------------------ */
/*
 *   truncating case-insensitive hash
 */
class CVmHashFuncCITrunc: public CVmHashFuncCI
{
public:
    CVmHashFuncCITrunc(size_t trunc_len) { trunc_len_ = trunc_len; }

    /* compute the hash value */
    unsigned int compute_hash(const char *str, size_t len) const
    {
        /* 
         *   if the length is longer than the truncation length, hash only
         *   to the truncation length 
         */
        if (len > trunc_len_)
            len = trunc_len_;

        /* use the normal case-insensitive hash */
        return CVmHashFuncCI::compute_hash(str, len);
    }

private:
    /* truncation length */
    size_t trunc_len_;
};


/* ------------------------------------------------------------------------ */
/*
 *   dictionary object statics 
 */

/* metaclass registration object */
static CVmMetaclassDict metaclass_reg_obj;
CVmMetaclass *CVmObjDict::metaclass_reg_ = &metaclass_reg_obj;

/* function table */
int (CVmObjDict::
     *CVmObjDict::func_table_[])(VMG_ vm_obj_id_t self,
                                 vm_val_t *retval, uint *argc) =
{
    &CVmObjDict::getp_undef,
    &CVmObjDict::getp_find,
    &CVmObjDict::getp_find_trunc,
    &CVmObjDict::getp_settrunc,
    &CVmObjDict::getp_add,
    &CVmObjDict::getp_del,
    &CVmObjDict::getp_is_defined,
    &CVmObjDict::getp_is_defined_trunc
};

/* ------------------------------------------------------------------------ */
/*
 *   Dictionary object implementation 
 */

/* 
 *   create dynamically using stack arguments 
 */
vm_obj_id_t CVmObjDict::create_from_stack(VMG_ const uchar **pc_ptr,
                                          uint argc)
{
    vm_obj_id_t id;
    size_t trunc_len;
    CVmObjDict *obj;
    
    /* check arguments */
    if (argc != 1)
        err_throw(VMERR_WRONG_NUM_OF_ARGS);

    /* pop the truncation length argument */
    trunc_len = (size_t)CVmBif::pop_long_val(vmg0_);

    /* 
     *   allocate the object ID - this type of construction never creates
     *   a root object 
     */
    id = vm_new_id(vmg_ FALSE);

    /* create the object */
    obj = new (vmg_ id) CVmObjDict(vmg0_);

    /* set the truncation length */
    obj->set_trunc_len(vmg_ trunc_len);

    /* build an empty initial hash table */
    obj->create_hash_table(vmg0_);

    /* 
     *   mark the object as modified since image load, since it doesn't
     *   come from the image file at all 
     */
    obj->get_ext()->modified_ = TRUE;
    
    /* return the new ID */
    return id;
}

/*
 *   constructor 
 */
CVmObjDict::CVmObjDict(VMG0_)
{
    /* allocate our extension structure from the variable heap */
    ext_ = (char *)G_mem->get_var_heap()
           ->alloc_mem(sizeof(vm_dict_ext), this);

    /* we have no image data yet */
    get_ext()->image_data_ = 0;
    get_ext()->image_data_size_ = 0;

    /* set a default truncation length */
    get_ext()->trunc_len_ = 6;

    /* no hash table yet */
    get_ext()->hashtab_ = 0;

    /* no non-image entries yet */
    get_ext()->modified_ = FALSE;
}

/* 
 *   notify of deletion 
 */
void CVmObjDict::notify_delete(VMG_ int /*in_root_set*/)
{
    /* free our additional data */
    if (ext_ != 0)
    {
        /* free our hash table */
        if (get_ext()->hashtab_ != 0)
            delete get_ext()->hashtab_;

        /* free the extension */
        G_mem->get_var_heap()->free_mem(ext_);
    }
}

/*
 *   Set the truncation length 
 */
void CVmObjDict::set_trunc_len(VMG_ size_t trunc_len)
{
    /* 
     *   if the truncation length isn't changing, we don't need to do any
     *   addition work 
     */
    if (get_ext()->trunc_len_ == trunc_len)
        return;
    
    /* set the new truncation length in the extension */
    get_ext()->trunc_len_ = trunc_len;

    /* 
     *   if we have a hash table, we must now rebuild it to reflect the
     *   new hash function implied by the new truncation length 
     */
    if (get_ext()->hashtab_ != 0)
        create_hash_table(vmg0_);
}

/*
 *   Rebuild the hash table 
 */
void CVmObjDict::create_hash_table(VMG0_)
{
    CVmHashTable *new_tab;
    CVmHashFunc *hash_func;
    
    /* 
     *   Allocate our new hash table.  If the truncation length is zero,
     *   it means that there's no truncation matching at all, so use an
     *   ordinary hash function; otherwise, use the truncating hash
     *   function. 
     */
    hash_func = (get_ext()->trunc_len_ != 0
                 ? new CVmHashFuncCITrunc(get_ext()->trunc_len_)
                 : new CVmHashFuncCI());
    new_tab = new CVmHashTable(256, hash_func, TRUE);

    /* if we had a previous hash table, copy its contents */
    if (get_ext()->hashtab_ != 0)
    {
        /* copy the old hash table to the new one */
        get_ext()->hashtab_->move_entries_to(new_tab);
        
        /* delete the old hash table */
        delete get_ext()->hashtab_;
    }

    /* store the new hash table in the extension */
    get_ext()->hashtab_ = new_tab;
}

/*
 *   Set a property.  We have no settable properties, so simply signal an
 *   error indicating that the set-prop call is invalid.  
 */
void CVmObjDict::set_prop(VMG_ CVmUndo *, vm_obj_id_t,
                          vm_prop_id_t, const vm_val_t *)
{
    /* no settable properties - throw an error */
    err_throw(VMERR_INVALID_SETPROP);
}

/* 
 *   get a property 
 */
int CVmObjDict::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *retval,
                         vm_obj_id_t self, vm_obj_id_t *source_obj,
                         uint *argc)
{
    ushort func_idx;
    
    /* if we succeed, the result will come directly from this object */
    *source_obj = self;

    /* translate the property into a function vector index */
    func_idx = G_meta_table
               ->prop_to_vector_idx(metaclass_reg_->get_reg_idx(), prop);

    /* call the appropriate function */
    if ((this->*func_table_[func_idx])(vmg_ self, retval, argc))
        return TRUE;

    /* inherit default handling from the base object class */
    return CVmObject::get_prop(vmg_ prop, retval, self, source_obj, argc);
}

/*
 *   evaluate property: lookUp(word_string, &vocab_prop) - returns a list
 *   of the objects defining the given vocabulary property with the given
 *   string 
 */
int CVmObjDict::getp_find(VMG_ vm_obj_id_t, vm_val_t *val, uint *argc)
{
    CVmHashEntryDict *entry;
    const char *str;
    size_t len;
    vm_prop_id_t voc_prop;
    vm_obj_id_t lst_obj;
    CVmObjList *lst;
    size_t cnt;
    vm_dict_entry *cur;

    /* check arguments */
    if (get_prop_check_argc(val, argc, 2))
        return TRUE;

    /* pop the word string and property ID arguments */
    str = CVmBif::pop_str_val(vmg0_);
    voc_prop = CVmBif::pop_propid_val(vmg0_);

    /* get the string's length prefix */
    len = vmb_get_len(str);
    str += VMB_LEN;

    /* look up the string */
    entry = (CVmHashEntryDict *)
            get_ext()->hashtab_->find(str, len);
    if (entry == 0)
    {
        /* this entry isn't in the table - there are no entries */
        cnt = 0;
    }
    else
    {
        /* count the entries matching the desired property */
        for (cnt = 0, cur = entry->get_head() ; cur != 0 ;
             cur = cur->nxt_)
        {
            /* if this one matches the desired property, count it */
            if (cur->prop_ == voc_prop)
                ++cnt;
        }
    }
    
    /* create a list with the required number of elements */
    lst_obj = CVmObjList::create(vmg_ FALSE, cnt);
    lst = (CVmObjList *)vm_objp(vmg_ lst_obj);

    /* set the elements, if there's anything to set */
    if (entry != 0)
    {
        /* loop over all items for this word entry */
        for (cnt = 0, cur = entry->get_head() ; cur != 0 ;
             cur = cur->nxt_)
        {
            /* if this one matches the desired property, add it to the list */
            if (cur->prop_ == voc_prop)
            {
                vm_val_t ele_val;
                
                /* set the element */
                ele_val.set_obj(cur->obj_);
                lst->cons_set_element(cnt, &ele_val);
                
                /* count it */
                ++cnt;
            }
        }
    }

    /* return the list */
    val->set_obj(lst_obj);
    
    /* successful evaluation */
    return TRUE;
}

/*
 *   enumeration callback context for find_trunc 
 */
struct find_trunc_ctx
{
    /* list object - if this is null, we're just counting */
    CVmObjList *lst;

    /* string to check, and its length */
    const char *str;
    size_t len;

    /* vocabulary property value to match */
    vm_prop_id_t voc_prop;

    /* number of matches processed on this phase so far */
    size_t cnt;

    /* callback to invoke on each match (ignored if null) */
    void (*cb)(void *ctx, const char *txt, size_t len,
               vm_prop_id_t prop);

    /* callback context */
    void *cbctx;
};

/*
 *   Callback for find_trunc hash match enumeration
 */
void CVmObjDict::find_trunc_cb(void *ctx0, CVmHashEntry *entry0)
{
    find_trunc_ctx *ctx = (find_trunc_ctx *)ctx0;
    CVmHashEntryDict *entry = (CVmHashEntryDict *)entry0;

    /* 
     *   if this entry matches, process it - note that we only match if
     *   the entry is longer than the given word, because we don't want to
     *   include exact matches for the given word 
     */
    if (entry->getlen() > ctx->len
        && memicmp(entry->getstr(), ctx->str, ctx->len) == 0)
    {
        vm_dict_entry *cur;
        
        /* process the items under this entry */
        for (cur = entry->get_head() ; cur != 0 ; cur = cur->nxt_)
        {
            /* 
             *   If it matches the target property, it's a hit.  If we're
             *   searching for any property at all (indicated by the
             *   context property being set to VM_INVALID_PROP), count
             *   everything.
             */
            if (cur->prop_ == ctx->voc_prop
                || ctx->voc_prop == VM_INVALID_PROP)
            {
                /* if there's a list, add it to the list */
                if (ctx->lst != 0)
                {
                    vm_val_t ele_val;

                    /* set the list element at the current index */
                    ele_val.set_obj(cur->obj_);
                    ctx->lst->cons_set_element(ctx->cnt, &ele_val);
                }

                /* if there's a callback, invoke it */
                if (ctx->cb != 0)
                    (*ctx->cb)(ctx->cbctx, entry->getstr(), entry->getlen(),
                               cur->prop_);
                
                /* count it */
                ++ctx->cnt;
            }
        }
    }
}


/*
 *   property evaluation - look up a word with truncation
 */
int CVmObjDict::getp_find_trunc(VMG_ vm_obj_id_t, vm_val_t *val, uint *argc)
{
    const char *str;
    size_t len;
    vm_prop_id_t voc_prop;
    find_trunc_ctx ctx;
    vm_obj_id_t lst_obj;

    /* check arguments */
    if (get_prop_check_argc(val, argc, 2))
        return TRUE;

    /* pop the word string and property ID arguments */
    str = CVmBif::pop_str_val(vmg0_);
    voc_prop = CVmBif::pop_propid_val(vmg0_);

    /* get the string's length prefix */
    len = vmb_get_len(str);
    str += VMB_LEN;

    /* 
     *   If the given word is shorter than the truncation length, there
     *   are necessarily no matches, because it's not long enough to be a
     *   truncated word.  (If the word is exactly the truncation length,
     *   it can match, since it could be a dictionary word truncated to
     *   exactly the truncation length.)  This isn't an error; simply
     *   return an empty list, since we know we won't be able to find any
     *   matches in this case.
     *   
     *   In addition, if the truncation length is zero, it means that we
     *   don't allow truncation at all, so there are likewise no matches.  
     */
    if (len < get_ext()->trunc_len_ || get_ext()->trunc_len_ == 0)
    {
        /* return an empty list */
        val->set_obj(CVmObjList::create(vmg_ FALSE, (size_t)0));
        return TRUE;
    }

    /* 
     *   enumerate everything that matches the truncated value - first
     *   phase: just count the matches 
     */
    ctx.str = str;
    ctx.len = len;
    ctx.voc_prop = voc_prop;
    ctx.lst = 0;
    ctx.cnt = 0;
    ctx.cb = 0;
    get_ext()->hashtab_->enum_hash_matches(str, len, &find_trunc_cb, &ctx);

    /* create the list to hold the result */
    lst_obj = CVmObjList::create(vmg_ FALSE, ctx.cnt);
    ctx.lst = (CVmObjList *)vm_objp(vmg_ lst_obj);

    /* enumerate everything again to build the list */
    ctx.cnt = 0;
    get_ext()->hashtab_->enum_hash_matches(str, len, &find_trunc_cb, &ctx);

    /* return the list */
    val->set_obj(lst_obj);

    /* success */
    return TRUE;
}

/*
 *   property evaluation - add a word
 */
int CVmObjDict::getp_add(VMG_ vm_obj_id_t self, vm_val_t *val, uint *argc)
{
    vm_obj_id_t obj;
    vm_val_t str_val;
    vm_prop_id_t voc_prop;
    CVmObject *objp;
    const char *lst;
    const char *str;

    /* check arguments */
    if (get_prop_check_argc(val, argc, 3))
        return TRUE;

    /* pop the object, word string, and property ID arguments */
    obj = CVmBif::pop_obj_val(vmg0_);
    G_stk->pop(&str_val);
    voc_prop = CVmBif::pop_propid_val(vmg0_);

    /* ensure that the object value isn't a string or list */
    objp = vm_objp(vmg_ obj);
    if (objp->get_as_list() != 0 || objp->get_as_string(vmg0_) != 0)
        err_throw(VMERR_OBJ_VAL_REQD);

    /* if the string argument is a list, add each word from the list */
    if ((lst = str_val.get_as_list(vmg0_)) != 0)
    {
        size_t cnt;
        size_t idx;
        vm_val_t ele_val;

        /* get and skip the list count prefix */
        cnt = vmb_get_len(lst);

        /* iterate over the list, using 1-based indices */
        for (idx = 1 ; idx <= cnt ; ++idx)
        {
            /* get the next value from the list */
            CVmObjList::index_list(vmg_ &ele_val, lst, idx);

            /* get the next string value */
            if ((str = ele_val.get_as_string(vmg0_)) == 0)
                err_throw(VMERR_STRING_VAL_REQD);

            /* set this string */
            getp_add_string(vmg_ self, str, obj, voc_prop);

            /* 
             *   refresh the list pointer, in case it was const pool data
             *   and got swapped out by resolving the string pointer 
             */
            lst = str_val.get_as_list(vmg0_);
        }
    }
    else if ((str = str_val.get_as_string(vmg0_)) != 0)
    {
        /* add the string */
        getp_add_string(vmg_ self, str, obj, voc_prop);
    }
    else
    {
        /* string value required */
        err_throw(VMERR_STRING_VAL_REQD);
    }

    /* success */
    return TRUE;
}

/*
 *   Service routine for getp_add() - add a single string 
 */
void CVmObjDict::getp_add_string(VMG_ vm_obj_id_t self,
                                 const char *str, vm_obj_id_t obj,
                                 vm_prop_id_t voc_prop)
{
    size_t len;
    int added;

    /* get the string's length prefix */
    len = vmb_get_len(str);
    str += 2;

    /* add the entry */
    added = add_hash_entry(vmg_ str, len, TRUE, obj, voc_prop, FALSE);

    /* 
     *   if there's a global undo object, and we actually added a new
     *   entry, add undo for the change 
     */
    if (added && G_undo != 0)
    {
        dict_undo_rec *undo_rec;

        /* create the undo record */
        undo_rec = alloc_undo_rec(DICT_UNDO_ADD, str, len);
        undo_rec->obj = obj;
        undo_rec->prop = voc_prop;

        /* add the undo record */
        add_undo_rec(vmg_ self, undo_rec);
    }

    /* mark the object as modified since load */
    get_ext()->modified_ = TRUE;
}

/*
 *   property evaluation - delete a word 
 */
int CVmObjDict::getp_del(VMG_ vm_obj_id_t self, vm_val_t *val, uint *argc)
{
    vm_obj_id_t obj;
    vm_val_t str_val;
    vm_prop_id_t voc_prop;
    const char *lst;
    const char *str;

    /* check arguments */
    if (get_prop_check_argc(val, argc, 3))
        return TRUE;

    /* pop the object, word string, and property ID arguments */
    obj = CVmBif::pop_obj_val(vmg0_);
    G_stk->pop(&str_val);
    voc_prop = CVmBif::pop_propid_val(vmg0_);

    /* if the string argument is a list, add each word from the list */
    if ((lst = str_val.get_as_list(vmg0_)) != 0)
    {
        size_t cnt;
        size_t idx;
        vm_val_t ele_val;

        /* get and skip the list count prefix */
        cnt = vmb_get_len(lst);

        /* iterate over the list, using 1-based indices */
        for (idx = 1 ; idx <= cnt ; ++idx)
        {
            /* get the next value from the list */
            CVmObjList::index_list(vmg_ &ele_val, lst, idx);

            /* get the next string value */
            if ((str = ele_val.get_as_string(vmg0_)) == 0)
                err_throw(VMERR_STRING_VAL_REQD);

            /* remove this string */
            getp_del_string(vmg_ self, str, obj, voc_prop);

            /* 
             *   refresh the list pointer, in case it was const pool data
             *   and got swapped out by resolving the string pointer 
             */
            lst = str_val.get_as_list(vmg0_);
        }
    }
    else if ((str = str_val.get_as_string(vmg0_)) != 0)
    {
        /* remove the string */
        getp_del_string(vmg_ self, str, obj, voc_prop);
    }
    else
    {
        /* string value required */
        err_throw(VMERR_STRING_VAL_REQD);
    }

    /* success */
    return TRUE;
}

/*
 *   Service routine for getp_del - delete a string value 
 */
void CVmObjDict::getp_del_string(VMG_ vm_obj_id_t self,
                                 const char *str, vm_obj_id_t obj,
                                 vm_prop_id_t voc_prop)
{
    size_t len;
    int deleted;

    /* get the string's length prefix */
    len = vmb_get_len(str);
    str += 2;

    /* delete this entry */
    deleted = del_hash_entry(vmg_ str, len, obj, voc_prop);

    /* 
     *   if there's a global undo object, and we actually deleted an
     *   entry, add undo for the change 
     */
    if (deleted && G_undo != 0)
    {
        dict_undo_rec *undo_rec;

        /* create the undo record */
        undo_rec = alloc_undo_rec(DICT_UNDO_DEL, str, len);
        undo_rec->obj = obj;
        undo_rec->prop = voc_prop;

        /* add the undo record */
        add_undo_rec(vmg_ self, undo_rec);
    }

    /* mark the object as modified since load */
    get_ext()->modified_ = TRUE;
}

/*
 *   property evaluation - set the truncation length
 */
int CVmObjDict::getp_settrunc(VMG_ vm_obj_id_t self,
                              vm_val_t *val, uint *argc)
{
    /* check arguments */
    if (get_prop_check_argc(val, argc, 1))
        return TRUE;

    /* if there's a global undo object, add undo for the change */
    if (G_undo != 0)
    {
        dict_undo_rec *undo_rec;
        
        /* create the undo record with the original truncation length */
        undo_rec = alloc_undo_rec(DICT_UNDO_TRUNCLEN, 0, 0);
        undo_rec->trunc_len = get_ext()->trunc_len_;

        /* add the undo record */
        add_undo_rec(vmg_ self, undo_rec);
    }

    /* set the new truncation length from the argument */
    set_trunc_len(vmg_ (size_t)CVmBif::pop_int_val(vmg0_));

    /* mark the object as modified since load */
    get_ext()->modified_ = TRUE;

    /* no return value - just return nil */
    val->set_nil();

    /* success */
    return TRUE;
}

/*
 *   property evaluation - check to see if a word is defined 
 */
int CVmObjDict::getp_is_defined(VMG_ vm_obj_id_t self, vm_val_t *val,
                                uint *argc)
{
    const char *str;
    size_t len;
    CVmHashEntryDict *entry;
    
    /* check arguments */
    if (get_prop_check_argc(val, argc, 1))
        return TRUE;

    /* pop the word string */
    str = CVmBif::pop_str_val(vmg0_);

    /* get and skip the length prefix */
    len = vmb_get_len(str);
    str += VMB_LEN;

    /* look up the word */
    entry = (CVmHashEntryDict *)get_ext()->hashtab_->find(str, len);

    /* 
     *   if the entry exists, and has any associated objects, the word is
     *   defined, so return true; otherwise, return nil 
     */
    val->set_logical(entry != 0 && entry->get_head() != 0);

    /* success */
    return TRUE;
}

/*
 *   property evaluation - check to see if a word is defined 
 */
int CVmObjDict::getp_is_defined_trunc(VMG_ vm_obj_id_t self, vm_val_t *val,
                                      uint *argc)
{
    const char *str;
    size_t len;
    find_trunc_ctx ctx;

    /* check arguments */
    if (get_prop_check_argc(val, argc, 1))
        return TRUE;

    /* pop the word string */
    str = CVmBif::pop_str_val(vmg0_);

    /* get and skip the length prefix */
    len = vmb_get_len(str);
    str += VMB_LEN;

    /* 
     *   if the given word is shorter than the truncation length, there
     *   can be no match; likewise, if we're not using truncation at all,
     *   there can be no truncation match 
     */
    if (len <= get_ext()->trunc_len_ || get_ext()->trunc_len_ == 0)
    {
        /* indicate that the string is not a match for any truncated word */
        val->set_nil();
        return TRUE;
    }

    /* 
     *   enumerate everything that matches the truncated value - use the
     *   first phase of the enumerator, so we just count the matches 
     */
    ctx.str = str;
    ctx.len = len;
    ctx.voc_prop = VM_INVALID_PROP;
    ctx.lst = 0;
    ctx.cnt = 0;
    ctx.cb = 0;
    get_ext()->hashtab_->enum_hash_matches(str, len, &find_trunc_cb, &ctx);

    /* if we found any matches, return true; otherwise, return nil */
    val->set_logical(ctx.cnt != 0);

    /* success */
    return TRUE;
}

/*
 *   Create an undo record 
 */
dict_undo_rec *CVmObjDict::alloc_undo_rec(dict_undo_action action,
                                          const char *txt, size_t len)
{
    size_t alloc_size;
    dict_undo_rec *rec;
    
    /* 
     *   compute the allocation size - start with the structure size, and
     *   add in the length we need to store the string if one is present 
     */
    alloc_size = sizeof(dict_undo_rec) - 1;
    if (txt != 0)
        alloc_size += len;

    /* allocate the record */
    rec = (dict_undo_rec *)t3malloc(alloc_size);

    /* set the action */
    rec->action = action;

    /* if there's a word, copy the text */
    if (txt != 0)
        memcpy(rec->txt, txt, len);

    /* save the word length */
    rec->len = len;

    /* presume we won't store a truncation length, object, or property */
    rec->trunc_len = 0;
    rec->obj = VM_INVALID_OBJ;
    rec->prop = VM_INVALID_PROP;

    /* return the new record */
    return rec;
}

/*
 *   add an undo record 
 */
void CVmObjDict::add_undo_rec(VMG_ vm_obj_id_t self, dict_undo_rec *rec)
{
    vm_val_t val;

    /* add the record with an empty value */
    val.set_empty();
    if (!G_undo->add_new_record_ptr_key(vmg_ self, rec, &val))
    {
        /* 
         *   we didn't add an undo record, so our extra undo information
         *   isn't actually going to be stored in the undo system - hence
         *   we must delete our extra information 
         */
        t3free(rec);
    }
}

/* 
 *   apply undo 
 */
void CVmObjDict::apply_undo(VMG_ CVmUndoRecord *undo_rec)
{
    if (undo_rec->id.ptrval != 0)
    {
        dict_undo_rec *rec;

        /* get my undo record */
        rec = (dict_undo_rec *)undo_rec->id.ptrval;

        /* take the appropriate action */
        switch(rec->action)
        {
        case DICT_UNDO_ADD:
            /* we added the word, so we must now delete it */
            del_hash_entry(vmg_ rec->txt, rec->len, rec->obj, rec->prop);
            break;

        case DICT_UNDO_DEL:
            /* we deleted the word, so now we must add it */
            add_hash_entry(vmg_ rec->txt, rec->len, TRUE,
                           rec->obj, rec->prop, FALSE);
            break;

        case DICT_UNDO_TRUNCLEN:
            /* we changed the truncation length, so change it back */
            set_trunc_len(vmg_ rec->trunc_len);
            break;
        }

        /* discard my record */
        t3free(rec);

        /* clear the pointer in the undo record so we know it's gone */
        undo_rec->id.ptrval = 0;
    }
}

/*
 *   discard extra undo information 
 */
void CVmObjDict::discard_undo(VMG_ CVmUndoRecord *rec)
{
    /* delete our extra information record */
    if (rec->id.ptrval != 0)
    {
        /* free the record */
        t3free((dict_undo_rec *)rec->id.ptrval);

        /* clear the pointer so we know it's gone */
        rec->id.ptrval = 0;
    }
}

/* 
 *   remove stale weak references from an undo record
 */
void CVmObjDict::remove_stale_undo_weak_ref(VMG_ CVmUndoRecord *undo_rec)
{
    if (undo_rec->id.ptrval != 0)
    {
        dict_undo_rec *rec;

        /* get my undo record */
        rec = (dict_undo_rec *)undo_rec->id.ptrval;

        /* take the appropriate action */
        switch(rec->action)
        {
        case DICT_UNDO_ADD:
        case DICT_UNDO_DEL:
            /* check to see if our object is being deleted */
            if (rec->obj != VM_INVALID_OBJ
                && G_obj_table->is_obj_deletable(rec->obj))
            {
                /* 
                 *   the object is being deleted - since we only weakly
                 *   reference our objects, we must forget about this
                 *   object now, which means we can forget this entire
                 *   record 
                 */
                t3free(rec);
                undo_rec->id.ptrval = 0;
            }
            break;

        case DICT_UNDO_TRUNCLEN:
            /* 
             *   this doesn't store an object reference - there's nothing
             *   to do here 
             */
            break;
        }
    }
}

/*
 *   callback context for weak reference function 
 */
struct enum_hashtab_ctx
{
    /* global context pointer */
    vm_globals *vmg;

    /* dictionary object */
    CVmObjDict *dict;
};

/* 
 *   Remove weak references that have become stale.  For each object for
 *   which we have a weak reference, check to see if the object is marked
 *   for deletion; if so, remove our reference to the object.  
 */
void CVmObjDict::remove_stale_weak_refs(VMG0_)
{
    enum_hashtab_ctx ctx;
    
    /* iterate over our hash table */
    ctx.vmg = VMGLOB_ADDR;
    ctx.dict = this;
    get_ext()->hashtab_->enum_entries(&remove_weak_ref_cb, &ctx);
}

/*
 *   enumeration callback to eliminate stale weak references 
 */
void CVmObjDict::remove_weak_ref_cb(void *ctx0, CVmHashEntry *entry0)
{
    enum_hashtab_ctx *ctx = (enum_hashtab_ctx *)ctx0;
    CVmHashEntryDict *entry = (CVmHashEntryDict *)entry0;
    vm_dict_entry *cur;
    vm_dict_entry *nxt;

    /* set up for global access through the context */
    VMGLOB_PTR(ctx->vmg);

    /* scan the entry's list */
    for (cur = entry->get_head() ; cur != 0 ; cur = nxt)
    {
        /* remember the next entry in case we delete this one */
        nxt = cur->nxt_;
        
        /* check to see if this object is free - if so, remove this record */
        if (G_obj_table->is_obj_deletable(cur->obj_))
        {
            /* delete the entry */
            entry->del_entry(ctx->dict->get_ext()->hashtab_,
                             cur->obj_, cur->prop_);
        }
    }
}


/* 
 *   load from an image file 
 */
void CVmObjDict::load_from_image(VMG_ vm_obj_id_t,
                                 const char *ptr, size_t siz)
{
    /* remember where our image data comes from */
    get_ext()->image_data_ = ptr;
    get_ext()->image_data_size_ = siz;

    /* build the initial hash table from the image data */
    build_hash_from_image(vmg0_);
}

/*
 *   Build the hash table from the image data 
 */
void CVmObjDict::build_hash_from_image(VMG0_)
{
    uint cnt;
    uint i;
    const char *p;
    const char *endp;

    /* start reading at the start of the image data */
    p = get_ext()->image_data_;
    endp = p + get_ext()->image_data_size_;

    /* set the truncation length from the image data */
    get_ext()->trunc_len_ = osrp2(p);
    p += 2;

    /* create the new hash table */
    create_hash_table(vmg0_);

    /* read the entry count */
    cnt = osrp2(p);
    p += 2;

    /* scan the entries */
    for (i = 0 ; p < endp && i < cnt ; ++i)
    {
        vm_obj_id_t obj;
        vm_prop_id_t prop;
        uint key_len;
        uint item_cnt;
        const char *key;
        CVmHashEntryDict *entry;
        char key_buf[256];
        char *keyp;
        size_t key_rem;

        /* read the key length */
        key_len = *(unsigned char *)p++;

        /* remember the key pointer */
        key = p;
        p += key_len;

        /* copy the key to our buffer for decoding */
        memcpy(key_buf, key, key_len);

        /* decode the key */
        for (keyp = key_buf, key_rem = key_len ; key_rem != 0 ;
             --key_rem, ++keyp)
            *keyp ^= 0xBD;

        /* read the key from the decoded buffer now */
        key = key_buf;

        /* read the item count */
        item_cnt = osrp2(p);
        p += 2;

        /* find an existing entry for this key */
        entry = (CVmHashEntryDict *)get_ext()->hashtab_->find(key, key_len);

        /* if there's no existing entry, add one */
        if (entry == 0)
        {
            /* create a new entry */
            entry = new CVmHashEntryDict(key, key_len, TRUE, TRUE);

            /* add it to the table */
            get_ext()->hashtab_->add(entry);
        }

        /* read the items */
        for ( ; item_cnt != 0 ; --item_cnt)
        {
            /* read the object ID and property ID */
            obj = (vm_obj_id_t)osrp4(p);
            prop = (vm_prop_id_t)osrp2(p + 4);

            /* 
             *   add the entry - this entry is from the image file, so
             *   mark it as such 
             */
            entry->add_entry(obj, prop, TRUE);

            /* move on to the next item */
            p += 6;
        }
    }
}

/*
 *   Add an entry 
 */
int CVmObjDict::add_hash_entry(VMG_ const char *p, size_t len, int copy,
                               vm_obj_id_t obj, vm_prop_id_t prop,
                               int from_image)
{
    CVmHashEntryDict *entry;
    
    /* find an existing entry for this key */
    entry = (CVmHashEntryDict *)get_ext()->hashtab_->find(p, len);

    /* if we didn't find an entry, add one */
    if (entry == 0)
    {
        /* create an entry */
        entry = new CVmHashEntryDict(p, len, copy, from_image);

        /* add it to the table */
        get_ext()->hashtab_->add(entry);
    }

    /* add the obj/prop to the entry's item list */
    return entry->add_entry(obj, prop, from_image);
}

/*
 *   delete an entry 
 */
int CVmObjDict::del_hash_entry(VMG_ const char *str, size_t len,
                               vm_obj_id_t obj, vm_prop_id_t voc_prop)
{
    CVmHashEntryDict *entry;

    /* find the hash table entry for the word */
    entry = (CVmHashEntryDict *)get_ext()->hashtab_->find(str, len);

    /* if we found it, delete the obj/prop entry */
    if (entry != 0)
    {
        /* delete the entry */
        return entry->del_entry(get_ext()->hashtab_, obj, voc_prop);
    }

    /* we didn't find anything to delete */
    return FALSE;
}

/* 
 *   restore to image file state 
 */
void CVmObjDict::reset_to_image(VMG0_)
{
    /* delete the hash table */
    if (get_ext()->hashtab_ != 0)
    {
        /* delete it */
        delete get_ext()->hashtab_;

        /* forget it */
        get_ext()->hashtab_ = 0;
    }

    /* rebuild the hash table from the image file data */
    build_hash_from_image(vmg0_);
}

/* 
 *   determine if the object has been changed since it was loaded 
 */
int CVmObjDict::is_changed_since_load() const
{
    /* 
     *   return true if our non-image-entry count is non-zero - if we have
     *   any non-image entries, we must have been changed since we were
     *   loaded 
     */
    return (get_ext()->modified_ != 0);
}

/*
 *   save-file enumeration context 
 */
struct save_file_ctx
{
    /* file object */
    CVmFile *fp;

    /* entry count */
    ulong cnt;

    /* globals */
    vm_globals *vmg;
};

/* 
 *   save to a file 
 */
void CVmObjDict::save_to_file(VMG_ CVmFile *fp)
{
    long cnt_pos;
    long end_pos;
    save_file_ctx ctx;

    /* write the current truncation length */
    fp->write_int2(get_ext()->trunc_len_);
    
    /* remember the starting seek position and write a placeholder count */
    cnt_pos = fp->get_pos();
    fp->write_int4(0);

    /* enumerate the entries to write out each one */
    ctx.fp = fp;
    ctx.cnt = 0;
    ctx.vmg = VMGLOB_ADDR;
    get_ext()->hashtab_->enum_entries(&save_file_cb, &ctx);

    /* remember our position for a moment */
    end_pos = fp->get_pos();

    /* go back and write out the symbol count prefix */
    fp->set_pos(cnt_pos);
    fp->write_int4(ctx.cnt);

    /* seek back to the end */
    fp->set_pos(end_pos);
}

/*
 *   save file enumeration callback 
 */
void CVmObjDict::save_file_cb(void *ctx0, CVmHashEntry *entry0)
{
    save_file_ctx *ctx = (save_file_ctx *)ctx0;
    CVmHashEntryDict *entry = (CVmHashEntryDict *)entry0;
    vm_dict_entry *cur;
    uint cnt;

    /* set up global access */
    VMGLOB_PTR(ctx->vmg);

    /* write out this entry's name */
    ctx->fp->write_int2(entry->getlen());
    ctx->fp->write_bytes(entry->getstr(), entry->getlen());

    /* count the items in this entry's list */
    for (cnt = 0, cur = entry->get_head() ; cur != 0 ; cur = cur->nxt_)
    {
        /* 
         *   if the referenced object is persistent, count this item;
         *   don't count items that refer to non-persistent items, because
         *   these objects will not be saved to the file and hence will
         *   not be present when the saved state is restored 
         */
        if (G_obj_table->is_obj_persistent(cur->obj_))
            ++cnt;
    }

    /* 
     *   if there are no saveable items for this entry, do not write out
     *   the entry at all 
     */
    if (cnt == 0)
        return;

    /* count this entry */
    ++ctx->cnt;

    /* write the item count */
    ctx->fp->write_int2(cnt);

    /* write out each item in this entry's list */
    for (cur = entry->get_head() ; cur != 0 ; cur = cur->nxt_)
    {
        /* if this item's object is persistent, write out the item */
        if (G_obj_table->is_obj_persistent(cur->obj_))
        {
            /* write out the object and property for this entry */
            ctx->fp->write_int4((long)cur->obj_);
            ctx->fp->write_int2((int)cur->prop_);
        }
    }
}

/* 
 *   restore from a file 
 */
void CVmObjDict::restore_from_file(VMG_ vm_obj_id_t,
                                   CVmFile *fp, CVmObject *obj)
{
    ulong cnt;

    /* delete the old hash table if we have one */
    if (get_ext()->hashtab_ != 0)
    {
        delete get_ext()->hashtab_;
        get_ext()->hashtab_ = 0;
    }

    /* read the truncation length */
    get_ext()->trunc_len_ = fp->read_uint2();

    /* create the new, empty hash table */
    create_hash_table(vmg0_);

    /* read the number of symbols */
    cnt = fp->read_int4();

    /* read the symbols */
    for ( ; cnt != 0 ; --cnt)
    {
        uint len;
        uint read_len;
        char buf[256];
        uint item_cnt;
        
        /* read the symbol length */
        len = fp->read_uint2();

        /* limit the reading to our buffer size */
        read_len = len;
        if (read_len > sizeof(buf))
            read_len = sizeof(buf);

        /* read the string */
        fp->read_bytes(buf, read_len);

        /* skip any extra data */
        if (len > read_len)
            fp->set_pos(fp->get_pos() + len - read_len);

        /* read the item count */
        item_cnt = fp->read_uint2();

        /* read the items */
        for ( ; item_cnt != 0 ; --item_cnt)
        {
            vm_obj_id_t obj;
            vm_prop_id_t prop;
            
            /* read the object and property ID's */
            obj = (vm_obj_id_t)fp->read_int4();
            prop = (vm_prop_id_t)fp->read_uint2();

            /* add the entry */
            add_hash_entry(vmg_ buf, len, TRUE, obj, prop, FALSE);
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   impedence-matcher callback context 
 */
struct enum_word_props_trunc_ctx
{
    /* client callback to invoke */
    void (*cb_func)(VMG_ void *ctx, vm_prop_id_t prop);

    /* client callback context */
    void *cb_ctx;

    /* globals */
    vm_globals *vmg;
};

/*
 *   impedence-matcher callback 
 */
static void enum_word_props_trunc_cb(void *ctx0, const char *txt, size_t len,
                                     vm_prop_id_t prop)
{
    enum_word_props_trunc_ctx *ctx = (enum_word_props_trunc_ctx *)ctx0;
    
    /* set up access to the VM globals through my stashed context */
    VMGLOB_PTR(ctx->vmg);

    /* invoke the callback */
    (*ctx->cb_func)(vmg_ ctx->cb_ctx, prop);
}

/*
 *   Enumerate the properties for which a word is defined 
 */
void CVmObjDict::enum_word_props(VMG_
                                 void (*cb_func)(VMG_ void *, vm_prop_id_t),
                                 void *cb_ctx, const char *txt, size_t len,
                                 int include_trunc)
{
    CVmHashEntryDict *entry;

    /* find the hash table entry for the word */
    entry = (CVmHashEntryDict *)get_ext()->hashtab_->find(txt, len);

    /* if we found an entry, enumerate its properties */
    if (entry != 0)
    {
        vm_dict_entry *cur;
        vm_dict_entry *prv;
        
        /* scan the entry's list of associations */
        for (prv = 0, cur = entry->get_head() ; cur != 0 ;
             prv = cur, cur = cur->nxt_)
        {
            /* 
             *   Unless this is the first item, check to see if its
             *   property matches the previous item's property; if so,
             *   don't bother calling the callback.  Because we keep the
             *   list grouped by property ID, this ensures that we only
             *   call the callback once per unique property in our list. 
             */
            if (prv == 0 || cur->prop_ != prv->prop_)
                (*cb_func)(vmg_ cb_ctx, cur->prop_);
        }
    }

    /* enumerate the truncated matches as well, if desired */
    if (include_trunc)
    {
        find_trunc_ctx ctx;
        enum_word_props_trunc_ctx trunc_ctx;

        /* set up the context for our impedence-matcher callback */
        trunc_ctx.cb_func = cb_func;
        trunc_ctx.cb_ctx = cb_ctx;
        trunc_ctx.vmg = VMGLOB_ADDR;

        /* set up the callback context */
        ctx.str = txt;
        ctx.len = len;
        ctx.voc_prop = VM_INVALID_PROP;
        ctx.lst = 0;
        ctx.cnt = 0;
        ctx.cb = &enum_word_props_trunc_cb;
        ctx.cbctx = &trunc_ctx;
        
        /* find truncated matches */
        get_ext()->hashtab_
            ->enum_hash_matches(txt, len, &find_trunc_cb, &ctx);
    }
}

