#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/vmerr.cpp,v 1.4 1999/07/11 00:46:59 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmerr.cpp - VM error handling
Function
  This module contains global variables required for the error handler.
Notes
  
Modified
  10/20/98 MJRoberts  - Creation
*/

#include <stdarg.h>
#include <stdlib.h>

#include "t3std.h"
#include "vmtype.h"
#include "vmerr.h"

/*
 *   Global error context pointer
 */
err_context_t *G_err;

/*
 *   Initialize the global error context 
 */
void err_init(size_t param_stack_size)
{
    /* ignore this if we're already initialized */
    if (G_err != 0)
        return;
    
    /* 
     *   allocate the error context, adding in room for the parameter
     *   stack 
     */
    G_err = (err_context_t *)t3malloc(sizeof(err_context_t)
                                      + param_stack_size);

    /* initialize the parameter stack pointers */
    G_err->param_free_ = G_err->param_stack_;
    G_err->param_stack_size_ = param_stack_size;
    G_err->cur_exc_ = 0;

    /* we have no active frame yet */
    G_err->cur_frame_ = 0;
}

/*
 *   Delete the global error context 
 */
void err_terminate()
{
    t3free(G_err);
    G_err = 0;
}

/*
 *   Throw the error currently on the stack 
 */
static void err_throw_current()
{
    err_state_t new_state;

    /*
     *   Figure out what state the enclosing frame will be in after this
     *   jump.  If we're currently in a 'trying' block, we'll now be in an
     *   'exception' state.  Otherwise, we must have been in a 'catch' or
     *   'finally' already - in these cases, the new state is 'rethrown',
     *   because we have an exception within an exception handler.  
     */
    if (G_err->cur_frame_->state_ == ERR_STATE_TRYING)
        new_state = ERR_STATE_EXCEPTION;
    else
        new_state = ERR_STATE_RETHROWN;

    /* jump to the enclosing frame's exception handler */
    longjmp(G_err->cur_frame_->jmpbuf_, new_state);
}

/*
 *   Throw an exception 
 */
void err_throw(err_id_t error_code)
{
    /* throw the error, with no parameters */
    err_throw_a(error_code, 0);
}

/*
 *   Throw an exception with parameters in va_list format
 */
void err_throw_v(err_id_t error_code, int param_count, va_list va)
{
    size_t siz;
    int i;
    va_list marker;
    CVmException *exc;
    CVmExcParam *param;
    char *vp;
    err_param_type typ;

    /*
     *   If the current stack frame is already handling an error, pop the
     *   current error, since the current error will no longer be
     *   accessible after this throw.  
     */
    if (G_err->cur_frame_->state_ == ERR_STATE_EXCEPTION
        || G_err->cur_frame_->state_ == ERR_STATE_CAUGHT
        || G_err->cur_frame_->state_ == ERR_STATE_RETHROWN)
        err_pop_exc();

    /* 
     *   Figure out how much space we need.  We always need a CVmException
     *   structure, and if we have parameters we need space for the
     *   parameters and their descriptors as well.  
     */
    siz = sizeof(CVmException);

    /* figure out how much space we need for the parametres */
    marker = va;
    for (i = 0 ; i < param_count ; ++i)
    {
        /* add in the size of the descriptor structure */
        siz += sizeof(CVmExcParam);

        /* get the type marker */
        typ = va_arg(marker, err_param_type);

        /* see what we have */
        switch(typ)
        {
        case ERR_TYPE_INT:
            /* skip the integer */
            va_arg(marker, int);

            /* we need space for an int */
            siz += osrndsz(sizeof(int));
            break;

        case ERR_TYPE_ULONG:
            /* skip the unsigned long */
            va_arg(marker, unsigned long);

            /* make room for it */
            siz += osrndsz(sizeof(unsigned long));
            break;

        case ERR_TYPE_TEXTCHAR:
            {
                textchar_t *s;

                /* get the string */
                s = va_arg(marker, textchar_t *);

                /* add in space for its bytes */
                siz += osrndsz(get_strlen(s) + sizeof(textchar_t));
            }
            break;

        case ERR_TYPE_TEXTCHAR_LEN:
            /* skip the string */
            va_arg(marker, textchar_t *);

            /* get the length and add space for it plus a null terminator */
            siz += osrndsz(va_arg(marker, size_t) + sizeof(textchar_t));
            break;

        case ERR_TYPE_CHAR:
            {
                char *s;

                /* get the string */
                s = va_arg(marker, char *);

                /* add in space for its bytes */
                siz += osrndsz(strlen(s) + 1);
            }
            break;
        }
    }

    /* if we don't have room, we have a serious problem */
    if ((G_err->param_stack_size_
         - (G_err->param_free_ - G_err->param_stack_)) < siz)
        err_abort("no space for exception parameters");

    /* put our new structure at the first free byte */
    exc = (CVmException *)G_err->param_free_;

    /* consume the space */
    G_err->param_free_ += siz;

    /* fill in our base structure */
    exc->error_code_ = error_code;
    exc->param_count_ = param_count;

    /* 
     *   link to the previous exception on the stack, so that we can pop
     *   this exception when we're done with it 
     */
    exc->prv_ = G_err->cur_exc_;

    /* this is now the current exception */
    G_err->cur_exc_ = exc;

    /* 
     *   Put varying-size parameters right after the exception structure.
     *   The end of the structure is the start of the next parameter array
     *   entry after the last entry we're actually using. 
     */
    vp = (char *)(&exc->params_[param_count]);

    /* fill in the parameters */
    marker = va;
    for (i = 0, param = exc->params_ ; i < param_count ; ++i, ++param)
    {
        /* get the type marker */
        typ = va_arg(marker, err_param_type);

        /* store the type */
        param->type_ = typ;

        /* see what we have */
        switch(typ)
        {
        case ERR_TYPE_INT:
            /* store the integer */
            param->val_.intval_ = va_arg(marker, int);
            break;

        case ERR_TYPE_ULONG:
            /* store the unsigned long */
            param->val_.ulong_ = va_arg(marker, unsigned long);
            break;

        case ERR_TYPE_TEXTCHAR:
            {
                textchar_t *s;
                size_t len;

                /* get the string */
                s = va_arg(marker, textchar_t *);
                len = get_strlen(s) + sizeof(textchar_t);

                /* store it in the variable part */
                param->val_.strval_ = (textchar_t *)vp;
                memcpy(vp, s, len);

                /* consume the space */
                vp += osrndsz(len);
            }
            break;

        case ERR_TYPE_TEXTCHAR_LEN:
            {
                textchar_t *s;
                size_t len;

                /* get the string and its length */
                s = va_arg(marker, textchar_t *);
                len = va_arg(marker, size_t);

                /* store the string and add a null terminator */
                param->val_.strval_ = (textchar_t *)vp;
                memcpy(vp, s, len);
                vp[len] = '\0';

                /* consume the space */
                vp += osrndsz(len + sizeof(textchar_t));

                /* 
                 *   change the type to a regular textchar string now,
                 *   since we've converted the value to a null-terminated
                 *   string 
                 */
                param->type_ = ERR_TYPE_TEXTCHAR;
            }
            break;

        case ERR_TYPE_CHAR:
            {
                char *s;
                size_t len;

                /* get the string */
                s = va_arg(marker, char *);
                len = strlen(s) + 1;

                /* store it in the variable part */
                param->val_.charval_ = (char *)vp;
                memcpy(vp, s, len);

                /* consume the space */
                vp += osrndsz(len);
            }
            break;
        }
    }

    /* throw the error that we just pushed */
    err_throw_current();
}

/*
 *   Throw an exception with parameters 
 */
void err_throw_a(err_id_t error_code, int param_count, ...)
{
    va_list marker;

    /* build the va_list and throw the error */
    va_start(marker, param_count);
    err_throw_v(error_code, param_count, marker);
    va_end(marker);
}

/*
 *   Re-throw the current exception.  This is valid only from 'catch'
 *   blocks.  
 */
void err_rethrow()
{
    /* throw the error currently on the stack */
    err_throw_current();
}

/*
 *   Pop the current exception from the exception stack. 
 */
void err_pop_exc()
{
    /* 
     *   if there's anything on the stack, remove it and replace it with
     *   the previous item on the stack 
     */
    if (G_err->cur_exc_ != 0)
    {
        /* 
         *   since we're removing this element, its space can now be
         *   re-used for the next exception allocation (since exceptions
         *   are always stacked, the space we use to consume them can be
         *   treated as a stack as well) 
         */
        G_err->param_free_ = (char *)G_err->cur_exc_;

        /* remove the top element of the stack */
        G_err->cur_exc_ = G_err->cur_exc_->prv_;
    }
    else
    {
        /* no errors are on the stack, so the parameter space is all free */
        G_err->param_free_ = G_err->param_stack_;
    }
}

/*
 *   Abort the program with a serious, unrecoverable error
 */
void err_abort(const char *message)
{
    printf("%s\n", message);
    exit(2);
}

/*
 *   Determine the current error stack depth 
 */
int err_stack_depth()
{
    int cnt;
    CVmException *exc;

    /* count exceptions in the stack */
    for (cnt = 0, exc = G_err->cur_exc_ ; exc != 0 ; exc = exc->prv_)
    {
        /* count this exception */
        ++cnt;
    }

    /* return the number of exceptions in the stack */
    return cnt;
}

/* ------------------------------------------------------------------------ */
/*
 *   Find an error message 
 */
const char *err_get_msg(const err_msg_t *msg_array, size_t msg_count,
                        int msgnum, int verbose)
{
    int hi, lo, cur;

    /* perform a binary search of the message list */
    lo = 0;
    hi = msg_count - 1;
    while (lo <= hi)
    {
        /* split the difference */
        cur = lo + (hi - lo)/2;

        /* is it a match? */
        if (msg_array[cur].msgnum == msgnum)
        {
            /* it's the one - return the text */
            return (verbose
                    ? msg_array[cur].long_msgtxt
                    : msg_array[cur].short_msgtxt);
        }
        else if (msgnum > msg_array[cur].msgnum)
        {
            /* we need to go higher */
            lo = (cur == lo ? cur + 1 : cur);
        }
        else
        {
            /* we need to go lower */
            hi = (cur == hi ? hi - 1 : cur);
        }
    }

    /* no such message */
    return 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Format a message with the parameters contained in an exception
 *   object.  Suports the following format codes:
 *   
 *   %s - String.  Formats an ERR_TYPE_CHAR or ERR_TYPE_TEXTCHAR value.
 *   
 *   %d, %u, %x - signed/unsigned decimal integer, hexadecimal integer.
 *   Formats an ERR_TYPE_INT value or an ERR_TYPE_ULONG value.
 *   Automatically uses the correct size for the argument.
 *   
 *   %% - Formats as a single percent sign.  
 */
void err_format_msg(char *outbuf, size_t outbuflen,
                    const char *msg, const CVmException *exc)
{
    int curarg;
    const char *p;
    char *dst;
    int exc_argc;

    /* if there's no space at all, ignore the request */
    if (outbuf == 0 || outbuflen == 0)
        return;

    /* get the number of parameters in the exception object */
    exc_argc = (exc == 0 ? 0 : exc->get_param_count());

    /* start with the first parameter */
    curarg = 0;

    /* start at the beginning of the buffer */
    dst = outbuf;

    /* if there's no message, there's nothing to return */
    if (msg == 0)
    {
        *dst = '\0';
        return;
    }

    /* scan the format string for formatting codes */
    for (p = msg ; *p != '\0' ; ++p)
    {
        /* 
         *   If we're out of space, stop now.  Make sure we leave room for
         *   the terminating null byte. 
         */
        if (dst + 1 >= outbuf + outbuflen)
            break;
        
        /* if it's a format specifier, translate it */
        if (*p == '%')
        {
            const char *src;
            char srcbuf[30];
            err_param_type typ;
            size_t len;
            
            /* 
             *   if no more parameters are available, ignore the
             *   formatting code entirely, and leave it in the string as
             *   it is 
             */
            if (curarg >= exc_argc)
            {
                *dst++ = *p;
                continue;
            }

            /* get the type of the current parameter */
            typ = exc->get_param_type(curarg);
            
            /* skip the '%' and determine what follows */
            ++p;
            switch (*p)
            {
            case 's':
                /* get the string value using the appropriate type */
                if (typ == ERR_TYPE_TEXTCHAR)
                    src = exc->get_param_text(curarg);
                else if (typ == ERR_TYPE_CHAR)
                    src = exc->get_param_char(curarg);
                else
                    src = "s";
                break;

            case 'd':
                src = srcbuf;
                if (typ == ERR_TYPE_INT)
                    sprintf(srcbuf, "%d", exc->get_param_int(curarg));
                else if (typ == ERR_TYPE_ULONG)
                    sprintf(srcbuf, "%ld", exc->get_param_ulong(curarg));
                else
                    src = "d";
                break;

            case 'u':
                src = srcbuf;
                if (typ == ERR_TYPE_INT)
                    sprintf(srcbuf, "%u", exc->get_param_int(curarg));
                else if (typ == ERR_TYPE_ULONG)
                    sprintf(srcbuf, "%lu", exc->get_param_ulong(curarg));
                else
                    src = "u";
                break;

            case 'x':
                src = srcbuf;
                if (typ == ERR_TYPE_INT)
                    sprintf(srcbuf, "%x", exc->get_param_int(curarg));
                else if (typ == ERR_TYPE_ULONG)
                    sprintf(srcbuf, "%lx", exc->get_param_ulong(curarg));
                else
                    src = "x";
                break;

            case '%':
                /* add a single percent sign */
                src = "%";
                break;
            }

            /* figure out how much of the value we can copy */
            len = strlen(src);
            if (len > outbuflen - (dst - outbuf) - 1)
                len = outbuflen - (dst - outbuf) - 1;

            /* copy the value and advance past it in the output buffer */
            memcpy(dst, src, len);
            dst += len;

            /* consume the argument */
            ++curarg;
        }
        else
        {
            /* just copy the current character as it is */
            *dst++ = *p;
        }
    }

    /* add the trailing null */
    *dst++ = '\0';
}


/* ------------------------------------------------------------------------ */
/*
 *   T3 VM Error Messages 
 *   
 *   The messages must be sorted by message number, so that we can perform
 *   a binary search to look up a message by number.  
 */
const err_msg_t vm_messages[] =
{
    { VMERR_READ_FILE,
    "error reading file",
    "Error reading file.  The file might be corrupted or a media error "
    "might have occurred." },

    { VMERR_WRITE_FILE,
    "error writing file",
    "Error writing file.  The media might be full, or another media "
    "error might have occurred." },

    { VMERR_FILE_NOT_FOUND,
    "file not found",
    "Error opening file.  The specified file might not exist, you "
    "might not have sufficient privileges to open the file, or "
    "a sharing violation might have occurred." },

    { VMERR_CREATE_FILE,
    "error creating file",
    "Error creating file.  You might not have sufficient privileges "
    "to open the file, or a sharing violation might have occurred." },

    { VMERR_OBJ_IN_USE,
    "object ID in use - the image/save file might be corrupted",
    "An object ID requested by the image/save file is already in use "
    "and cannot be allocated to the file.  This might indicate that "
    "the file is corrupted." },

    { VMERR_OUT_OF_MEMORY,
    "out of memory",
    "Out of memory.  Try making more memory available by closing "
    "other applications if possible." },

    { VMERR_NO_MEM_FOR_PAGE,
    "out of memory allocating pool page",
    "Out of memory allocating pool page.  Try making more memory available "
    "by closing other applications." },

    { VMERR_BAD_POOL_PAGE_SIZE,
    "invalid page size - file is not valid",
    "Invalid page size.  The file being loaded is not valid." },

    { VMERR_UNKNOWN_METACLASS,
    "intrinsic class \"%s\" is not available",
    "This image file requires an intrinsic class with the identifier \"%s\", "
    "but the class is not available in this interpreter.  This program "
    "cannot be executed with this interpreter." },

    { VMERR_UNKNOWN_FUNC_SET,
    "intrinsic function set \"%s\" is not available",
    "This image file requires a function set with the identifier \"%s\", "
    "but the function set is not available in this intepreter.  This program "
    "cannot be executed with this interpreter." },

    { VMERR_READ_PAST_IMG_END,
    "reading past end of image file",
    "Reading past end of image file.  The image file might be corrupted." },

    { VMERR_NOT_AN_IMAGE_FILE,
    "not an image file (invalid signature)",
    "This file is not a valid image file - the file has an invalid "
    "signature.  The image file might be corrupted." },

    { VMERR_UNKNOWN_IMAGE_BLOCK,
    "unknown block type - image file is incompatible with this interpreter "
        "version",
    "Unknown block type.  This image file is incompatible with this "
    "version of the interpreter." },

    { VMERR_IMAGE_BLOCK_TOO_SMALL,
    "data block too small",
    "A data block in the image file is too small.  The image file might "
    "be corrupted." },

    { VMERR_IMAGE_POOL_BEFORE_DEF,
    "invalid image file: pool page before pool definition",
    "This image file is invalid because it specifies a pool page "
    "before the pool's definition.  The image file might be corrupted." },

    { VMERR_IMAGE_POOL_BAD_PAGE,
    "invalid image file: pool page our of range of definition",
    "This image file is invalid because it specifies a pool page "
    "outside of the range of the pool's definition.  The image file "
    "might be corrupted." },

    { VMERR_IMAGE_BAD_POOL_ID,
    "invalid image file: invalid pool ID",
    "This image file is invalid because it specifies an invalid "
    "pool ID.  The image file might be corrupted." },

    { VMERR_LOAD_BAD_PAGE_IDX,
    "invalid image file: bad page index",
    "This image file is invalid because it specifies an invalid "
    "page index.  The image file might be corrupted." },

    { VMERR_LOAD_UNDEF_PAGE,
    "loading undefined pool page",
    "The program is attempting to load a pool page that is not present "
    "in the image file.  The image file might be corrupted." },

    { VMERR_IMAGE_POOL_REDEF,
    "invalid image file: pool is defined more than once",
    "This image file is invalid because it defines a pool more than "
    "once.  The image file might be corrupted." },

    { VMERR_IMAGE_METADEP_REDEF,
    "invalid image file: multiple intrinsic class dependency tables found",
    "This image file is invalid because it contains multiple intrinsic class "
    "tables.  The image file might be corrupted." },

    { VMERR_IMAGE_NO_METADEP,
    "invalid image file: no intrinsic class dependency table found",
    "This image file is invalid because it contains no intrinsic class "
    "tables.  The image file might be corrupted." },

    { VMERR_IMAGE_FUNCDEP_REDEF,
    "invalid image file: multiple function set dependency tables found",
    "This image file is invalid because it contains multiple function set "
    "tables.  The image file might be corrupted." },

    { VMERR_IMAGE_NO_FUNCDEP,
    "invalid image file: no function set dependency table found",
    "This image file is invalid because it contains no function set "
    "tables.  The image file might be corrupted." },

    { VMERR_IMAGE_ENTRYPT_REDEF,
    "invalid image file: multiple entrypoints found",
    "This image file is invalid because it contains multiple entrypoint "
    "definitions.  The image file might be corrupted." },

    { VMERR_IMAGE_NO_ENTRYPT,
    "invalid image file: no entrypoint found",
    "This image file is invalid because it contains no entrypoint "
    "specification.  The image file might be corrupted." },

    { VMERR_IMAGE_INCOMPAT_VSN,
    "incompatible image file format version",
    "This image file has an incompatible format version.  You must "
    "obtain a newer version of the interpreter to execute this "
    "program." },

    { VMERR_IMAGE_NO_CODE,
    "image contains no code",
    "This image file contains no executable code.  The file might "
    "be corrupted." },

    { VMERR_IMAGE_INCOMPAT_HDR_FMT,
    "incomptabile image file format: method header too old",
    "This image file has an incompatible method header format.  "
    "This is an older image file version which this interpreter "
    "does not support." },

    { VMERR_UNAVAIL_INTRINSIC,
    "unavailable intrinsic funtion called (index %d in function set \"%s\"",
    "Unavailable intrinsic function called (the function is at "
    "index %d in function set \"%s\").  This function is not available "
    "in this version of the interpreter and cannot be called when running "
    "the program with this version.  This normally indicates that the "
    "\"preinit\" function, or code invoked by preinit, called an intrinsic "
    "that is invalid during compile-time pre-initialization." },

    { VMERR_UNKNOWN_METACLASS_INTERNAL,
    "unknown internal intrinsic class ID %x",
    "Unknown internal intrinsic class ID %x.  This indicates an internal "
    "error in the interpreter.  Please report this error to the "
    "interpreter's maintainer." },

    { VMERR_XOR_MASK_BAD_IN_MEM,
    "page mask is not allowed for in-memory image file",
    "This image file cannot be loaded from memory because it contains "
    "masked data.  Masked data is not valid with in-memory files.  This "
    "probably indicates that the program file was not installed properly; "
    "you must convert this program file for in-memory use before you can "
    "load the program with this version of the interpreter." },

    { VMERR_NO_IMAGE_IN_EXE,
    "no embedded image file found in executable",
    "This executable does not contain an embedded image file.  The "
    "application might not be configured properly or might need to "
    "be rebuilt.  Re-install the application or obtain an updated "
    "version from the application's author." },

    { VMERR_OBJ_SIZE_OVERFLOW,
    "object size exceeds hardware limits of this computer",
    "An object defined in this program file exceeds the hardware limits "
    "of this computer.  This program cannot be executed on this type "
    "of computer or operating system.  Contact the program's author "
    "for assistance." },

    { VMERR_METACLASS_TOO_OLD,
    "intrinsic class \"%s\" version is not available - latest available "
        "version is \"%s\"",
    "This program needs the intrinsic class \"%s\".  This VM implementation "
    "does not provide a sufficiently recent version of this intrinsic class; "
    "the latest version available in this VM is \"%s\".  This program "
    "cannot run with this version of the VM; you must use a more recent "
    "version of the VM to execute this program." },

    { VMERR_INVAL_METACLASS_DATA,
    "invalid intrinsic class data - image file may be corrupted",
    "Invalid data were detected in an intrinsic class.  This might indicate "
    "that the image file has been corrupted.  You might need to "
    "re-install the program." },

    { VMERR_BAD_STATIC_NEW,
    "invalid object - class does not allow loading",
    "An object in the image file cannot be loaded because its class "
    "does not allow creation of objects of the class.  This usually "
    "means that the class is abstract and cannot be instantiated "
    "as a concrete object." },

    { VMERR_FUNCSET_TOO_OLD,
    "function set \"%s\" version is not available - latest available "
        "version is \"%s\"",
    "This program needs the function set \"%s\".  This VM implementation "
    "does not provide a sufficiently recent version of this function set; "
    "the latest version available in this VM is \"%s\".  This program "
    "cannot run with this version of the VM; you must use a more recent "
    "version of the VM to execute this program." },
    

    { VMERR_INVALID_SETPROP,
    "property cannot be set for object",
    "Invalid property change - this property cannot be set for this "
    "object.  This normally indicates that the object is of a type "
    "that does not allow setting of properties at all, or at least "
    "of certain properties.  For example, a string object does not "
    "allow setting properties at all." },

    { VMERR_NOT_SAVED_STATE,
    "file is not a valid saved state file",
    "This file is not a valid saved state file.  Either the file was "
    "not created as a saved state file, or its contents have been "
    "corrupted." },

    { VMERR_WRONG_SAVED_STATE,
    "saved state is for a different program or a different version of "
        "this program",
    "This file does not contain saved state information for this program.  "
    "The file was saved by another program, or by a different version of "
    "this program; in either case, it cannot be restored with this version "
    "of this program." },

    { VMERR_SAVED_META_TOO_LONG,
    "intrinsic class name in saved state file is too long",
    "An intrinsic class name in the saved state file is too long.  The file "
    "might be corrupted, or might have been saved by an incompatible "
    "version of the interpreter." },

    { VMERR_RECURSIVE_VM_CALL,
    "operation is not valid in a recursive VM invocation",
    "The operation is not valid in the current "
    "context, because the VM has been invoked recursively.  This "
    "indicates an error in the program." },

    { VMERR_SAVED_STACK_OVERFLOW,
    "stack overflow restoring saved state file",
    "The stack stored in the saved state file is too large for this "
    "interpreter.  The state file might have been saved by an incompatible "
    "version of the interpreter." },

    { VMERR_SAVED_OBJ_ID_INVALID,
    "invalid object ID in saved state file",
    "The saved state file contains an invalid object ID.  The saved "
    "state file might be corrupted." },

    { VMERR_NO_STR_CONV,
    "cannot convert value to string",
    "This value cannot be converted to a string." },

    { VMERR_CONV_BUF_OVF,
    "string conversion buffer overflow",
    "An internal buffer overflow occurred converting this value to "
    "a string." },

    { VMERR_BAD_TYPE_ADD,
    "invalid datatypes for addition operator",
    "Invalid datatypes for addition operator.  The values being added "
    "cannot be combined in this manner." },

    { VMERR_NUM_VAL_REQD,
    "numeric value required",
    "Invalid value type - a numeric value is required." },

    { VMERR_INT_VAL_REQD,
    "integer value required",
    "Invalid value type - an integer value is required." },

    { VMERR_NO_LOG_CONV,
    "cannot convert value to logical (true/nil)",
    "This value cannot be converted to a logical (true/nil) value." },

    { VMERR_BAD_TYPE_SUB,
    "invalid datatypes for subtraction operator",
    "Invalid datatypes for subtraction operator.  The values used "
    "cannot be combined in this manner." },

    { VMERR_DIVIDE_BY_ZERO,
    "division by zero",
    "Arithmetic error - Division by zero." },

    { VMERR_INVALID_COMPARISON,
    "invalid comparison",
    "Invalid comparison - these values cannot be compared to one another." },

    { VMERR_OBJ_VAL_REQD,
    "object value required",
    "An object value is required." },

    { VMERR_PROPPTR_VAL_REQD,
    "property pointer required",
    "A property pointer value is required." },

    { VMERR_LOG_VAL_REQD,
    "logical value required",
    "A logical (true/nil) value is required." },

    { VMERR_FUNCPTR_VAL_REQD,
    "function pointer required",
    "A function pointer value is required." },

    { VMERR_CANNOT_INDEX_TYPE,
    "invalid type for indexing - value cannot be index",
    "This type of value cannot be indexed." },

    { VMERR_INDEX_OUT_OF_RANGE,
    "index out of range",
    "The index value is out of range for the value being indexed (it is "
    "too low or too high)." },

    { VMERR_BAD_METACLASS_INDEX,
    "invalid intrinsic class index",
    "The intrinsic class index is out of range.  This probably indicates "
    "that the image file is corrupted." },

    { VMERR_BAD_DYNAMIC_NEW,
    "invalid dynamic object creation (intrinsic class does not support NEW)",
    "This type of object cannot be dynamically created, because the "
    "intrinsic class does not support dynamic creation." },

    { VMERR_OBJ_VAL_REQD_SC,
    "object value required for base class",
    "An object value must be specified for the base class of a dynamic "
    "object creation operation.  The superclass value is of a non-object "
    "type." },

    { VMERR_STRING_VAL_REQD,
    "string value required",
    "A string value is required." },

    { VMERR_LIST_VAL_REQD,
    "list value required",
    "A list value is required." },

    { VMERR_DICT_NO_CONST,
    "list or string reference found in dictionary (entry \"%s\") - this "
        "dictionary cannot be saved in the image file",
    "A dictionary entry (for the string \"%s\") referred to a string or "
    "list value for its associated value data.  This dictionary cannot "
    "be stored in the image file, so the image file cannot be created.  "
    "Check dictionary word additions and ensure that only objects are "
    "added to the dictionary." },

    { VMERR_INVAL_OBJ_TYPE,
    "invalid object type - cannot convert to required object type",
    "An object is not of the correct type.  The object specified "
    "cannot be converted to the required object type." },

    { VMERR_NUM_OVERFLOW,
    "numeric overflow",
    "A numeric calculation overflowed the limits of the datatype." },

    { VMERR_BAD_TYPE_MUL, 
    "invalid datatypes for multiplication operator",
    "Invalid datatypes for multiplication operator.  The values being added "
    "cannot be combined in this manner." },

    { VMERR_BAD_TYPE_DIV, 
    "invalid datatypes for division operator",
    "Invalid datatypes for division operator.  The values being added "
    "cannot be combined in this manner." },

    { VMERR_BAD_TYPE_NEG,
    "invalid datatype for arithmetic negation operator",
    "Invalid datatype for arithmetic negation operator.  The value "
    "cannot be negated." },

    { VMERR_OUT_OF_RANGE,
    "value is out of range",
    "A value that was outside of the legal range of inputs was specified "
    "for a calculation." },
        
    { VMERR_WRONG_NUM_OF_ARGS,
    "wrong number of arguments",
    "The wrong number of arguments was passed to a function or method "
    "in the invocation of the function or method." },

    { VMERR_WRONG_NUM_OF_ARGS_CALLING,
    "argument mismatch calling %s - function definition is incorrect",
    "The number of arguments doesn't match the number expected "
    "calling %s.  Check the function or method and correct the "
    "number of parameters that it is declared to receive." },

    { VMERR_CANNOT_CREATE_INST,
    "cannot create instance of object - object is not a class",
    "An instance of this object cannot be created, because this "
    "object is not a class." },

    { VMERR_ILLEGAL_NEW,
    "cannot create instance - class does not allow dynamic construction",
    "An instance of this class cannot be created, because this class "
    "does not allow dynamic construction." },

    { VMERR_INVALID_OPCODE,
    "invalid opcode - possible image file corruption",
    "Invalid instruction opcode - the image file might be corrupted." },

    { VMERR_UNHANDLED_EXC,
    "unhandled exception",
    "An exception was thrown but was not caught by the program.  "
    "The interpreter is terminating execution of the program." },

    { VMERR_STACK_OVERFLOW,
    "stack overflow",
    "Stack overflow.  This indicates that function or method calls "
    "were nested too deeply; this might have occurred because of "
    "unterminated recursion, which can happen when a function or method "
    "calls itself (either directly or indirectly)." },

    { VMERR_BAD_TYPE_BIF,
    "invalid type for intrinsic function argument",
    "An invalid datatype was provided for an intrinsic function argument." },

    { VMERR_SAY_IS_NOT_DEFINED,
    "default output function is not defined",
    "The default output function is not defined.  Implicit string "
    "display is not allowed until a default output function "
    "is specified." },

    { VMERR_BAD_VAL_BIF,
    "invalid value for intrinsic function argument",
    "An invalid value was specified for an intrinsic function argument.  "
    "The value is out of range or is not an allowed value." },

    { VMERR_BREAKPOINT,
    "breakpoint encountered",
    "A breakpoint instruction was encountered, and no debugger is "
    "active.  The compiler might have inserted this breakpoint to "
    "indicate an invalid or unreachable location in the code, so "
    "executing this instruction probably indicates an error in the "
    "program." },

    { VMERR_CALLEXT_NOT_IMPL,
    "external function calls are not implemented in this version",
    "This version of the interpreter does not implement external function "
    "calls.  This program requires an interpreter that provides external "
    "function call capabilities, so this program is not compatible "
    "with this interpreter." },

    { VMERR_INVALID_OPCODE_MOD,
    "invalid opcode modifier - possible image file corruption",
    "Invalid instruction opcode modifier - the image file might "
    "be corrupted." },

    { VMERR_NO_CHARMAP_FILE,
    "No mapping file available for local character set \"%.32s\"",
    "[Warning: no mapping file is available for the local character set "
    "\"%.32s\".  The system will use a default ASCII character set mapping "
    "instead, so accented characters will be displayed without their "
    "accents.]" },

    { VMERR_DBG_QUIT,
    "'quit' signal", "'quit' signal" },

    { VMERR_DBG_ABORT,
    "'abort' signal", "'abort' signal" },

    { VMERR_DBG_RESTART,
    "'restart' signal", "'restart' signal" },

    { VMERR_BAD_FILE_HANDLE,
    "invalid file handle",
    "An invalid value was specified for the file handle in a call to "
    "a file manipulation function." },

    { VMERR_BAD_FRAME,
    "invalid frame in debugger local/parameter evaluation",
    "An invalid stack frame was specified in a debugger local/parameter "
    "evaluation.  This probably indicates an internal problem in the "
    "debugger." },

    { VMERR_BAD_SPEC_EVAL,
    "invalid speculative expression"
    "This expression cannot be executed speculatively.  (This does not "
    "indicate a problem; it's merely an internal condition in the "
    "debugger.)" },

    { VMERR_INVAL_DBG_EXPR,
    "invalid debugger expression",
    "This expression cannot be evaluated in the debugger." },

    { VMERR_NO_IMAGE_DBG_INFO,
    "image file has no debugging information - recompile for debugging",
    "The image file has no debugging information.  You must recompile the "
    "source code for this program with debugging information in order to "
    "run the program under the debugger." },

    { VMERR_BIGNUM_NO_REGS,
    "out of temporary floating point registers (calculation too complex)",
    "The interpreter is out of temporary floating point registers.  "
    "This probably indicates that an excessively complex calculation "
    "has been attempted." }
};

/* message count */
size_t vm_message_count = sizeof(vm_messages)/sizeof(vm_messages[0]);

