#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/VMINIT.CPP,v 1.3 1999/07/11 00:46:58 MJRoberts Exp $";
#endif

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vminit.cpp - initialize and terminate VM processing
Function
  Provides functions to create and destroy global objects for a VM
  session.
Notes
  
Modified
  04/06/99 MJRoberts  - Creation
*/

#include <string.h>

#include "t3std.h"
#include "charmap.h"
#include "vminit.h"
#include "vmerr.h"
#include "vmfile.h"
#include "vmimage.h"
#include "vmpool.h"
#include "vmobj.h"
#include "vmstack.h"
#include "vmundo.h"
#include "vmmeta.h"
#include "vmbif.h"
#include "vmrun.h"
#include "vmpredef.h"
#include "vmmcreg.h"
#include "vmbiftad.h"
#include "resload.h"
#include "vmhost.h"
#include "vmconsol.h"
#include "vmbignum.h"


/* ------------------------------------------------------------------------ */
/*
 *   Perform base initialization.  This is an internal routine called only
 *   by higher-level initialization routines; we perform all of the
 *   generic, configuration-independent initialization.
 */
void vm_init_base(vm_globals **vmg, class CVmHostIfc *hostifc,
                  const char *charset)
{
    vm_globals *vmg__;
    char disp_mapname[32];
    char file_mapname[32];
    CResLoader *map_loader;
    int file_map_err;
    int disp_map_err;
    
    /* 
     *   Allocate globals according to build-time configuration, then
     *   assign the global pointer to a local named vmg__.  This will
     *   ensure that the globals are accessible for all of the different
     *   build-time configurations.  
     */
    vmg__ = *vmg = vmglob_alloc();

    /* 
     *   initialize error stack if it hasn't already been initialized;
     *   note if it's already set up so that we don't tear it down 
     */
    G_err_pre_inited = (G_err != 0);
    if (!G_err_pre_inited)
        err_init(1024);

    /* remember the host interface */
    G_host_ifc = hostifc;

    /* we don't have a resource loader for program resources yet */
    G_res_loader = 0;

    /* create the object table */
    G_obj_table = new CVmObjTable();

    /* create the memory manager */
    G_varheap = new CVmVarHeapHybrid();
//    G_varheap = new CVmVarHeapMalloc();
    G_mem = new CVmMemory(vmg_ G_varheap);

    /* create a stack */
    G_stk = new CVmStack(1024);

    /* create the undo manager */
    G_undo = new CVmUndo(100, 10);

    /* create the metafile and function set tables */
    G_meta_table = new CVmMetaTable(5);
    G_bif_table = new CVmBifTable(5);

    /* initialize the metaclass registration tables */
    vm_register_metaclasses();

    /* create the byte-code interpreter */
    G_interpreter = new CVmRun();

    /* presume we won't create debugger information */
    G_debugger = 0;
    G_srcf_table = 0;

    /* initialize the debugger if present */
    vm_init_debugger(vmg0_);

    /* create the pre-defined object mapper */
    G_predef = new CVmPredef;

    /* presume we're in normal execution mode (not preinit) */
    G_preinit_mode = FALSE;

    /* allocate the TADS intrinsic function set's globals */
    G_bif_tads_globals = new CVmBifTADSGlobals();

    /* allocate the BigNumber register cache */
    G_bignum_cache = new CVmBigNumCache(32);

    /* no image loader yet */
    G_image_loader = 0;

    /* get the character map loader from the host interface */
    map_loader = hostifc->get_cmap_res_loader();

    /*
     *   If the caller explicitly specified a character set, use it.
     *   Otherwise, ask the OS layer for the default character set we
     *   should use. 
     */
    if (charset == 0)
    {
        /* the user did not specify a mapping - ask the OS for the default */
        os_get_charmap(disp_mapname, OS_CHARMAP_DISPLAY);

        /* use the name we got from the OS */
        charset = disp_mapname;
    }

    /* create the display character maps */
    G_cmap_from_ui = CCharmapToUni::load(map_loader, charset);
    G_cmap_to_ui = CCharmapToLocal::load(map_loader, charset);

    /* create the filename character maps */
    os_get_charmap(file_mapname, OS_CHARMAP_FILENAME);
    G_cmap_from_fname = CCharmapToUni::load(map_loader, file_mapname);
    G_cmap_to_fname = CCharmapToLocal::load(map_loader, file_mapname);

    /* make a note of whether we had any problems loading the maps */
    disp_map_err = (G_cmap_from_ui == 0 || G_cmap_to_ui == 0);
    file_map_err = (G_cmap_from_fname == 0 || G_cmap_to_fname == 0);

    /* if we failed to create any of the maps, load defaults */
    if (G_cmap_from_ui == 0)
        G_cmap_from_ui = CCharmapToUni::load(map_loader, "asc7dflt");
    if (G_cmap_to_ui == 0)
        G_cmap_to_ui = CCharmapToLocal::load(map_loader, "asc7dflt");
    if (G_cmap_from_fname == 0)
        G_cmap_from_fname = CCharmapToUni::load(map_loader, "asc7dflt");
    if (G_cmap_to_fname == 0)
        G_cmap_to_fname = CCharmapToLocal::load(map_loader, "asc7dflt");    

    /* create the primary console */
    G_console = new CVmConsole();

    /* 
     *   if we had any trouble opening the display character mapping file,
     *   make a note that we are using a default mapping 
     */
    if (disp_map_err)
    {
        const char *msg;
        char buf[256];

        /* get the message */
        msg = err_get_msg(vm_messages, vm_message_count,
                          VMERR_NO_CHARMAP_FILE, TRUE);

        /* format it */
        sprintf(buf, msg, charset);

        /* display it */
        G_console->format_text(vmg_ buf);
        G_console->write_blank_line(vmg0_);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Terminate the VM 
 */
void vm_terminate(vm_globals *vmg__)
{
    /* tell the debugger to shut down, if necessary */
    vm_terminate_debug_shutdown(vmg0_);
    
    /* drop all undo information */
    G_undo->drop_undo(vmg0_);

    /* delete the output formatter */
    delete G_console;
    
    /* delete the character mappers */
    delete G_cmap_from_fname;
    delete G_cmap_to_fname;
    delete G_cmap_from_ui;
    delete G_cmap_to_ui;
    
    /* delete the BigNumber register cache */
    delete G_bignum_cache;

    /* delete the TADS intrinsic function set's globals */
    delete G_bif_tads_globals;

    /* delete the predefined object table */
    delete G_predef;

    /* delete the interpreter */
    delete G_interpreter;

    /* delete debugger objects */
    vm_terminate_debug_delete(vmg0_);

    /* delete the constant pools */
    delete G_code_pool;
    delete G_const_pool;

    /* delete the dependency tables */
    delete G_meta_table;
    delete G_bif_table;

    /* delete the undo manager */
    delete G_undo;

    /* delete the stack */
    delete G_stk;

    /* delete the object table */
    G_obj_table->delete_obj_table(vmg0_);

    /* delete the memory manager and heap manager */
    delete G_mem;
    delete G_varheap;

    /* delete the error context */
    if (!G_err_pre_inited)
        err_terminate();

    /* delete the globals */
    vmglob_delete(vmg__);
}

