#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 2000 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmintcls.cpp - T3 metaclass - intrinsic class
Function
  
Notes
  
Modified
  03/08/00 MJRoberts  - Creation
*/

#include <stdlib.h>
#include <assert.h>

#include "vmtype.h"
#include "vmobj.h"
#include "vmglob.h"
#include "vmintcls.h"
#include "vmmeta.h"
#include "vmfile.h"

/* ------------------------------------------------------------------------ */
/*
 *   statics 
 */

/* metaclass registration object */
static CVmMetaclassClass metaclass_reg_obj;
CVmMetaclass *CVmObjClass::metaclass_reg_ = &metaclass_reg_obj;


/* ------------------------------------------------------------------------ */
/* 
 *   create dynamically using stack arguments 
 */
vm_obj_id_t CVmObjClass::create_from_stack(VMG_ const uchar **pc_ptr,
                                           uint argc)
{
    /* it is illegal to create this type of object dynamically */
    err_throw(VMERR_ILLEGAL_NEW);
    return VM_INVALID_OBJ;
}

/* 
 *   create with no initial contents 
 */
vm_obj_id_t CVmObjClass::create(VMG_ int in_root_set)
{
    vm_obj_id_t id = vm_new_id(vmg_ in_root_set);
    new (vmg_ id) CVmObjClass();
    return id;
}

/*
 *   create with a given dependency index 
 */
vm_obj_id_t CVmObjClass::create_dyn(VMG_ uint meta_idx)
{
    vm_obj_id_t id = vm_new_id(vmg_ FALSE);
    new (vmg_ id) CVmObjClass(vmg_ FALSE, meta_idx, id);
    return id;
}

/*
 *   create with a given dependency index 
 */
CVmObjClass::CVmObjClass(VMG_ int in_root_set, uint meta_idx,
                         vm_obj_id_t self)
{
    /* calls of this form can't come from the image file */
    assert(!in_root_set);

    /* allocate the extension */
    ext_ = (char *)G_mem->get_var_heap()->alloc_mem(8, this);

    /* set up the extension - write the length and dependency index */
    oswp2(ext_, 8);
    oswp2(ext_ + 2, meta_idx);
    oswp4(ext_ + 4, VM_INVALID_OBJ);

    /* register myself with the metaclass table */
    register_meta(vmg_ self);
}

/* 
 *   notify of deletion 
 */
void CVmObjClass::notify_delete(VMG_ int in_root_set)
{
    /* free our extension */
    if (ext_ != 0 && !in_root_set)
        G_mem->get_var_heap()->free_mem(ext_);
}

/* set a property */
void CVmObjClass::set_prop(VMG_ CVmUndo *undo,
                           vm_obj_id_t self, vm_prop_id_t prop,
                           const vm_val_t *val)
{
    vm_obj_id_t mod_obj;
    
    /* if I have a modifier object, pass the setprop to the modifier */
    if ((mod_obj = get_mod_obj()) != VM_INVALID_OBJ)
    {
        /* set the property in the modifier object */
        vm_objp(vmg_ mod_obj)->set_prop(vmg_ undo, mod_obj, prop, val);
    }
    else
    {
        /* if we don't have a modifier, we can't set the property */
        err_throw(VMERR_INVALID_SETPROP);
    }
}

/* 
 *   get a property 
 */
int CVmObjClass::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *val,
                          vm_obj_id_t self, vm_obj_id_t *source_obj,
                          uint *argc)
{
    uint meta_idx;
    vm_obj_id_t mod_obj;

    /* 
     *   pass the property request as a static property of the metaclass
     *   with which we're associated 
     */
    meta_idx = osrp2(ext_ + 2);
    if (meta_idx < G_meta_table->get_count())
    {
        /* call the static property in our metaclass */
        if (G_meta_table->get_entry(meta_idx)->meta_
            ->call_stat_prop(vmg_ val, 0, argc, prop))
        {
            /* the metaclass object handled it */
            return TRUE;
        }
    }

    /* try handling it through the modifier object, if we have one */
    if ((mod_obj = get_mod_obj()) != VM_INVALID_OBJ)
    {
        /* let the modifier have a crack at it */
        if (vm_objp(vmg_ mod_obj)->get_prop(vmg_ prop, val, mod_obj,
                                            source_obj, argc))
            return TRUE;
    }

    /* inherit default handling */
    return CVmObject::get_prop(vmg_ prop, val, self, source_obj, argc);
}

/* 
 *   save to a file 
 */
void CVmObjClass::save_to_file(VMG_ class CVmFile *fp)
{
    size_t len;
    
    /* write our data */
    len = osrp2(ext_);
    fp->write_bytes(ext_, len);
}

/* 
 *   restore from a file 
 */
void CVmObjClass::restore_from_file(VMG_ vm_obj_id_t self,
                                    class CVmFile *fp, CVmObject *obj)
{
    size_t len;

    /* read the length */
    len = fp->read_uint2();

    /* free any existing extension */
    if (ext_ != 0)
    {
        G_mem->get_var_heap()->free_mem(ext_);
        ext_ = 0;
    }

    /* allocate the space */
    ext_ = (char *)G_mem->get_var_heap()->alloc_mem(len, this);

    /* store our length */
    vmb_put_len(ext_, len);

    /* read the contents */
    fp->read_bytes(ext_ + VMB_LEN, len);

    /* register myself with the metaclass table */
    register_meta(vmg_ self);
}

/* 
 *   check for equality 
 */
int CVmObjClass::equals(VMG_ vm_obj_id_t self, const vm_val_t *val) const
{
    /* we equal only ourself */
    return (val->typ == VM_OBJ && val->val.obj == self);
}

/* load from an image file */
void CVmObjClass::load_from_image(VMG_ vm_obj_id_t self,
                                  const char *ptr, size_t siz)
{
    /* save a pointer to the image file data as our extension */
    ext_ = (char *)ptr;

    /* make sure the length is valid */
    if (siz < 8)
        err_throw(VMERR_INVAL_METACLASS_DATA);

    /* register myself */
    register_meta(vmg_ self);
}

/*
 *   Register myself with the metaclass dependency table - this
 *   establishes the association between the metaclass in the depencency
 *   table and this instance, so that the metaclass can figure out what
 *   instance represents it.  
 */
void CVmObjClass::register_meta(VMG_ vm_obj_id_t self)
{
    uint meta_idx;

    /* determine which metaclass we represent */
    meta_idx = osrp2(ext_ + 2);

    /* 
     *   if this is a valid metaclass index, store a reference to myself
     *   in the metaclass table for this index - this will let the
     *   metaclass that we represent find us when asked for its class
     *   object 
     */
    if (meta_idx < G_meta_table->get_count())
        G_meta_table->get_entry(meta_idx)->class_obj_ = self;
}

/*
 *   Get the number of superclasses 
 */
int CVmObjClass::get_superclass_count(VMG_ vm_obj_id_t) const
{
    uint meta_idx;

    /* determine which metaclass we represent */
    meta_idx = osrp2(ext_ + 2);

    /* 
     *   if this is a valid metaclass index, ask the metaclass object to
     *   tell us the superclass count 
     */
    if (meta_idx < G_meta_table->get_count())
        return G_meta_table->get_entry(meta_idx)->meta_
            ->get_supermeta_count(vmg0_);
    else
        return 0;
}

/*
 *   Get a superclass 
 */
vm_obj_id_t CVmObjClass::get_superclass(VMG_ vm_obj_id_t,
                                        int sc_idx) const
{
    uint meta_idx;

    /* determine which metaclass we represent */
    meta_idx = osrp2(ext_ + 2);

    /* 
     *   if this is a valid metaclass index, ask the metaclass object to
     *   retrieve the superclass
     */
    if (meta_idx < G_meta_table->get_count())
        return G_meta_table->get_entry(meta_idx)->meta_
            ->get_supermeta(vmg_ sc_idx);
    else
        return VM_INVALID_OBJ;
}

/*
 *   Determine if I'm an instance of the given object 
 */
int CVmObjClass::is_instance_of(VMG_ vm_obj_id_t obj)
{
    uint meta_idx;

    /* determine which metaclass we represent */
    meta_idx = osrp2(ext_ + 2);

    /* if this is a valid metaclass index, ask the metaclass object */
    if (meta_idx < G_meta_table->get_count())
    {
        /* ask the metaclass if it's an instance of the given object */
        return G_meta_table->get_entry(meta_idx)->meta_
            ->is_meta_instance_of(vmg_ obj);
    }
    else
    {
        /* not a valid metaclass - we can't make sense of this */
        return FALSE;
    }
}

