#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/VMRUN.CPP,v 1.4 1999/07/11 00:46:58 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmrun.cpp - VM Execution
Function
  
Notes
  
Modified
  11/12/98 MJRoberts  - Creation
*/

#include "t3std.h"
#include "os.h"
#include "vmrun.h"
#include "vmdbg.h"
#include "vmop.h"
#include "vmstack.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmobj.h"
#include "vmlst.h"
#include "vmstr.h"
#include "vmtobj.h"
#include "vmfunc.h"
#include "vmmeta.h"
#include "vmbif.h"
#include "vmpredef.h"
#include "vmfile.h"
#include "vmsave.h"


/* ------------------------------------------------------------------------ */
/*
 *   Initialize 
 */
CVmRun::CVmRun()
{
    /* start out with nothing in R0 */
    r0_.set_nil();

    /* there's no frame yet */
    frame_ptr_ = 0;
    reset_ptr_ = 0;

    /* there's no entry pointer yet */
    entry_ptr_ = 0;

    /* clear R0 */
    r0_.set_nil();

    /* function header size is not yet known */
    funchdr_size_ = 0;

    /* the 'say' function hasn't been initialized yet */
    say_func_ = 0;
    say_func_inited_ = FALSE;

    /* no default 'say' method */
    say_method_ = VM_INVALID_PROP;

    /* the program counter isn't stored yet */
    pc_ptr_ = 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Terminate 
 */
CVmRun::~CVmRun()
{
}

/* ------------------------------------------------------------------------ */
/*
 *   Set the function header size 
 */
void CVmRun::set_funchdr_size(size_t siz)
{
    /* remember the new size */
    funchdr_size_ = siz;

    /* 
     *   ensure that the size is at least as large as our required
     *   function header block - if it's not, this version of the VM can't
     *   run this image file 
     */
    if (siz < VMFUNC_HDR_MIN_SIZE)
        err_throw(VMERR_IMAGE_INCOMPAT_HDR_FMT);
}

/* ------------------------------------------------------------------------ */
/*
 *   Add two values and push the result 
 */
void CVmRun::add_and_push(VMG_ vm_val_t *val1, vm_val_t *val2)
{
    /* compute the sum, leaving the result in val1 */
    compute_sum(vmg_ val1, val2);

    /* push the result */
    pushval(vmg_ val1);
}

/*
 *   Add two values, leaving the result in *val1 
 */
void CVmRun::compute_sum(VMG_ vm_val_t *val1, vm_val_t *val2)
{    
    /* the meaning of "add" depends on the type of the first operand */
    switch(val1->typ)
    {
    case VM_SSTRING:
        /* 
         *   string constant - add the second value to the string, using
         *   the static string add method 
         */
        CVmObjString::add_to_str(vmg_ val1, VM_INVALID_OBJ,
                                 get_const_ptr(vmg_ val1->val.ofs), val2);
        break;

    case VM_LIST:
        /* 
         *   list constant - add the second value to the list, using the
         *   static list add method 
         */
        CVmObjList::add_to_list(vmg_ val1, VM_INVALID_OBJ,
                                get_const_ptr(vmg_ val1->val.ofs), val2);
        break;

    case VM_OBJ:
        /*
         *   object - add the second value to the object, using the
         *   object's virtual metaclass add method 
         */
        vm_objp(vmg_ val1->val.obj)->add_val(vmg_ val1, val1->val.obj, val2);
        break;

    case VM_INT:
        /* make sure the other value is a number as well */
        if (!val2->is_numeric())
            err_throw(VMERR_NUM_VAL_REQD);

        /* compute the sum */
        val1->val.intval += val2->num_to_int();
        break;

    default:
        /* other types cannot be added */
        err_throw(VMERR_BAD_TYPE_ADD);
        break;
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Compute the difference of two values and push the result 
 */
void CVmRun::sub_and_push(VMG_ vm_val_t *val1, vm_val_t *val2)
{
    /* compute the difference, leaving the value in val1 */
    compute_diff(vmg_ val1, val2);
    
    /* push the result */
    pushval(vmg_ val1);
}

/*
 *   Compute the difference of two values, leaving the result in *val1 
 */
void CVmRun::compute_diff(VMG_ vm_val_t *val1, vm_val_t *val2)
{
    /* the meaning of "subtract" depends on the type of the first operand */
    switch(val1->typ)
    {
    case VM_LIST:
        /* 
         *   list constant - remove the second value from the list, using
         *   the static list subtraction method 
         */
        CVmObjList::sub_from_list(vmg_ val1, val1,
                                  get_const_ptr(vmg_ val1->val.ofs), val2);
        break;

    case VM_OBJ:
        /* object - use the object's virtual subtraction method */
        vm_objp(vmg_ val1->val.obj)->sub_val(vmg_ val1, val1->val.obj, val2);
        break;

    case VM_INT:
        /* make sure the other value is a number as well */
        if (!val2->is_numeric())
            err_throw(VMERR_NUM_VAL_REQD);

        /* compute the difference */
        val1->val.intval -= val2->num_to_int();
        break;

    default:
        /* other types cannot be subtracted */
        err_throw(VMERR_BAD_TYPE_SUB);
    }

}

/* ------------------------------------------------------------------------ */
/*
 *   Multiply two values and push the result 
 */
void CVmRun::mul_and_push(VMG_ vm_val_t *val1, vm_val_t *val2)
{
    switch(val1->typ)
    {
    case VM_OBJ:
        /* use the object's virtual multiplication method */
        vm_objp(vmg_ val1->val.obj)->mul_val(vmg_ val1, val1->val.obj, val2);
        break;

    case VM_INT:
        /* make sure the other value is a number as well */
        if (!val2->is_numeric())
            err_throw(VMERR_NUM_VAL_REQD);

        /* compute the product */
        val1->val.intval *= val2->num_to_int();
        break;

    default:
        /* other types are invalid */
        err_throw(VMERR_BAD_TYPE_MUL);
    }

    /* push the result */
    pushval(vmg_ val1);
}

/* ------------------------------------------------------------------------ */
/*
 *   Divide two values and push the result 
 */
void CVmRun::div_and_push(VMG_ vm_val_t *val1, vm_val_t *val2)
{
    switch(val1->typ)
    {
    case VM_OBJ:
        /* use the object's virtual division method */
        vm_objp(vmg_ val1->val.obj)->div_val(vmg_ val1, val1->val.obj, val2);
        break;

    case VM_INT:
        /* make sure the other value is a number as well */
        if (!val2->is_numeric())
            err_throw(VMERR_NUM_VAL_REQD);

        /* check for divide by zero */
        if (val2->num_to_int() == 0)
            err_throw(VMERR_DIVIDE_BY_ZERO);

        /* compute the product */
        val1->val.intval = os_divide_long(val1->val.intval,
                                          val2->num_to_int());
        break;

    default:
        /* other types are invalid */
        err_throw(VMERR_BAD_TYPE_DIV);
    }

    /* push the result */
    pushval(vmg_ val1);
}

/* ------------------------------------------------------------------------ */
/*
 *   XOR two values and push the result.  The values can be numeric or
 *   logical.  If either value is logical, the result will be logical;
 *   otherwise, the result will be a bitwise XOR of the integers.  
 */
void CVmRun::xor_and_push(VMG_ vm_val_t *val1, vm_val_t *val2)
{
    /* if either value is logical, compute the logical XOR */
    if (val1->is_logical() && val2->is_logical())
    {
        /* both values are logical - compute the logical XOR */
        val1->set_logical(val1->get_logical() ^ val2->get_logical());
    }
    else if (val1->is_logical() || val2->is_logical())
    {
        /* 
         *   one value is logical, but not both - convert the other value
         *   from a number to a logical and compute the result as a
         *   logical value 
         */
        if (!val1->is_logical())
            val1->num_to_logical();
        else if (!val2->is_logical())
            val2->num_to_logical();

        /* compute the logical xor */
        val1->set_logical(val1->get_logical() ^ val2->get_logical());
    }
    else if (val1->typ == VM_INT && val2->typ == VM_INT)
    {
        /* compute and store the bitwise XOR */
        val1->val.intval = val1->val.intval ^ val2->val.intval;
    }
    else
    {
        /* no logical conversion */
        err_throw(VMERR_NO_LOG_CONV);
    }

    /* push the result */
    pushval(vmg_ val1);
}


/* ------------------------------------------------------------------------ */
/*
 *   Index a value and push the result.
 */
void CVmRun::index_and_push(VMG_ const vm_val_t *container_val,
                            const vm_val_t *index_val)
{
    vm_val_t result;
    
    /* check the type of the value we're indexing */
    switch(container_val->typ)
    {
    case VM_LIST:
        /* list constant - use the static list indexing method */
        CVmObjList::index_list(vmg_ &result,
                               get_const_ptr(vmg_ container_val->val.ofs),
                               index_val);
        break;

    case VM_OBJ:
        /* object - use the object's virtual indexing method */
        vm_objp(vmg_ container_val->val.obj)
            ->index_val(vmg_ &result, container_val->val.obj, index_val);
        break;

    default:
        /* other values cannot be indexed */
        err_throw(VMERR_CANNOT_INDEX_TYPE);
    }

    /* push the result, which is now in result */
    pushval(vmg_ &result);
}

/* ------------------------------------------------------------------------ */
/*
 *   Set an indexed value and push the modified container.
 */
void CVmRun::set_index_and_push(VMG_ vm_val_t *container_val,
                                const vm_val_t *index_val,
                                const vm_val_t *new_val)
{
    switch(container_val->typ)
    {
    case VM_LIST:
        /* list constant - use the static list set-index method */
        CVmObjList::set_index_list(vmg_ container_val,
                                   get_const_ptr(vmg_ container_val->val.ofs),
                                   index_val, new_val);
        break;

    case VM_OBJ:
        /* object - use the object's virtual set-index method */
        vm_objp(vmg_ container_val->val.obj)
            ->set_index_val(vmg_ container_val,
                            container_val->val.obj, index_val, new_val);
        break;

    default:
        /* other values cannot be indexed */
        err_throw(VMERR_CANNOT_INDEX_TYPE);
    }

    /* push the new container value */
    pushval(vmg_ container_val);
}

/* ------------------------------------------------------------------------ */
/*
 *   Create a new object and store it in R0
 */
void CVmRun::new_and_store_r0(VMG_ const uchar **pc_ptr,
                              uint metaclass_idx, uint argc)
{
    vm_obj_id_t obj;
    
    /* create the object */
    obj = G_meta_table->create_from_stack(vmg_ pc_ptr, metaclass_idx, argc);

    /* if we got a valid object, store a reference to it in R0 */
    if (obj != VM_INVALID_OBJ)
        r0_.set_obj(obj);
    else
        r0_.set_nil();
}

/* ------------------------------------------------------------------------ */
/*
 *   Resume execution from a saved state 
 */
void CVmRun::run_restore(VMG_ const char *saved_state)
{
    osfildef *save_fp;
    CVmFile *save_file;
    const uchar *pc;
    int err;

    /* open the saved state file */
    save_fp = osfoprb(saved_state, OSFTT3SAV);
    if (save_fp == 0)
        err_throw(VMERR_FILE_NOT_FOUND);

    /* set up the file reader */
    save_file = new CVmFile();
    save_file->set_file(save_fp, 0);

    err_try
    {
        /* 
         *   before restoring, set up the PC pointer to refer to our PC
         *   register - this will make sure that the restoration fills in
         *   our PC register with the address of the code where execution
         *   is to resume 
         */
        pc_ptr_ = &pc;

        /* restore the state */
        err = CVmSaveFile::restore(vmg_ save_file);

        /* clear out the PC pointer now that we're finished restoring */
        pc_ptr_ = 0;
    }
    err_finally
    {
        /* close the file and delete the file object */
        delete save_file;
    }
    err_end;

    /* if we failed to restore, throw the error */
    if (err != 0)
        err_throw(err);

    /* 
     *   Store 1 in r0 - we're restoring directly from the command line,
     *   so we don't have a value passed into the restore() function.
     *   This is the value that it will look like save() is returning.  We
     *   don't want to return nil, because the caller would then think it
     *   just saved rather than just restored.  
     */
    G_interpreter->get_r0()->set_int(1);

    /* begin program execution from the resumed state */
    run(vmg_ pc);
}

/* ------------------------------------------------------------------------ */
/*
 *   Execute byte code 
 */
void CVmRun::run(VMG_ const uchar *p)
{
    vm_val_t *valp;
    vm_val_t val;
    vm_val_t val2;
    vm_val_t val3;
    int done;
    vm_obj_id_t obj;
    vm_prop_id_t prop;
    uint argc;
    uint idx;
    uint set_idx;
    pool_ofs_t ofs;
    uint cnt;
    vm_obj_id_t unhandled_exc;
    const uchar **old_pc_ptr;
    int level;

    /* save the enclosing program counter pointer, and remember the new one */
    old_pc_ptr = pc_ptr_;
    pc_ptr_ = &p;

    /* we're not done yet */
    done = FALSE;

    /* no unhandled exception yet */
    unhandled_exc = VM_INVALID_OBJ;

    /*
     *   Come back here whenever we catch a run-time exception and find a
     *   byte-code error handler to process it in the stack.  We'll
     *   re-enter our exception handler and resume byte-code execution at
     *   the handler.  
     */
resume_execution:

    /*
     *   Execute all code within an exception frame.  If any routine we
     *   call throws an exception, we'll catch the exception and process
     *   it as a run-time error.  
     */
    err_try
    {
        /* execute code until something makes us stop */
        for (;;)
        {
            /* if we're single-stepping, break into the debugger */
            VM_IF_DEBUGGER(if (G_debugger->is_single_step())
                G_debugger->step(vmg_ &p, entry_ptr_, FALSE));

            /* execute the current instruction */
        exec_instruction:
            switch(*p++)
            {
            case OPC_PUSH_0:
                /* push the constant value 0 */
                push_int(vmg_ 0);
                break;

            case OPC_PUSH_1:
                /* push the constant value 1 */
                push_int(vmg_ 1);
                break;
                
            case OPC_PUSHINT8:
                /* push an SBYTE operand value */
                push_int(vmg_ get_op_int8(&p));
                break;

            case OPC_PUSHINT:
                /* push a UINT4 operand value */
                push_int(vmg_ get_op_int32(&p));
                break;

            case OPC_PUSHENUM:
                /* push a UINT4 operand value */
                push_enum(vmg_ get_op_uint32(&p));
                break;

            case OPC_PUSHSTR:
                /* push UINT4 offset operand as a string */
                val.set_sstring(get_op_uint32(&p));
                pushval(vmg_ &val);
                break;

            case OPC_PUSHSTRI:
                /* inline string - get the length prefix */
                cnt = get_op_uint16(&p);

                /* create the new string from the inline data */
                val.set_obj(CVmObjString::create(vmg_ FALSE,
                    (const char *)p, cnt));

                /* skip past the string's bytes */
                p += cnt;

                /* push the new string */
                pushval(vmg_ &val);
                break;

            case OPC_PUSHLST:
                /* push UINT4 offset operand as a list */
                val.set_list(get_op_uint32(&p));
                pushval(vmg_ &val);
                break;

            case OPC_PUSHOBJ:
                /* push UINT4 object ID operand */
                val.set_obj(get_op_uint32(&p));
                pushval(vmg_ &val);
                break;

            case OPC_PUSHNIL:
                /* push nil */
                push_nil(vmg0_);
                break;

            case OPC_PUSHTRUE:
                /* push true */
                val.set_true();
                pushval(vmg_ &val);
                break;

            case OPC_PUSHPROPID:
                /* push UINT2 property ID operand */
                val.set_propid(get_op_uint16(&p));
                pushval(vmg_ &val);
                break;

            case OPC_PUSHFNPTR:
                /* push a function pointer operand */
                val.set_fnptr(get_op_uint32(&p));
                pushval(vmg_ &val);
                break;

            case OPC_PUSHPARLST:
                /* get the number of fixed parameters */
                cnt = *p++;

                /* make sure we have enough stack space available */
                if (!G_stk->check_space(get_cur_argc(vmg0_) - cnt))
                    err_throw(VMERR_STACK_OVERFLOW);

                /* push from the last list parameter backwards */
                for (idx = get_cur_argc(vmg0_) ; idx > cnt ; )
                {
                    /* move on to the previous parameter */
                    --idx;
                    
                    /* push this parameter onto the stack */
                    pushval(vmg_ get_param(vmg_ idx));
                }

                /* construct a new list */
                obj = CVmObjList::
                      create_from_stack(vmg_ &p, get_cur_argc(vmg0_) - cnt);

                /* push the new list */
                push_obj(vmg_ obj);
                break;

            case OPC_MAKELSTPAR:
                {
                    const char *lstp;
                    uint i;
                    uint hdr_depth;
                    CVmFuncPtr hdr_ptr;

                    /* pop the value */
                    popval(vmg_ &val);

                    /* pop the argument counter so far */
                    pop_int(vmg_ &val2);

                    /* if it's not a list, just push it again unchanged */
                    if ((lstp = val.get_as_list(vmg0_)) == 0)
                    {
                        /* put it back on the stack */
                        pushval(vmg_ &val);

                        /* increment the argument count and push it */
                        ++val2.val.intval;
                        pushval(vmg_ &val2);

                        /* our work here is done */
                        break;
                    }

                    /* set up a pointer to the current function header */
                    hdr_ptr.set((const uchar *)
                                G_code_pool->get_ptr(entry_ptr_));

                    /* get the depth required for the header */
                    hdr_depth = hdr_ptr.get_stack_depth();

                    /* 
                     *   deduct the amount stack space we've already used
                     *   from the amount noted in the header, because
                     *   that's the amount more that we could need for the
                     *   fixed stuff
                     */
                    hdr_depth -= (G_stk->get_depth_rel(frame_ptr_) - 1);

                    /* get the number of elements in the list */
                    cnt = vmb_get_len(lstp);
                      
                    /* make sure we have enough stack space available */
                    if (!G_stk->check_space(cnt + hdr_depth))
                        err_throw(VMERR_STACK_OVERFLOW);

                    /* push the elements of the list from last to first */
                    for (i = cnt ; i != 0 ; --i)
                    {
                        /* get this element's value */
                        CVmObjList::index_list(vmg_ &val, lstp, i);

                        /* push the value */
                        pushval(vmg_ &val);
                    }

                    /* increment and push the argument count */
                    val2.val.intval += cnt;
                    pushval(vmg_ &val2);
                }
                break;

            case OPC_NEG:
                /* check for an object */
                if (G_stk->get(0)->typ == VM_OBJ)
                {
                    /* pop the object */
                    popval(vmg_ &val);
                    
                    /* call the object's negate method */
                    vm_objp(vmg_ val.val.obj)
                        ->neg_val(vmg_ &val2, val.val.obj);

                    /* push the result */
                    pushval(vmg_ &val2);
                }
                else
                {
                    /* negate number on top of stack */
                    pop_num(vmg_ &val);
                    val.val.intval = -val.val.intval;
                    pushval(vmg_ &val);
                }
                break;

            case OPC_BNOT:
                /* bitwise NOT integer on top of stack */
                pop_int(vmg_ &val);
                val.val.intval = ~val.val.intval;
                pushval(vmg_ &val);
                break;

            case OPC_ADD:
                /* if they're both integers, add them the quick way */
                if (G_stk->get(0)->typ == VM_INT
                    && G_stk->get(1)->typ == VM_INT)
                {
                    /* add the two values */
                    G_stk->get(1)->val.intval += G_stk->get(0)->val.intval;

                    /* discard the second value */
                    G_stk->discard();
                }
                else
                {
                    /* add two values on top of stack */
                    popval_2(vmg_ &val, &val2);
                    add_and_push(vmg_ &val, &val2);
                }
                break;

            case OPC_INC:
                /* 
                 *   Increment the value at top of stack.  We must perform
                 *   the same type conversions as the ADD instruction
                 *   does.  As an optimization, check to see if we have an
                 *   integer on top of the stack, and if so simply
                 *   increment its value without popping and repushing.  
                 */
                if (G_stk->get(0)->typ == VM_INT)
                {
                    /* it's an integer - increment it, and we're done */
                    ++(G_stk->get(0)->val.intval);
                }
                else
                {
                    /* pop the value, add 1, and push the result */
                    popval(vmg_ &val);
                    val2.set_int(1);
                    add_and_push(vmg_ &val, &val2);
                }
                break;

            case OPC_DEC:
                /* 
                 *   Decrement the value at top of stack.  We must perform
                 *   the same type conversions as the SUB instruction
                 *   does.  As an optimization, check to see if we have an
                 *   integer on top of the stack, and if so simply
                 *   decrement its value without popping and repushing.  
                 */
                if (G_stk->get(0)->typ == VM_INT)
                {
                    /* it's an integer - decrement it, and we're done */
                    --(G_stk->get(0)->val.intval);
                }
                else
                {
                    /* decrement top of stack and push the result */
                    popval(vmg_ &val);
                    val2.set_int(1);
                    sub_and_push(vmg_ &val, &val2);
                }
                break;

            case OPC_SUB:
                /* if they're both integers, subtract them the quick way */
                if (G_stk->get(0)->typ == VM_INT
                    && G_stk->get(1)->typ == VM_INT)
                {
                    /* compute the difference */
                    G_stk->get(1)->val.intval -= G_stk->get(0)->val.intval;

                    /* discard the second value */
                    G_stk->discard();
                }
                else
                {
                    /* subtract two values on top of stack */
                    popval_2(vmg_ &val, &val2);
                    sub_and_push(vmg_ &val, &val2);
                }
                break;

            case OPC_MUL:
                /* if they're both integers, this is easy */
                if (G_stk->get(0)->typ == VM_INT
                    && G_stk->get(1)->typ == VM_INT)
                {
                    /* compute the difference */
                    G_stk->get(1)->val.intval *= G_stk->get(0)->val.intval;

                    /* discard the second value */
                    G_stk->discard();
                }
                else
                {
                    /* multiply two values on top of stack */
                    popval_2(vmg_ &val, &val2);
                    mul_and_push(vmg_ &val, &val2);
                }
                break;

            case OPC_DIV:
                /* if they're both integers, divide them the quick way */
                if (G_stk->get(0)->typ == VM_INT
                    && G_stk->get(1)->typ == VM_INT)
                {
                    /* check for division by zero */
                    if (G_stk->get(0)->val.intval == 0)
                        err_throw(VMERR_DIVIDE_BY_ZERO);

                    /* compute the result of the division */
                    G_stk->get(1)->val.intval =
                        os_divide_long(G_stk->get(1)->val.intval,
                                       G_stk->get(0)->val.intval);

                    /* discard the second value */
                    G_stk->discard();
                }
                else
                {
                    /* divide number at (TOS-1) by number at top of stack */
                    popval_2(vmg_ &val, &val2);
                    div_and_push(vmg_ &val, &val2);
                }
                break;

            case OPC_BAND:
                /* bitwise AND two integers on top of stack */
                pop_int_2(vmg_ &val, &val2);
                val.val.intval &= val2.val.intval;
                pushval(vmg_ &val);
                break;

            case OPC_BOR:
                /* bitwise OR two integers on top of stack */
                pop_int_2(vmg_ &val, &val2);
                val.val.intval |= val2.val.intval;
                pushval(vmg_ &val);
                break;

            case OPC_SHL:
                /* 
                 *   bit-shift left integer at (TOS-1) by integer at top
                 *   of stack 
                 */
                pop_int_2(vmg_ &val, &val2);
                val.val.intval <<= val2.val.intval;
                pushval(vmg_ &val);
                break;

            case OPC_SHR:
                /* 
                 *   bit-shift right integer at (TOS-1) by integer at top
                 *   of stack 
                 */
                pop_int_2(vmg_ &val, &val2);
                val.val.intval >>= val2.val.intval;
                pushval(vmg_ &val);
                break;

            case OPC_XOR:
                /* XOR two values at top of stack */
                popval_2(vmg_ &val, &val2);
                xor_and_push(vmg_ &val, &val2);
                break;

            case OPC_MOD:
                /* remainder number at (TOS-1) by number at top of stack */
                pop_num_2(vmg_ &val, &val2);
                if (val2.val.intval == 0)
                    err_throw(VMERR_DIVIDE_BY_ZERO);
                val.val.intval =
                    os_remainder_long(val.val.intval, val2.val.intval);
                pushval(vmg_ &val);
                break;

            case OPC_NOT:
                /* 
                 *   invert the logic value; if the value is a number,
                 *   treat 0 as nil and non-zero as true 
                 */
                G_stk->pop(&val);
                switch(val.typ)
                {
                case VM_NIL:
                    push_bool(vmg_ TRUE);
                    break;

                case VM_TRUE:
                    push_bool(vmg_ FALSE);
                    break;

                case VM_INT:
                    push_bool(vmg_ val.val.intval == 0);
                    break;

                default:
                    err_throw(VMERR_NO_LOG_CONV);
                }
                break;

            case OPC_BOOLIZE:
                /* set to a boolean value */
                switch(G_stk->get(0)->typ)
                {
                case VM_NIL:
                case VM_TRUE:
                    /* it's already a logical value - leave it alone */
                    break;

                case VM_INT:
                    /* pop the old value */
                    G_stk->pop(&val);
                    
                    /* push the new value: 0 -> nil, non-zero -> true */
                    push_bool(vmg_ val.val.intval);
                    break;

                case VM_ENUM:
                    /* pop the old value */
                    G_stk->pop(&val);

                    /* push the new value - an enum is always non-nil */
                    push_bool(vmg_ TRUE);
                    break;

                default:
                    err_throw(VMERR_NO_LOG_CONV);
                }
                break;

            case OPC_EQ:
                /* compare two values at top of stack for equality */
                push_bool(vmg_ pop2_equal(vmg0_));
                break;

            case OPC_NE:
                /* compare two values at top of stack for inequality */
                push_bool(vmg_ !pop2_equal(vmg0_));
                break;

            case OPC_LT:
                /* compare values at top of stack - true if (TOS-1) < TOS */
                push_bool(vmg_ pop2_compare(vmg0_) < 0);
                break;

            case OPC_LE:
                /* compare values at top of stack - true if (TOS-1) <= TOS */
                push_bool(vmg_ pop2_compare(vmg0_) <= 0);
                break;

            case OPC_GT:
                /* compare values at top of stack - true if (TOS-1) > TOS */
                push_bool(vmg_ pop2_compare(vmg0_) > 0);
                break;

            case OPC_GE:
                /* compare values at top of stack - true if (TOS-1) >= TOS */
                push_bool(vmg_ pop2_compare(vmg0_) >= 0);
                break;

            case OPC_VARARGC:
                {
                    uchar opc;

                    /* get the modified opcode */
                    opc = *p++;

                    /* 
                     *   skip the immediate data argument count - this is
                     *   superseded by our dynamic argument counter 
                     */
                    ++p;
                    
                    /* pop the argument counter */
                    pop_int(vmg_ &val);
                    argc = val.val.intval;

                        /* execute the appropriate next opcode */
                    switch(opc)
                    {
                    case OPC_CALL:
                        goto do_opc_call;

                    case OPC_PTRCALL:
                        goto do_opc_ptrcall;

                    case OPC_CALLPROP:
                        goto do_opc_callprop;

                    case OPC_PTRCALLPROP:
                        goto do_opc_ptrcallprop;

                    case OPC_CALLPROPSELF:
                        goto do_opc_callpropself;

                    case OPC_PTRCALLPROPSELF:
                        goto do_opc_ptrcallpropself;

                    case OPC_OBJCALLPROP:
                        goto do_opc_objcallprop;

                    case OPC_INHERIT:
                        goto do_opc_inherit;

                    case OPC_PTRINHERIT:
                        goto do_opc_ptrinherit;

                    case OPC_EXPINHERIT:
                        goto do_opc_expinherit;

                    case OPC_PTREXPINHERIT:
                        goto do_opc_ptrexpinherit;

                    case OPC_BUILTIN_A:
                        goto do_opc_builtin_a;

                    case OPC_BUILTIN_B:
                        goto do_opc_builtin_b;

                    case OPC_BUILTIN_C:
                        goto do_opc_builtin_c;

                    case OPC_BUILTIN_D:
                        goto do_opc_builtin_d;

                    case OPC_BUILTIN1:
                        goto do_opc_builtin1;

                    case OPC_BUILTIN2:
                        goto do_opc_builtin2;

                    case OPC_NEW1:
                        goto do_opc_new1;

                    case OPC_NEW2:
                        goto do_opc_new2;

                    default:
                        err_throw(VMERR_INVALID_OPCODE_MOD);
                        break;
                    }
                }
                break;

            case OPC_CALL:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_call:
                /* get the code offset to invoke */
                ofs = get_op_int32(&p);

                /* call it */
                do_call(vmg_ &p, ofs, argc, VM_INVALID_OBJ, FALSE, 0);
                break;

            case OPC_PTRCALL:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_ptrcall:
                /* retrieve the target of the call */
                popval(vmg_ &val);

                /* 
                 *   if it's a prop ID, and there's a valid "self" object,
                 *   treat it as a PTRCALLPROPSELF 
                 */
                if (val.typ == VM_PROP && get_self(vmg0_) != VM_INVALID_OBJ)
                    goto do_opc_ptrcallpropself_val;
                
                /* call the function */
                call_func_ptr(vmg_ &val, argc, FALSE, 0, &p);
                break;

            case OPC_RETVAL:
                /* pop the return value into R0 */
                popval(vmg_ &r0_);

                /* return */
                if (!do_return(vmg_ &p))
                    goto exit_loop;
                break;

            case OPC_RET:
                /* return with no value - do not change R0 */
                if (!do_return(vmg_ &p))
                    goto exit_loop;
                break;

            case OPC_RETNIL:
                /* store nil in R0 */
                r0_.set_nil();

                /* return */
                if (!do_return(vmg_ &p))
                    goto exit_loop;
                break;

            case OPC_RETTRUE:
                /* store true in R0 */
                r0_.set_true();

                /* retur */
                if (!do_return(vmg_ &p))
                    goto exit_loop;
                break;

            case OPC_GETPROP:
                /* get the object whose property we're fetching */
                G_stk->pop(&val);

                /* evaluate the property given by the immediate data */
                get_prop(vmg_ &p, &val, get_op_uint16(&p), &val, 0);
                break;

            case OPC_CALLPROP:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_callprop:
                /* pop the object whose property we're fetching */
                G_stk->pop(&val);

                /* evaluate the property given by the immediate data */
                get_prop(vmg_ &p, &val, get_op_uint16(&p), &val, argc);
                break;

            case OPC_PTRCALLPROP:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_ptrcallprop:
                /* 
                 *   pop the property to be evaluated, and the object
                 *   whose property we're evaluating 
                 */
                pop_prop(vmg_ &val);
                G_stk->pop(&val2);
                
                /* evaluate the property */
                get_prop(vmg_ &p, &val2, val.val.prop, &val2, argc);
                break;

            case OPC_GETPROPSELF:
                /* evaluate the property of 'self' */
                val.set_obj(get_self(vmg0_));
                get_prop(vmg_ &p, &val, get_op_uint16(&p), &val, 0);
                break;

            case OPC_CALLPROPSELF:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_callpropself:
                /* evaluate the property of 'self' */
                val.set_obj(get_self(vmg0_));
                get_prop(vmg_ &p, &val, get_op_uint16(&p), &val, argc);
                break;

            case OPC_PTRCALLPROPSELF:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_ptrcallpropself:
                /* get the property to be evaluated */
                pop_prop(vmg_ &val);

            do_opc_ptrcallpropself_val:
                /* evaluate the property of 'self' */
                val2.set_obj(get_self(vmg0_));
                get_prop(vmg_ &p, &val2, val.val.prop, &val2, argc);
                break;

            case OPC_OBJGETPROP:
                /* get the object */
                val.set_obj((vm_obj_id_t)get_op_uint32(&p));
                
                /* evaluate the property */
                get_prop(vmg_ &p, &val, get_op_uint16(&p), &val, 0);
                break;

            case OPC_OBJCALLPROP:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_objcallprop:
                /* get the object */
                val.set_obj((vm_obj_id_t)get_op_uint32(&p));
                
                /* evaluate the property */
                get_prop(vmg_ &p, &val, get_op_uint16(&p), &val, argc);
                break;

            case OPC_GETPROPDATA:
                /* get the object whose property we're fetching */
                G_stk->pop(&val);

                /* 
                 *   if the object is not an object, it's one of the
                 *   native types, in which case we'll definitely run
                 *   native code to evaluate the property, in which case
                 *   it's not valid for speculative evaluation 
                 */
                if (val.typ != VM_OBJ)
                    err_throw(VMERR_BAD_SPEC_EVAL);

                /* get the property */
                prop = (vm_prop_id_t)get_op_uint16(&p);

                /* check validity for speculative evaluation */
                check_prop_spec_eval(vmg_ val.val.obj, prop);

                /* evaluate the property given by the immediate data */
                get_prop(vmg_ &p, &val, prop, &val, 0);
                break;

            case OPC_PTRGETPROPDATA:
                /* get the property and object to evaluate */
                pop_prop(vmg_ &val);
                G_stk->pop(&val2);

                /* 
                 *   if the object is not an object, it's one of the
                 *   native types, in which case we'll definitely run
                 *   native code to evaluate the property, in which case
                 *   it's not valid for speculative evaluation 
                 */
                if (val2.typ != VM_OBJ)
                    err_throw(VMERR_BAD_SPEC_EVAL);

                /* check validity for speculative evaluation */
                check_prop_spec_eval(vmg_ val2.val.obj, val.val.prop);

                /* evaluate it */
                get_prop(vmg_ &p, &val2, val.val.prop, &val2, 0);
                break;

            case OPC_GETLCL1:
                /* push the local */
                pushval(vmg_ get_local(vmg_ get_op_uint8(&p)));
                break;

            case OPC_GETLCL2:
                /* push the local */
                pushval(vmg_ get_local(vmg_ get_op_uint16(&p)));
                break;

            case OPC_GETARG1:
                /* push the argument */
                pushval(vmg_ get_param(vmg_ get_op_uint8(&p)));
                break;

            case OPC_GETARG2:
                /* push the argument */
                pushval(vmg_ get_param(vmg_ get_op_uint16(&p)));
                break;

            case OPC_PUSHSELF:
                /* push 'self' */
                push_obj(vmg_ get_self(vmg0_));
                break;

            case OPC_SETSELF:
                /* pop the value */
                pop_obj(vmg_ &val);

                /* set 'self' to the value */
                set_self(vmg_
                         val.typ == VM_NIL ? VM_INVALID_OBJ : val.val.obj);
                break;

            case OPC_GETARGC:
                /* push the argument counter */
                push_int(vmg_ get_cur_argc(vmg0_));
                break;

            case OPC_DUP:
                /* re-push the item at top of stack */
                pushval(vmg_ G_stk->get(0));
                break;

            case OPC_SWAP:
                /* pop the top two elements of the stack */
                popval(vmg_ &val);
                popval(vmg_ &val2);

                /* re-push the elements, reversing their order on the stack */
                pushval(vmg_ &val);
                pushval(vmg_ &val2);
                break;

            case OPC_DISC:
                /* discard the item at the top of the stack */
                G_stk->discard();
                break;

            case OPC_DISC1:
                /* discard n items */
                G_stk->discard(get_op_uint8(&p));
                break;

            case OPC_GETR0:
                /* push the contents of R0 */
                pushval(vmg_ &r0_);
                break;

            case OPC_GETDBARGC:
                /* push the argument count from the selected frame */
                push_int(vmg_ get_argc_at_level(vmg_ get_op_uint16(&p) + 1));
                break;

            case OPC_GETDBLCL:
                /* get the local variable number and stack level */
                idx = get_op_uint16(&p);
                level = get_op_uint16(&p);

                /* push the value */
                pushval(vmg_ get_local_at_level(vmg_ idx, level + 1));
                break;
                
            case OPC_GETDBARG:
                /* get the parameter variable number and stack level */
                idx = get_op_uint16(&p);
                level = get_op_uint16(&p);

                /* push the value */
                pushval(vmg_ get_param_at_level(vmg_ idx, level + 1));
                break;

            case OPC_SETDBLCL:
                /* get the local variable number and stack level */
                idx = get_op_uint16(&p);
                level = get_op_uint16(&p);

                /* get the local pointer */
                valp = get_local_at_level(vmg_ idx, level + 1);

                /* pop the value into the local */
                popval(vmg_ valp);
                break;

            case OPC_SETDBARG:
                /* get the parameter variable number and stack level */
                idx = get_op_uint16(&p);
                level = get_op_uint16(&p);

                /* get the parameter pointer */
                valp = get_param_at_level(vmg_ idx, level + 1);

                /* pop the value into the local */
                popval(vmg_ valp);
                break;

            case OPC_SWITCH:
                /* pop the control value */
                popval(vmg_ &val);

                /* get the case count */
                cnt = get_op_uint16(&p);

                /* iterate through the case table */
                for ( ; cnt != 0 ; p += 7, --cnt)
                {
                    /* get this value */
                    vmb_get_dh((const char *)p, &val2);

                    /* check if the values match */
                    if (val.equals(vmg_ &val2))
                    {
                        /* it matches - jump to this offset */
                        p += VMB_DATAHOLDER;
                        p += osrp2s(p);

                        /* no need to look any further */
                        break;
                    }
                }

                /* if we didn't find it, jump to the default case */
                if (cnt == 0)
                    p += osrp2s(p);
                break;

            case OPC_JMP:
                /* unconditionally jump to the given offset */
                p += osrp2s(p);
                break;

            case OPC_JT:
                /* pop the value */
                popval(vmg_ &val);

                /* if it's true, or a non-zero numeric value, jump */
                if (val.is_logical())
                    p += (val.typ == VM_TRUE ? osrp2s(p) : 2);
                else if (val.is_numeric())
                    p += (!val.num_is_zero() ? osrp2s(p) : 2);
                else
                    err_throw(VMERR_LOG_VAL_REQD);
                break;

            case OPC_JF:
                /* pop the value */
                popval(vmg_ &val);

                /* if it's nil or zero, jump */
                if (val.is_logical())
                    p += (val.typ == VM_NIL ? osrp2s(p) : 2);
                else if (val.is_numeric())
                    p += (val.num_is_zero() ? osrp2s(p) : 2);
                else
                {
                    /* it's not nil or zero - consider it non-false */
                    p += 2;
                }
                break;

            case OPC_JE:
                /* jump if the two values at top of stack are equal */
                p += (pop2_equal(vmg0_) ? osrp2s(p) : 2);
                break;

            case OPC_JNE:
                /* jump if the two values at top of stack are not equal */
                p += (!pop2_equal(vmg0_) ? osrp2s(p) : 2);
                break;

            case OPC_JGT:
                /* jump if greater */
                p += (pop2_compare(vmg0_) > 0 ? osrp2s(p) : 2);
                break;

            case OPC_JGE:
                /* jump if greater or equal */
                p += (pop2_compare(vmg0_) >= 0 ? osrp2s(p) : 2);
                break;

            case OPC_JLT:
                /* jump if less */
                p += (pop2_compare(vmg0_) < 0 ? osrp2s(p) : 2);
                break;

            case OPC_JLE:
                /* jump if less or equal */
                p += (pop2_compare(vmg0_) <= 0 ? osrp2s(p) : 2);
                break;

            case OPC_JST:
                /* get (do not remove) the element at top of stack */
                valp = G_stk->get(0);

                /* 
                 *   if it's true or a non-zero number, jump, saving the
                 *   value; otherwise, require that it be a logical value,
                 *   pop it, and proceed 
                 */
                if (valp->typ == VM_TRUE
                    || valp->typ == VM_ENUM
                    || valp->typ == VM_INT && !valp->num_is_zero())
                {
                    /* it's true - save it and jump */
                    p += osrp2s(p);
                }
                else
                {
                    /* 
                     *   it's not true - discard the value, but require
                     *   that it be a valid logical value 
                     */
                    if (valp->typ != VM_NIL && valp->typ != VM_INT)
                        err_throw(VMERR_LOG_VAL_REQD);
                    G_stk->discard();

                    /* skip to the next instruction */
                    p += 2;
                }
                break;

            case OPC_JSF:
                /* get (do not remove) the element at top of stack */
                valp = G_stk->get(0);

                /* 
                 *   if it's nil or zero, jump, saving the value;
                 *   otherwise, discard the value and proceed 
                 */
                if (valp->typ == VM_NIL
                    || valp->typ == VM_INT && valp->num_is_zero())
                {
                    /* it's nil or zero - save it and jump */
                    p += osrp2s(p);
                }
                else
                {
                    /* it's something non-false - discard it */
                    G_stk->discard();

                    /* skip to the next instruction */
                    p += 2;
                }
                break;

            case OPC_LJSR:
                /* 
                 *   compute and push the offset of the next instruction
                 *   (at +2 because of the branch offset operand) from our
                 *   method header - this will be the return address,
                 *   which in this offset format will survive any code
                 *   swapping that might occur in subsequent execution 
                 */
                push_int(vmg_ pcptr_to_method_ofs(vmg_ p + 2));

                /* jump to the target address */
                p += osrp2s(p);
                break;

            case OPC_LRET:
                /* get the indicated local variable */
                valp = get_local(vmg_ get_op_uint16(&p));
                
                /* the value must be an integer */
                if (valp->typ != VM_INT)
                    err_throw(VMERR_INT_VAL_REQD);
                
                /* 
                 *   jump to the code address obtained from adding the
                 *   integer value in the given local variable to the
                 *   current method header pointer 
                 */
                p = (const uchar *)G_code_pool->get_ptr(entry_ptr_)
                    + valp->val.intval;
                break;

            case OPC_SAY:
                /* get the string offset */
                ofs = get_op_int32(&p);

                /* display it */
                disp_dstring(vmg_ ofs, &p, get_self(vmg0_));
                break;

            case OPC_SAYVAL:
                /* invoke the default string display function */
                disp_string_val(vmg_ &p, get_self(vmg0_));
                break;

            case OPC_THROW:
                /* pop the exception object */
                pop_obj(vmg_ &val);

                /* throw it */
                if (!do_throw(vmg_ &p, val.val.obj))
                {
                    /* remember the unhandled exception for re-throwing */
                    unhandled_exc = val.val.obj;

                    /* terminate execution */
                    goto exit_loop;
                }
                break;

            case OPC_INHERIT:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_inherit:
                /* inherit the property */
                inh_prop(vmg_ &p, (vm_prop_id_t)get_op_uint16(&p), argc);
                break;

            case OPC_PTRINHERIT:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_ptrinherit:
                /* pop the property to be inherited */
                pop_prop(vmg_ &val);

                /* inherit it */
                inh_prop(vmg_ &p, val.val.prop, argc);
                break;

            case OPC_EXPINHERIT:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_expinherit:
                /* get the property to inherit */
                prop = (vm_prop_id_t)get_op_uint16(&p);

                /* get the superclass to inherit it from */
                val.set_obj((vm_obj_id_t)get_op_uint32(&p));

                /* 
                 *   inherit it -- process this essentially the same way
                 *   as a normal CALLPROP, since we're going to evaluate
                 *   the given property of the given object, but retain
                 *   the current 'self' object 
                 */
                val2.set_obj(get_self(vmg0_));
                get_prop(vmg_ &p, &val, prop, &val2, argc);
                break;

            case OPC_PTREXPINHERIT:
                /* get the argument count */
                argc = get_op_uint8(&p);

            do_opc_ptrexpinherit:
                /* pop the property to inherit */
                pop_prop(vmg_ &val);

                /* get the superclass to inherit it from */
                val3.set_obj((vm_obj_id_t)get_op_uint32(&p));

                /* inherit it */
                val2.set_obj(get_self(vmg0_));
                get_prop(vmg_ &p, &val3, val.val.prop, &val2, argc);
                break;

            case OPC_BUILTIN_A:
                /* get the function index and argument count */
                argc = get_op_uint8(&p);

            do_opc_builtin_a:
                idx = get_op_uint8(&p);

                /* call the function in set #0 */
                call_bif(vmg_ 0, idx, argc, &p);
                break;

            case OPC_BUILTIN_B:
                /* get the function index and argument count */
                argc = get_op_uint8(&p);

            do_opc_builtin_b:
                idx = get_op_uint8(&p);

                /* call the function in set #1 */
                call_bif(vmg_ 1, idx, argc, &p);
                break;

            case OPC_BUILTIN_C:
                /* get the function index and argument count */
                argc = get_op_uint8(&p);

            do_opc_builtin_c:
                idx = get_op_uint8(&p);

                /* call the function in set #2 */
                call_bif(vmg_ 2, idx, argc, &p);
                break;

            case OPC_BUILTIN_D:
                /* get the function index and argument count */
                argc = get_op_uint8(&p);

            do_opc_builtin_d:
                idx = get_op_uint8(&p);

                /* call the function in set #3 */
                call_bif(vmg_ 3, idx, argc, &p);
                break;

            case OPC_BUILTIN1:
                /* get the function index and argument count */
                argc = get_op_uint8(&p);

            do_opc_builtin1:
                idx = get_op_uint8(&p);

                /* get the function set ID */
                set_idx = get_op_uint8(&p);

                /* call the function in set #0 */
                call_bif(vmg_ set_idx, idx, argc, &p);
                break;

            case OPC_BUILTIN2:
                /* get the function index and argument count */
                argc = get_op_uint8(&p);

            do_opc_builtin2:
                idx = get_op_uint16(&p);

                /* get the function set ID */
                set_idx = get_op_uint8(&p);

                /* call the function in set #0 */
                call_bif(vmg_ set_idx, idx, argc, &p);
                break;

            case OPC_CALLEXT:
                //$$$
                err_throw(VMERR_CALLEXT_NOT_IMPL);
                break;

            case OPC_INDEX:
                /* pop the index value */
                popval(vmg_ &val2);

                /* pop the object to index */
                popval(vmg_ &val);

                /* look up the indexed value */
                index_and_push(vmg_ &val, &val2);
                break;

            case OPC_BP:
                /* step back to the breakpoint location itself */
                VM_IF_DEBUGGER(--p);

                /* let the debugger take control */
                VM_IF_DEBUGGER(G_debugger->step(vmg_ &p, entry_ptr_, TRUE));

                /* if there's no debugger, it's an error */
                VM_IF_NOT_DEBUGGER(err_throw(VMERR_BREAKPOINT));

                /* 
                 *   go back and execute the current instruction - bypass
                 *   single-step tracing into the debugger in this case,
                 *   since the debugger expects when it returns that one
                 *   instruction will always be traced before the debugger
                 *   is re-entered 
                 */
                goto exec_instruction;

            case OPC_NOP:
                /* NO OP - no effect */
                break;

            case OPC_NEW1:
                /* get the metaclass ID and argument count */
                argc = get_op_uint8(&p);

            do_opc_new1:
                idx = get_op_uint8(&p);
                
                /* create the new object */
                new_and_store_r0(vmg_ &p, idx, argc);
                break;

            case OPC_NEW2:
                /* get the argument count and metaclass ID */
                argc = get_op_uint16(&p);

            do_opc_new2:
                idx = get_op_uint16(&p);

                /* create the new object */
                new_and_store_r0(vmg_ &p, idx, argc);
                break;

            case OPC_INCLCL:
                /* get the local */
                valp = get_local(vmg_ get_op_uint16(&p));
                
                /* the local must be a number */
                if (!valp->is_numeric())
                    err_throw(VMERR_NUM_VAL_REQD);

                /* increment it */
                ++(valp->val.intval);
                break;

            case OPC_DECLCL:
                /* get the local */
                valp = get_local(vmg_ get_op_uint16(&p));

                /* the local must be a number */
                if (!valp->is_numeric())
                    err_throw(VMERR_NUM_VAL_REQD);

                /* decrement it */
                --(valp->val.intval);
                break;

            case OPC_ADDILCL1:
                /* get the local */
                valp = get_local(vmg_ get_op_uint8(&p));

                /* get the number to add */
                val2.set_int(get_op_int8(&p));

                /* compute the sum, leaving the result in the local */
                compute_sum(vmg_ valp, &val2);
                break;

            case OPC_ADDILCL4:
                /* get the local */
                valp = get_local(vmg_ get_op_uint16(&p));

                /* get the number to add */
                val2.set_int(get_op_int32(&p));

                /* compute the sum, leaving the result in the local */
                compute_sum(vmg_ valp, &val2);
                break;

            case OPC_ADDTOLCL:
                /* get the local */
                valp = get_local(vmg_ get_op_uint16(&p));

                /* pop the value to add */
                popval(vmg_ &val2);

                /* compute the sum, leaving the result in the local */
                compute_sum(vmg_ valp, &val2);
                break;

            case OPC_SUBFROMLCL:
                /* get the local */
                valp = get_local(vmg_ get_op_uint16(&p));

                /* pop the value to add */
                popval(vmg_ &val2);

                /* subtract the values, leaving the result in the local */
                compute_diff(vmg_ valp, &val2);
                break;

            case OPC_ZEROLCL1:
                /* get the local and set it to zero */
                get_local(vmg_ get_op_uint8(&p))->set_int(0);
                break;

            case OPC_ZEROLCL2:
                /* get the local and set it to zero */
                get_local(vmg_ get_op_uint16(&p))->set_int(0);
                break;

            case OPC_NILLCL1:
                /* get the local and set it to zero */
                get_local(vmg_ get_op_uint8(&p))->set_nil();
                break;

            case OPC_NILLCL2:
                /* get the local and set it to zero */
                get_local(vmg_ get_op_uint16(&p))->set_nil();
                break;

            case OPC_ONELCL1:
                /* get the local and set it to zero */
                get_local(vmg_ get_op_uint8(&p))->set_int(1);
                break;

            case OPC_ONELCL2:
                /* get the local and set it to zero */
                get_local(vmg_ get_op_uint16(&p))->set_int(1);
                break;

            case OPC_SETLCL1:
                /* get a pointer to the local */
                valp = get_local(vmg_ get_op_uint8(&p));
                
                /* pop the value into the local */
                popval(vmg_ valp);
                break;

            case OPC_SETLCL2:
                /* get a pointer to the local */
                valp = get_local(vmg_ get_op_uint16(&p));

                /* pop the value into the local */
                popval(vmg_ valp);
                break;

            case OPC_SETARG1:
                /* get a pointer to the parameter */
                valp = get_param(vmg_ get_op_uint8(&p));

                /* pop the value into the parameter */
                popval(vmg_ valp);
                break;

            case OPC_SETARG2:
                /* get a pointer to the parameter */
                valp = get_param(vmg_ get_op_uint16(&p));

                /* pop the value into the parameter */
                popval(vmg_ valp);
                break;

            case OPC_SETIND:
                /* pop the index */
                popval(vmg_ &val2);

                /* pop the value to be indexed */
                popval(vmg_ &val);

                /* pop the value to assign */
                popval(vmg_ &val3);

                /* assign into the index, pushing the result */
                set_index_and_push(vmg_ &val, &val2, &val3);
                break;

            case OPC_SETPROP:
                /* get the object whose property we're setting */
                pop_obj(vmg_ &val);

                /* pop the value we're setting */
                popval(vmg_ &val2);

                /* set the value */
                set_prop(vmg_ val.val.obj, get_op_uint16(&p), &val2);
                break;

            case OPC_PTRSETPROP:
                /* get the property and object to set */
                pop_prop(vmg_ &val);
                pop_obj(vmg_ &val2);

                /* get the value to set */
                popval(vmg_ &val3);

                /* set it */
                set_prop(vmg_ val2.val.obj, val.val.prop, &val3);
                break;

            case OPC_SETPROPSELF:
                /* get the value to set */
                popval(vmg_ &val);

                /* set it */
                set_prop(vmg_ get_self(vmg0_), get_op_uint16(&p), &val);
                break;

            case OPC_OBJSETPROP:
                /* get the objet */
                obj = (vm_obj_id_t)get_op_uint32(&p);

                /* get the new value */
                popval(vmg_ &val);

                /* set the property */
                set_prop(vmg_ obj, get_op_uint16(&p), &val);
                break;

            default:
                /* unrecognized opcode */
                err_throw(VMERR_INVALID_OPCODE);
                break;
            }
        }

        /*
         *   We jump to this label when it's time to terminate execution
         *   and return to the host environment which called us. 
         */
    exit_loop:
        /* note that we're ready to return */
        done = TRUE;
    }
    err_catch(err)
    {
        int i;

        err_try
        {
            /* 
             *   Create a new exception object to describe the error.  The
             *   arguments to the constructor are the error number and the
             *   error parameters.
             *   
             *   If the error code is "unhandled exception," it means that
             *   an exception occurred in a recursive interpreter
             *   invocation, and the exception wasn't handled within the
             *   code called recursively; in this case, we can simply
             *   re-throw the original error, and perhaps handle it in the
             *   context of the current code.  
             */
            if (err->get_error_code() == VMERR_UNHANDLED_EXC)
            {
                /* get the original exception object from the error stack */
                obj = (vm_obj_id_t)err->get_param_ulong(0);
            }
            else
            {
                /* push the error parameters (in reverse order) */
                for (i = err->get_param_count() ; i > 0 ; )
                {
                    /* go to the next parameter */
                    --i;
                    
                    /* see what we have and push an appropriate value */
                    switch(err->get_param_type(i-1))
                    {
                    case ERR_TYPE_INT:
                        /* push the integer value */
                        push_int(vmg_ err->get_param_int(i));
                        break;
                        
                    case ERR_TYPE_ULONG:
                        /* push the value */
                        push_int(vmg_ (int32)err->get_param_ulong(i));
                        break;
                        
                    case ERR_TYPE_TEXTCHAR:
                        /* push a new string with the text */
                        push_obj(vmg_ CVmObjString::create(vmg_ FALSE,
                            err->get_param_text(i),
                            get_strlen(err->get_param_text(i))));
                        break;
                        
                    case ERR_TYPE_CHAR:
                        /* push a new string with the text */
                        push_obj(vmg_ CVmObjString::create(vmg_ FALSE,
                            err->get_param_char(i),
                            strlen(err->get_param_char(i))));
                        break;
                        
                    default:
                        /* unrecognized type - push nil for now */
                        push_nil(vmg0_);
                        break;
                    }
                }
                
                /* 
                 *   if there's a RuntimeError base class defined, create
                 *   an instance; otherwise, create a simple instance of
                 *   the basic object type to throw as a placeholder,
                 *   since the program hasn't made any provision to catch
                 *   run-time errors 
                 */
                if (G_predef->rterr != VM_INVALID_OBJ)
                {
                    /* push the error number */
                    push_int(vmg_ err->get_error_code());
                    
                    /* 
                     *   Create the new RuntimeException instance.  Run
                     *   the constructor in a recursive invocation of the
                     *   interpreter (by passing a null pc_ptr).  
                     */
                    vm_objp(vmg_ G_predef->rterr)
                        ->create_instance(vmg_ G_predef->rterr, 0,
                                          err->get_param_count() + 1);
                    
                    /* get the object from R0 */
                    if (r0_.typ != VM_OBJ)
                        err_throw(VMERR_OBJ_VAL_REQD);
                    obj = r0_.val.obj;
                }
                else
                {
                    /* 
                     *   There's no RuntimeError object defined by the
                     *   image file, so create a basic object to throw.
                     *   This won't convey any information to the program
                     *   except that it's not one of the errors they're
                     *   expecting; this is fine, since they have made no
                     *   provisions to catch VM errors, as demonstrated by
                     *   their lack of a RuntimeError definition.  
                     */
                    obj = CVmObjTads::create(vmg_ FALSE, 0, 1);
                }

                /* 
                 *   if possible, set the exceptionMessage property in the
                 *   new exception object to the default error message for
                 *   the run-time error we're processing 
                 */
                if (G_predef->rterrmsg_prop != VM_INVALID_PROP)
                {
                    const char *msg;
                    char buf[256];
                    vm_obj_id_t str_obj;
                    
                    /* format the message text */
                    msg = err_get_msg(vm_messages, vm_message_count,
                                      err->get_error_code(), FALSE);
                    err_format_msg(buf, sizeof(buf), msg, err);
                    
                    /* 
                     *   momentarily push the new exception object, so we
                     *   don't lose track of it if we run garbage
                     *   collection here 
                     */
                    push_obj(vmg_ obj);
                    
                    /* create a string object with the message text */
                    str_obj =
                        CVmObjString::create(vmg_ FALSE, buf, strlen(buf));

                    /* set the property in the new object */
                    val.set_obj(str_obj);
                    vm_objp(vmg_ obj)
                        ->set_prop(vmg_ G_undo, obj,
                                   G_predef->rterrmsg_prop, &val);

                    /* we don't need gc protection any more */
                    G_stk->discard();
                }
            }

            /* discard the exception, since we're going to deal with it */
            err_pop_exc();

            /* throw the exception */
            if (!do_throw(vmg_ &p, obj))
            {
                /* remember the unhandled exception for a moment */
                unhandled_exc = obj;
            }
        }
        err_catch(exc2)
        {
            /* 
             *   we got another exception trying to handle the first
             *   exception - just throw the error again, but at least
             *   clean up statics on the way out 
             */
            pc_ptr_ = old_pc_ptr;

            /* re-throw the error */
            err_rethrow();
        }
        err_end;
    }
    err_end;

    /* if an unhandled exception occurred, re-throw it */
    if (unhandled_exc != VM_INVALID_OBJ)
        err_throw_a(VMERR_UNHANDLED_EXC, 1,
                    ERR_TYPE_ULONG, (unsigned long)unhandled_exc);

    /* if we're not done, go back and resume execution */
    if (!done)
        goto resume_execution;

    /* restore the enclosing PC pointer */
    pc_ptr_ = old_pc_ptr;
}


/* ------------------------------------------------------------------------ */
/*
 *   Throw an exception.  Returns true if an exception handler was found,
 *   which means that execution can proceed; returns false if no handler
 *   was found, in which case the execution loop must throw the exception
 *   to its caller.  
 */
int CVmRun::do_throw(VMG_ const uchar **pcptr, vm_obj_id_t exception_obj)
{
    /*
     *   Search the stack for a handler for this exception class.  Start
     *   at the current stack frame; if we find a handler here, use it;
     *   otherwise, unwind the stack to the enclosing frame and search for
     *   a handler there; repeat until we exhaust the stack.  
     */
    for (;;)
    {
        CVmExcTablePtr tab;
        const uchar *func_start;
        uint ofs;

        /* 
         *   Step back one byte in the byte-code stream.  This is
         *   necessary because the program counter always points to the
         *   *next* instruction to be executed; we want the exception
         *   handler that encloses the instruction that actually called
         *   the code that threw the exception.  It's adequate to back up
         *   by one byte, because the handler range will always include
         *   all of the bytes in the final instruction in the range.
         *   
         *   Note that this program counter value isn't ever going to be
         *   used to execute any code -- it's only going to be used to
         *   select an exception handler.  If we don't find an exception
         *   handler, we'll just unwind another level, so this pc value
         *   won't be used.  If we do find an exception handler, pc will
         *   jump to the exception handler, hence this pc value won't be
         *   used.  So, it's unimportant that pc may be pointing into the
         *   middle of an instruction here.  
         */
        --(*pcptr);

        /* get a pointer to the start of the current function */
        func_start = (const uchar *)G_code_pool->get_ptr(entry_ptr_);

        /* set up a pointer to the current exception table */
        if (tab.set(func_start))
        {
            size_t cnt;
            size_t i;
            CVmExcEntryPtr entry;

            /* calculate our offset in the current function */
            ofs = *pcptr - func_start;

            /* set up a pointer to the first table entry */
            tab.set_entry_ptr(vmg_ &entry, 0);
            
            /* loop through the entries */
            for (i = 0, cnt = tab.get_count() ; i < cnt ;
                 ++i, entry.inc(vmg0_))
            {
                /* 
                 *   Check to see if we're in the range for this entry.
                 *   If this entry covers the appropriate range, and the
                 *   exception we're handling is of the class handled by
                 *   this exception (or derives from that class), this
                 *   handler handles this exception. 
                 */
                if (ofs >= entry.get_start_ofs()
                    && ofs <= entry.get_end_ofs()
                    && (entry.get_exception() == VM_INVALID_OBJ
                        || exception_obj == entry.get_exception()
                        || (vm_objp(vmg_ exception_obj)
                            ->is_instance_of(vmg_ entry.get_exception()))))
                {
                    /* 
                     *   this is it - move the program counter to the
                     *   first byte of the handler's code 
                     */
                    *pcptr = func_start + entry.get_handler_ofs();

                    /* push the exception so that the handler can get at it */
                    push_obj(vmg_ exception_obj);

                    /* indicate that we found a handler for the exception */
                    return TRUE;
                }
            }
        }

        /* 
         *   We didn't find a handler in the current function - unwind the
         *   stack one level, using an ordinary RETURN operation (we're
         *   not really returning, though, so we don't need to provide a
         *   return value).  
         */
        if (!do_return(vmg_ pcptr))
        {
            /* 
             *   We were at the outermost stack frame already, so there's
             *   nothing left to unwind.  Return false to indicate that
             *   the exception was not handled. 
             */
            return FALSE;
        }
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Call a built-in function 
 */
void CVmRun::call_bif(VMG_ uint set_index, uint func_index, uint argc,
                      const uchar ** /*pc_ptr*/)
{
    /* 
     *   Call the function -- presume the compiler has ensured that the
     *   function set index is valid for the load image, and that the
     *   function index is valid for the function set; all of this can be
     *   determined at compile time, since function sets are statically
     *   defined. 
     */
    G_bif_table->call_func(vmg_ set_index, func_index, argc);
}


/* ------------------------------------------------------------------------ */
/*
 *   Call a function pointer 
 */
void CVmRun::call_func_ptr(VMG_ const vm_val_t *funcptr, uint argc,
                           int recurse, const char *recurse_name,
                           const uchar **pc_ptr)
{
    /* 
     *   if it's an object, and the predefined property ObjectCallProp is
     *   defined, and the object defines this property, call this property
     *   in the object 
     */
    if (funcptr->typ == VM_OBJ
        && G_predef->obj_call_prop != VM_INVALID_PROP)
    {
        vm_val_t prop_val;
        vm_obj_id_t srcobj;
        int found;
        uint objcall_argc = 0;
        
        /* make sure the object defines ObjectCallProp */
        found = vm_objp(vmg_ funcptr->val.obj)
                ->get_prop(vmg_ G_predef->obj_call_prop, &prop_val,
                           funcptr->val.obj, &srcobj, &objcall_argc);
        
        /* 
         *   if we didn't find it, this object can't be used in this
         *   fashion - throw an error 
         */
        if (!found)
            err_throw(VMERR_FUNCPTR_VAL_REQD);
        
        /* 
         *   if this is a function pointer, call the function pointer with
         *   the function object as 'self' 
         */
        if (prop_val.typ == VM_FUNCPTR)
        {
            /* call the function */
            do_call(vmg_ pc_ptr, prop_val.val.ofs, argc,
                    funcptr->val.obj, recurse, recurse_name);
            
            /* handled */
            return;
        }
        
        /* proceed with the new value */
        funcptr = &prop_val;
    }
    
    /* if it's not a function pointer, it's an error */
    if (funcptr->typ != VM_FUNCPTR)
        err_throw(VMERR_FUNCPTR_VAL_REQD);
    
    /* call the function */
    do_call(vmg_ pc_ptr, funcptr->val.ofs, argc, VM_INVALID_OBJ,
            recurse, recurse_name);
}

/* ------------------------------------------------------------------------ */
/*
 *   Call a function 
 */
void CVmRun::do_call(VMG_ const uchar **pc_ptr,
                     pool_ofs_t call_ofs, uint argc, vm_obj_id_t self,
                     int recurse, const char *recurse_name)
{
    pool_ofs_t old_ofs;
    const uchar *p;
    CVmFuncPtr hdr_ptr;
    uint i;

    /* store nil in R0 */
    r0_.set_nil();

    /* 
     *   If we have a debugger, and this is a recursive call, set up a
     *   frame for the recursive call, so that the debugger can look up
     *   the stack to the byte-code caller of the native code that's
     *   recursing into the VM.
     *   
     *   This is unnecessary if there's no debugger; the only reason we
     *   need a special frame on native recursion is to allow the debugger
     *   to traverse the stack correctly through the native call.  
     */
    VM_IF_DEBUGGER(if (pc_ptr == 0) enter_recursive_frame(vmg_ argc));

    /* 
     *   Compute the current code offset.  Note that this 'get_ptr'
     *   operation will definitely not involve swapping in any code, and
     *   hence won't involve swapping out any code, because the current
     *   method must necessarily be in memory.
     *   
     *   If pc_ptr is null, it means we're doing a recursive call; simply
     *   store zero as the caller's code offset in this case.  Zero is
     *   always an invalid method offset, since there's never executable
     *   code at offset zero in a method (since the method header is
     *   stored there).  The RETURN code will detect this condition and
     *   realize that it must return from the interpreter loop.  
     */
    old_ofs = (pool_ofs_t)(pc_ptr == 0
                           ? 0 : pcptr_to_method_ofs(vmg_ *pc_ptr));

    /* 
     *   We're done with the old code segment now, so we can safely
     *   translate a new address.  Get the physical address we're calling
     *   -- this will swap in the new code segment if necessary.  
     */
    p = (const uchar *)G_code_pool->get_ptr(call_ofs);

    /* set up a pointer to the new function header */
    hdr_ptr.set(p);

    /* 
     *   Get the space needs of the new function, and ensure we have
     *   enough stack space available.  Include the size of the frame that
     *   we store (the 'self' object, the caller's code offset, the
     *   caller's entrypoint offset, the actual parameter count, and the
     *   enclosing frame pointer) in our space needs.  
     */
    if (!G_stk->check_space(hdr_ptr.get_stack_depth() + 5))
        err_throw(VMERR_STACK_OVERFLOW);

    /* push 'self' */
    push_obj(vmg_ self);

    /* push the current code offset */
    push_codeofs(vmg_ old_ofs);

    /* push the current entrypoint code offset */
    push_codeofs(vmg_ entry_ptr_);

    /* push the actual parameter count */
    push_int(vmg_ (int32)argc);

    /* push the current frame pointer */
    push_stackptr(vmg_ frame_ptr_);

    /* 
     *   set up the new frame so that the frame pointer points to the old
     *   frame pointer stored in the stack 
     */
    frame_ptr_ = G_stk->get_sp();

    /* load EP with the new code offset */
    entry_ptr_ = call_ofs;

    /* check the argument count */
    if (!hdr_ptr.argc_ok(argc))
    {
        /* 
         *   if we're making a recursive call, throw an error indicating
         *   what kind of recursive call we're making 
         */
        if (recurse && recurse_name != 0)
        {
            /* throw the named generic argument mismatch error */
            err_throw_a(VMERR_WRONG_NUM_OF_ARGS_CALLING, 1,
                        ERR_TYPE_CHAR, recurse_name);
        }
        else
        {
            /* throw the generic argument mismatch error */
            err_throw(VMERR_WRONG_NUM_OF_ARGS);
        }
    }

    /* push nil for each local */
    for (i = hdr_ptr.get_local_cnt() ; i != 0 ; --i)
        push_nil(vmg0_);

    /* remember the reset pointer */
    reset_ptr_ = G_stk->get_sp();

    /* if the caller has no PC pointer, make a recursive call */
    if (recurse)
    {
        /* recursively call the interpreter loop */
        run(vmg_ p + get_funchdr_size());
    }
    else if (pc_ptr != 0)
    {
        /* 
         *   load the caller's program counter with the first byte of code
         *   in the new function, which immediately follows the header 
         */
        *pc_ptr = p + get_funchdr_size();
    }
    else
    {
        /*
         *   We're setting up a recursive call frame, because pc_ptr == 0,
         *   but we're not making a recursive call - simply return now
         *   with the new frame established but without having started
         *   execution in the recursive VM invocation.  
         */
    }

    /* 
     *   if this was a recursive call, and the debugger is present, remove
     *   the recursive frame we created 
     */
    VM_IF_DEBUGGER(if (pc_ptr == 0) leave_recursive_frame(vmg0_));
}

/*
 *   Set up a call frame for a recursive VM invocation, but do not
 *   actually invoke the execution loop.  
 */
void CVmRun::set_up_recursive_call_frame(VMG_ pool_ofs_t func_ofs, uint argc,
                                         vm_obj_id_t self)
{
    /* 
     *   Set up the call frame.  Since we're setting up a recursive VM
     *   invocation frame, pass null for the PC pointer so that we set the
     *   return address to return to the VM; however, tell do_call NOT to
     *   actually make the recursive call.  
     */
    do_call(vmg_ 0, func_ofs, argc, self, FALSE, 0);

    /* set the PC pointer to the start of the function */
    if (pc_ptr_ != 0)
    {
        /* point to the start of the function, following the header */
        *pc_ptr_ = (const uchar *)G_code_pool->get_ptr(entry_ptr_)
                   + get_funchdr_size();
    }
}

/*
 *   Determine if we're in a recursive VM invocation.  If this frame or
 *   any enclosing frame other than the outermost has a code offset of
 *   zero in the return address slot, we are in a recursive VM invocation.
 */
int CVmRun::is_recursive_invocation(VMG0_) const
{
    vm_val_t *p;
    
    /* start with the current frame */
    p = frame_ptr_;

    /* if there's no frame pointer, it's obviously not recursive */
    if (p == 0)
        return FALSE;

    /* scan frames until we get to the outermost frame */
    for (;;)
    {
        /* 
         *   If this is the outermost frame, we can stop now.  The
         *   outermost frame has an enclosing frame pointer value of null.
         *   (A given frame pointer always points directly to the
         *   enclosing frame pointer stored in the stack frame, so the
         *   offset from this frame pointer is zero.)  
         */
        if (get_enclosing_frame_ptr(vmg_ p) == 0)
            break;

        /* 
         *   Check the return address in this frame - if it's at offset
         *   zero, it means that this method was called directly as a
         *   recursive VM invocation.
         */
        if (get_return_addr_from_frame(vmg_ p) == 0)
            return TRUE;

        /* move to the enclosing frame */
        p = get_enclosing_frame_ptr(vmg_ p);
    }

    /* 
     *   we didn't find any direct invocations after the outermost frame,
     *   so this is the top-level VM invocation 
     */
    return FALSE;
}


/*
 *   Return from the current function.  Returns true if execution can
 *   proceed, false if this returns us out of the outermost function, in
 *   which case the execution loop must terminate and return control to
 *   the host environment.  
 */
int CVmRun::do_return(VMG_ const uchar **pc_ptr)
{
    vm_val_t val;
    int argc;
    pool_ofs_t caller_ofs;
    const uchar *p;
    CVmFuncPtr hdr_ptr;

    /*
     *   The frame pointer always points to the location on the stack
     *   where we pushed the enclosing frame pointer.  Reset the stack
     *   pointer to the current frame pointer, then pop the enclosing
     *   frame pointer.  
     */
    G_stk->set_sp(frame_ptr_);
    G_stk->pop(&val);
    frame_ptr_ = (vm_val_t *)val.val.ptr;

    /* pop the argument count */
    G_stk->pop(&val);
    argc = val.val.intval;

    /* pop the enclosing entry pointer */
    G_stk->pop(&val);
    entry_ptr_ = val.val.ofs;

    /* pop the enclosing code offset */
    G_stk->pop(&val);
    caller_ofs = val.val.ofs;

    /* discard the actual parameters, including the 'self' parameter */
    G_stk->discard(argc + 1);

    /* 
     *   If the enclosing code offset is invalid, we've returned from the
     *   outermost function invoked by the host environment.  0 is an
     *   invalid offset, since offset 0 in a method never contains valid
     *   code.  
     */
    if (caller_ofs == 0)
        return FALSE;

    /* translate the method entry pointer to a physical address */
    p = (const uchar *)G_code_pool->get_ptr(entry_ptr_);

    /* 
     *   get the new program counter - calculate the PC offset by adding
     *   the offset within the method to the entry pointer 
     */
    *pc_ptr =  p + caller_ofs;

    /* 
     *   calculate the reset pointer - this is the frame pointer offset by
     *   the locals 
     */
    hdr_ptr.set(p);
    reset_ptr_ = G_stk->get_from_frame(frame_ptr_, hdr_ptr.get_local_cnt());

    /* continue executing */
    return TRUE;
}


/* ------------------------------------------------------------------------ */
/*
 *   Recursive frame routines.  These are used only if the debugger is
 *   present. 
 */

#ifdef VM_DEBUGGER

/*
 *   Enter a recursive call frame from a native routine 
 */
void CVmRun::enter_recursive_frame(VMG_ int argc)
{
    pool_ofs_t old_ofs;
    int i;

    /* 
     *   don't bother setting up a recursive frame for a recursive call
     *   from the debugger itself - the only purpose of these frames is to
     *   aid the debugger in tracing the stack, which it obviously won't
     *   need to do when it's the native caller 
     */
    if (G_debugger->is_in_debugger())
        return;

    /* 
     *   if there's no global PC register, we're being called from the
     *   outermost native caller, so there's no need for a native frame 
     */
    if (pc_ptr_ == 0)
        return;
    
    /* get the return address from the global PC register */
    old_ofs = pcptr_to_method_ofs(vmg_ *pc_ptr_);

    /* make sure we have space for the native frame */
    if (!G_stk->check_space(5))
        err_throw(VMERR_STACK_OVERFLOW);

    /* there's no 'self' for a recursive caller */
    push_nil(vmg0_);

    /* push the caller's code offset */
    push_codeofs(vmg_ old_ofs);

    /* push the old entrypoint code offset */
    push_codeofs(vmg_ entry_ptr_);

    /* 
     *   push the argument count to the routine being invoked from the
     *   native code - this isn't actually the argument count to the
     *   native routine, which we don't know, but we must push it anyway
     *   because the arguments are arranged as though they're to this fake
     *   native frame 
     */
    push_int(vmg_ (int32)argc);

    /* push the current frame pointer */
    push_stackptr(vmg_ frame_ptr_);

    /* set up the new frame pointer */
    frame_ptr_ = G_stk->get_sp();

    /* there's no entrypoint address for the native code */
    entry_ptr_ = 0;

    /* 
     *   call the debugger to do a step trace - the debugger obviously
     *   can't really stop here, but what it can do is note that we've
     *   stepped through this native stack level for the purposes of
     *   determining when it should stop next for step-in, step-over, and
     *   step-out modes 
     */
    if (G_debugger->is_single_step())
        G_debugger->step(vmg_ 0, 0, FALSE);

    /* 
     *   copy the arguments from this frame - this is necessary so that
     *   the recursive frame we'll set up next (after we return) receives
     *   a copy of its arguments, which we hijacked by establishing this
     *   intermediate frame to represent the native caller 
     */
    for (i = 0 ; i < argc ; ++i)
        G_stk->push(get_param(vmg_ i));
}

/*
 *   Leave a recursive call frame on our way back out to a native routine 
 */
void CVmRun::leave_recursive_frame(VMG0_)
{
    vm_val_t val;
    int argc;
    
    /* 
     *   if we're in the debugger, we will not have set up a recursive
     *   call frame, so we will not need to remove one 
     */
    if (G_debugger->is_in_debugger())
        return;

    /* 
     *   if there's no global PC pointer, it means that we're at the
     *   outermost native frame, which we suppress 
     */
    if (pc_ptr_ == 0)
        return;

    /* re-activate the enclosing frame */
    G_stk->set_sp(frame_ptr_);
    G_stk->pop(&val);
    frame_ptr_ = (vm_val_t *)val.val.ptr;

    /* pop the argument count */
    G_stk->pop(&val);
    argc = val.val.intval;

    /* pop the enclosing entry pointer */
    G_stk->pop(&val);
    entry_ptr_ = val.val.ofs;

    /* 
     *   discard the enclosing code offset - since we know this is
     *   actually a native caller, we pushed the enclosing code offset
     *   only to enable the debugger to find the native caller 
     */
    G_stk->discard();

    /* discard the actual parameters and the 'self' parameter */
    G_stk->discard(argc + 1);
}

#endif /* VM_DEBUGGER */

/* ------------------------------------------------------------------------ */
/*
 *   Set a property of an object 
 */
void CVmRun::set_prop(VMG_ vm_obj_id_t obj, vm_prop_id_t prop,
                      const vm_val_t *new_val)
{
    /* set the property */
    vm_objp(vmg_ obj)->set_prop(vmg_ G_undo, obj, prop, new_val);
}

/* ------------------------------------------------------------------------ */
/*
 *   Evaluate a property of an object 
 */
void CVmRun::get_prop(VMG_ const uchar **pc_ptr,
                      const vm_val_t *target_obj, vm_prop_id_t target_prop,
                      const vm_val_t *self, uint argc)
{
    vm_val_t val;
    vm_obj_id_t srcobj;
    int found;

    /* 
     *   we can evaluate properties of regular objects, as well as string
     *   and list constants - see what we have 
     */
    switch(target_obj->typ)
    {
    case VM_LIST:
        /* 'self' must be the same as the target for a constant list */
        if (self->typ != target_obj->typ
            || self->val.ofs != target_obj->val.ofs)
            err_throw(VMERR_OBJ_VAL_REQD);

        /* evaluate the constant list property */
        if (CVmObjList::
            const_get_prop(vmg_ &val, target_obj,
                           G_const_pool->get_ptr(target_obj->val.ofs),
                           target_prop, &argc))
        {
            /* got it - store the result in r0 */
            r0_ = val;
        }
        else
        {
            /* property not found - make sure there are no arguments */
            if (argc != 0)
                err_throw(VMERR_WRONG_NUM_OF_ARGS);

            /* return nil */
            r0_.set_nil();
        }
        break;

    case VM_SSTRING:
        /* 'self' must be the same as the target for a constant string */
        if (self->typ != target_obj->typ
            || self->val.ofs != target_obj->val.ofs)
            err_throw(VMERR_OBJ_VAL_REQD);

        /* evaluate the constant string property */
        if (CVmObjString::
            const_get_prop(vmg_ &val, target_obj,
                           G_const_pool->get_ptr(target_obj->val.ofs),
                           target_prop, &argc))
        {
            /* got it - store the result in r0 */
            r0_ = val;
        }
        else
        {
            /* property not found - make sure there are no arguments */
            if (argc != 0)
                err_throw(VMERR_WRONG_NUM_OF_ARGS);
            
            /* return nil */
            r0_.set_nil();
        }
        break;

    case VM_OBJ:
        /* get the property value from the target object */
        found = vm_objp(vmg_ target_obj->val.obj)
                ->get_prop(vmg_ target_prop, &val, target_obj->val.obj,
                           &srcobj, &argc);

        /* 'self' must be an object as well */
        if (self->typ != VM_OBJ)
            err_throw(VMERR_OBJ_VAL_REQD);

        /* evaluate the result and store it in R0 */
        eval_prop_val(vmg_ found, pc_ptr, &val, self->val.obj, argc);
        break;

    default:
        /* we can't evaluate properties of anything else */
        err_throw(VMERR_OBJ_VAL_REQD);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Given a value that has been retrieved from an object property,
 *   evaluate the value.  If the value contains code, we'll execute the
 *   code; if it contains a self-printing string, we'll display the
 *   string; otherwise, we'll just store the value in R0.
 *   
 *   'found' indicates whether or not the property value is defined.
 *   False indicates that the property value is not defined by the object;
 *   true indicates that it is.  
 */
void CVmRun::eval_prop_val(VMG_ int found, const uchar **pc_ptr,
                           const vm_val_t *val, vm_obj_id_t self, uint argc)
{
    /* check whether or not the property is defined */
    if (found)
    {
        /* take appropriate action based on the datatype of the result */
        switch(val->typ)
        {
        case VM_CODEOFS:
            /* 
             *   It's a method - invoke the method.  This will set us up
             *   to start executing this new code, so there's nothing more
             *   we need to do here.  
             */
            do_call(vmg_ pc_ptr, val->val.ofs, argc, self, pc_ptr == 0, 0);
            break;
            
        case VM_DSTRING:
            /* no arguments are allowed */
            if (argc != 0)
                err_throw(VMERR_WRONG_NUM_OF_ARGS);
            
            /* 
             *   it's a self-printing string - invoke the default string
             *   output function (this is effectively a do_call()) 
             */
            disp_dstring(vmg_ val->val.ofs, pc_ptr, self);
            break;
            
        default:
            /* for any other value, no arguments are allowed */
            if (argc != 0)
                err_throw(VMERR_WRONG_NUM_OF_ARGS);
            
            /* store the result in R0 */
            r0_ = *val;
            break;
        }
    }
    else
    {
        /* 
         *   the property or method is not defined - discard arguments and
         *   set R0 to nil
         */
        G_stk->discard(argc);
        r0_.set_nil();
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Inherit a property or method from the appropriate superclass of the
 *   object that defines currently executing code.  
 */
void CVmRun::inh_prop(VMG_ const uchar **pc_ptr,
                      vm_prop_id_t prop, uint argc)
{
    CVmFuncPtr hdr_ptr;
    vm_obj_id_t cur_obj;
    vm_val_t val;
    vm_obj_id_t srcobj;
    int found;

    /* set up a pointer to the current function header */
    set_current_func_ptr(vmg_ &hdr_ptr);

    /* get the defining object from the function header */
    cur_obj = hdr_ptr.get_defining_obj();

    /* get the inherited property value */
    found = vm_objp(vmg_ cur_obj)
            ->inh_prop(vmg_ prop, &val, cur_obj, &srcobj, &argc);

    /* 
     *   evaluate and store the result - note that "self" remains the
     *   current "self" object, since we're inheriting within the context
     *   of the original method call 
     */
    eval_prop_val(vmg_ found, pc_ptr, &val, get_self(vmg0_), argc);
}


/* ------------------------------------------------------------------------ */
/*
 *   Display a dstring via the default string display mechanism 
 */
void CVmRun::disp_dstring(VMG_ pool_ofs_t ofs, const uchar **pc_ptr,
                          vm_obj_id_t self)
{
    vm_val_t val;
    
    /* push the string */
    val.set_sstring(ofs);
    pushval(vmg_ &val);

    /* invoke the default "say" function */
    disp_string_val(vmg_ pc_ptr, self);
}

/*
 *   Display the value at top of stack via the default string display
 *   mechanism 
 */
void CVmRun::disp_string_val(VMG_ const uchar **pc_ptr, vm_obj_id_t self)
{
    /* 
     *   if there's a valid 'self' object, and there's a default display
     *   method defined, and 'self' defines or inherits that method,
     *   invoke the method 
     */
    if (say_method_ != VM_INVALID_PROP && self != VM_INVALID_OBJ)
    {
        vm_obj_id_t src_obj;
        vm_val_t val;

        /* 
         *   look up the property - if we find it, and it's a regular
         *   method, invoke it 
         */
        if (vm_objp(vmg_ self)->get_prop(vmg_ say_method_, &val, self,
                                         &src_obj, 0)
            && val.typ == VM_CODEOFS)
        {
            /* there's a default display method - invoke it */
            eval_prop_val(vmg_ TRUE, pc_ptr, &val, self, 1);

            /* done */
            return;
        }
    }
    
    /* if the "say" function isn't initialized, it's an error */
    if (!say_func_inited_)
        err_throw(VMERR_SAY_IS_NOT_DEFINED);

    /* call the "say" function with the argument at top of stack */
    do_call(vmg_ pc_ptr, say_func_, 1, VM_INVALID_OBJ, pc_ptr == 0, 0);
}

/* ------------------------------------------------------------------------ */
/*
 *   Check a property for speculative evaluation 
 */
void CVmRun::check_prop_spec_eval(VMG_ vm_obj_id_t obj, vm_prop_id_t prop)
{
    vm_val_t val;
    vm_obj_id_t srcobj;

    /* get the property value */
    if (vm_objp(vmg_ obj)->get_prop(vmg_ prop, &val, obj, &srcobj, 0))
    {
        /* check the type of the value */
        switch(val.typ)
        {
        case VM_CODEOFS:
        case VM_DSTRING:
        case VM_NATIVE_CODE:
            /* 
             *   evaulating these types could result in side effects, so
             *   this property cannot be evaulated during a speculative
             *   evaluation 
             */
            err_throw(VMERR_BAD_SPEC_EVAL);
            break;

        default:
            /* evaluating other types causes no side effects, so proceed */
            break;
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Save the interpreter state to a file 
 */
void CVmRun::write_to_file(VMG_ CVmFile *fp)
{
    /* 
     *   write the frame pointer and the reset pointer; these are both
     *   offsets from the start of the stack 
     */
    fp->write_int4(G_stk->ptr_to_index(frame_ptr_));
    fp->write_int4(G_stk->ptr_to_index(reset_ptr_));

    /* save the entry pointer */
    fp->write_int4(entry_ptr_);

    /* write the "say" function and method information */
    fp->write_int2(say_func_inited_);
    fp->write_int4(say_func_);
    fp->write_int2(say_method_);

    /* save the program counter, as an offset from the entry pointer */
    if (pc_ptr_ != 0)
        fp->write_int4(pcptr_to_method_ofs(vmg_ *pc_ptr_));
    else
        fp->write_int4(0);
}

/*
 *   Load the interpreter state from a file 
 */
int CVmRun::read_from_file(VMG_ CVmFile *fp)
{
    ulong pc_ofs;
    
    /* read the frame pointer and reset pointer */
    frame_ptr_ = G_stk->index_to_ptr(fp->read_int4());
    reset_ptr_ = G_stk->index_to_ptr(fp->read_int4());

    /* read the entry pointer */
    entry_ptr_ = fp->read_int4();

    /* read the "say" function information */
    say_func_inited_ = fp->read_int2();
    say_func_ = fp->read_int4();
    say_method_ = (vm_prop_id_t)fp->read_int2();

    /* read the program counter as an offset from the entry pointer */
    pc_ofs = fp->read_int4();

    /* 
     *   store the new program counter, if we have a valid pc pointer
     *   stored (i.e., this is being invoked from an intrinsic), and we
     *   had a valid (non-zero) program counter offset stored in the file 
     */
    if (pc_ofs != 0 && pc_ptr_ != 0)
        *pc_ptr_ = (const uchar *)G_code_pool->get_ptr(entry_ptr_) + pc_ofs;
    
    /* success */
    return 0;
}

/*
 *   Reset the machine registers to initial conditions 
 */
void CVmRun::reset(VMG0_)
{
    /* there's no 'say' function yet */
    say_func_ = 0;
    say_func_inited_ = FALSE;
    say_method_ = VM_INVALID_PROP;

    /* there's no frame yet */
    frame_ptr_ = 0;
    reset_ptr_ = 0;

    /* we are not in a method currently */
    entry_ptr_ = 0;

    /* clear R0 */
    r0_.set_nil();
}

/* ------------------------------------------------------------------------ */
/*
 *   Set up a function header pointer for the current function 
 */
void CVmRun::set_current_func_ptr(VMG_ CVmFuncPtr *func_ptr)
{
    /* set up the pointer based on the current Entry Pointer register */
    func_ptr->set((const uchar *)G_code_pool->get_ptr(entry_ptr_));
}

/*
 *   Set up a function header pointer for the return address of the given
 *   stack frame 
 */
void CVmRun::set_return_funcptr_from_frame(VMG_ CVmFuncPtr *func_ptr,
                                           vm_val_t *frame_ptr)
{
    pool_ofs_t ep;
    
    /* get the enclosing entry pointer for the frame */
    ep = get_enclosing_entry_ptr_from_frame(vmg_ frame_ptr);

    /* set up the function pointer for the entry pointer */
    func_ptr->set((const uchar *)G_code_pool->get_ptr(ep));
}

/* ------------------------------------------------------------------------ */
/*
 *   Get the frame pointer at a given stack level 
 */
vm_val_t *CVmRun::get_fp_at_level(VMG_ int level) const
{
    vm_val_t *fp;
    
    /* walk up the stack to the desired level */
    for (fp = frame_ptr_ ; fp != 0 && level != 0 ;
         --level, fp = get_enclosing_frame_ptr(vmg_ fp));

    /* 
     *   if we ran out of frames before we reached the desired level,
     *   throw an error 
     */
    if (fp == 0)
        err_throw(VMERR_BAD_FRAME);

    /* return the frame */
    return fp;
}

/* ------------------------------------------------------------------------ */
/*
 *   Get the message from an "unhandled exception" error object
 */
const char *CVmRun::get_exc_message(VMG_ CVmException *exc, size_t *msg_len)
{
    vm_obj_id_t exc_obj;

    /* 
     *   if the error isn't "unhandled exception," there's not a stored
     *   exception object; likewise, if there's no object parameter in the
     *   exception, there's nothing to use to obtain the message 
     */
    if (exc->get_error_code() != VMERR_UNHANDLED_EXC
        || exc->get_param_count() < 1)
        return 0;
    
    /* get the exception object */
    exc_obj = (vm_obj_id_t)exc->get_param_ulong(0);

    /* get the message from the object */
    return get_exc_message(vmg_ exc_obj, msg_len);
}

/*
 *   Get the message from an exception object 
 */
const char *CVmRun::get_exc_message(VMG_ vm_obj_id_t exc_obj, size_t *msg_len)
{
    vm_val_t val;
    vm_obj_id_t src_obj;
    const char *str;
    uint argc;
    
    /* if there's no object, there's no message */
    if (exc_obj == VM_INVALID_OBJ)
        return 0;
    
    /* 
     *   get the exceptionMessage property value from the object; if
     *   there's not a valid exceptionMessage property defined, or the
     *   object doesn't have a value for the property, there's no message 
     */
    argc = 0;
    if (G_predef->rterrmsg_prop == VM_INVALID_PROP
        || (!vm_objp(vmg_ exc_obj)->get_prop(vmg_ G_predef->rterrmsg_prop,
                                             &val, exc_obj, &src_obj,
                                             &argc)))
        return 0;

    /* 
     *   We got the property.  If it's a string or an object containing a
     *   string, retrieve the string.
     */
    switch(val.typ)
    {
    case VM_SSTRING:
        /* get the constant string */
        str = G_const_pool->get_ptr(val.val.ofs);
        break;

    case VM_OBJ:
        /* get the string value of the object, if possible */
        str = vm_objp(vmg_ val.val.obj)->get_as_string(vmg0_);
        break;

    default:
        /* it's not a string - we can't use it */
        str = 0;
        break;
    }

    /* check to see if we got a string */
    if (str != 0)
    {
        /* 
         *   The string is in the standard VM internal format, which means
         *   it has a 2-byte length prefix followed by the bytes of the
         *   string (with no null termination).  Read the length prefix,
         *   then skip past it so the caller doesn't have to.  
         */
        *msg_len = osrp2(str);
        str += VMB_LEN;
    }

    /* return the string pointer */
    return str;
}

