/* $Header: d:/cvsroot/tads/tads3/vmrun.h,v 1.4 1999/07/11 00:46:59 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmrun.h - VM Execution
Function
  
Notes
  
Modified
  11/12/98 MJRoberts  - Creation
*/

#ifndef VMRUN_H
#define VMRUN_H

#include "vmglob.h"
#include "vmtype.h"
#include "vmstack.h"
#include "vmpool.h"
#include "vmobj.h"
#include "vmerr.h"
#include "vmerrnum.h"


/* ------------------------------------------------------------------------ */
/*
 *   Procedure activation frame.  The activation frame is arranged as
 *   follows (the stack index increases reading down the list):
 *   
 *.  argument N
 *.  argument N-1
 *.  ...
 *.  argument 2
 *.  argument 1
 *.  self
 *.  offset in calling method of next instruction to execute
 *.  caller's entry pointer (EP) register value
 *.  actual parameter count
 *.  caller's frame pointer <<<--- CURRENT FRAME POINTER
 *.  local variable 1
 *.  local variable 2
 *.  local variable 3
 *.                         <<<--- CURRENT RESET POINTER
 *   
 *   So, local variable 1 is at (FP+1), local variable 2 is at (FP+2), and
 *   so on; the argument count is at (FP-1); 'self' is at (FP-4); argument
 *   1 is at (FP-5), argument 2 is at (FP-6), and so on.  
 */

/* offset from FP of first argument */
const int VMRUN_FPOFS_ARG1 = -5;

/* offset from FP of 'self' */
const int VMRUN_FPOFS_SELF = -4;

/* offset from FP of return address */
const int VMRUN_FPOFS_RET = -3;

/* offset from FP of enclosing entry pointer */
const int VMRUN_FPOFS_ENC_EP = -2;

/* offset from FP of argument count */
const int VMRUN_FPOFS_ARGC = -1;

/* offset from FP of enclosing frame pointer */
const int VMRUN_FPOFS_ENC_FP = 0;

/* offset from FP of first local variable */
const int VMRUN_FPOFS_LCL1 = 1;


/* ------------------------------------------------------------------------ */
/*
 *   VM Execution Engine class.  This class handles execution of byte
 *   code.  
 */
class CVmRun
{
    friend class CVmDebug;
    
public:
    CVmRun();
    ~CVmRun();

    /*
     *   Get/set the method header size.  This size is stored in the image
     *   file; the image loader sets this at load time to the value
     *   retrieved from the image file.  All method headers in an image
     *   file use the same size.  
     */
    void set_funchdr_size(size_t siz);
    size_t get_funchdr_size() const { return funchdr_size_; }
    
    /* 
     *   Call a function or method.  'ofs' is the offset (in the code
     *   pool) of the code to invoke, and 'argc' is the number of
     *   arguments that the caller has pushed onto the stack.  'pc_ptr' is
     *   a pointer to the program counter; on return, this will point to
     *   the first byte of code in the new function.
     *   
     *   When calling a function, 'self' should be VM_INVALID_OBJ.
     *   Otherwise, this value gives the object whose method is being
     *   invoked.  
     *   
     *   If 'recurse' is true, we'll make a recursive call to the
     *   interpreter loop to execute the code; thus, the code will have
     *   run to completion by the time we return in this case.  If
     *   'recurse' is false, we'll simply set up the frame and return.
     *   
     *   pc_ptr should be null to set up a recursive VM invocation frame,
     *   whether or not the caller wants us to invoke the execution loop.  
     */
    void do_call(VMG_ const uchar **pc_ptr, pool_ofs_t ofs, uint argc,
                 vm_obj_id_t self, int recurse, const char *recurse_calling);

    /*
     *   Call a function pointer value.  If 'funcptr' contains a function
     *   pointer, we'll simply call the function; if it contains an
     *   anonymous function object, we'll call the anonymous function.  
     */
    void call_func_ptr(VMG_ const vm_val_t *funcptr, uint argc,
                       int recurse, const char *recurse_name,
                       const uchar **pc_ptr);
        
    /* 
     *   Set up a call frame for a recursive VM invocation, without
     *   actually invoking the VM.  This works the same as do_call() with
     *   a null PC pointer, but rather than invoking the VM, simply sets
     *   up the PC register to point to the first byte of the function to
     *   execute.  
     */
    void set_up_recursive_call_frame(VMG_ pool_ofs_t func_ofs, uint argc,
                                     vm_obj_id_t self);

    /*
     *   Get the descriptive message, if any, from an exception object.
     *   The returned string will not be null-terminated, but the length
     *   will be stored in *msg_len.  The returned string might point to
     *   constant pool data or data in an object, so it might not remain
     *   valid after a constant pool translation or garbage collection
     *   operation.  If the exception has no message, we will return a
     *   null pointer.  
     */
    static const char *get_exc_message(VMG_ CVmException *exc,
                                       size_t *msg_len);
    static const char *get_exc_message(VMG_ vm_obj_id_t obj, size_t *msg_len);

    /*
     *   Evaluate a property of an object.  'target_obj' is the object
     *   whose property is to be evaluated, 'target_prop' is the ID of the
     *   property to evaluate, and 'argc' is the number of arguments to
     *   the method.  'pc_ptr' is a pointer to the program counter; on
     *   return, this may point to the first byte of code in the method,
     *   if the property is a method.
     *   
     *   'self' is the object in whose context we're to perform the code
     *   execution, if the property is a method.  Note that 'self' may
     *   differ from 'target_obj' in some cases, particularly when
     *   inheriting.
     *   
     *   If 'pc_ptr' is null, we'll make a recursive call to the
     *   interpreter loop to execute any method code; thus, any method
     *   code will have run to completion by the time we return in this
     *   case.
     */
    void get_prop(VMG_ const uchar **pc_ptr,
                  const vm_val_t *target_obj, vm_prop_id_t target_prop,
                  const vm_val_t *self, uint argc);

    /*
     *   Set a property of an object 
     */
    void set_prop(VMG_ vm_obj_id_t obj, vm_prop_id_t prop,
                  const vm_val_t *new_val);

    /* 
     *   Display a dstring via the default string display function.  This
     *   function pushes a string value (with the given constant pool
     *   offset), then does the same work as do_call() to invoke a
     *   function with one argument.
     *   
     *   The string is identified by its offset in the constant pool.  
     */
    void disp_dstring(VMG_ pool_ofs_t ofs, const uchar **pc_ptr,
                      vm_obj_id_t self);

    /*
     *   Display the value at top of stack via the default string display
     *   function.  does the same work as do_call() to invoke the function
     *   with one argument, which must already be on the stack.  
     */
    void disp_string_val(VMG_ const uchar **pc_ptr, vm_obj_id_t self);

    /* get data register 0 (R0) */
    vm_val_t *get_r0() { return &r0_; }

    /* set the default "say" function */
    void set_say_func(pool_ofs_t code_pool_ofs)
    {
        /* set the offset */
        say_func_ = code_pool_ofs;

        /* note that it's been initialized */
        say_func_inited_ = TRUE;
    }

    /* set the default "say" method */
    void set_say_method(vm_prop_id_t prop)
    {
        /* remember the property */
        say_method_ = prop;
    }

    /* pop an integer value; throws an error if the value is not an integer */
    void pop_int(VMG_ vm_val_t *val)
    {
        G_stk->pop(val);
        if (val->typ != VM_INT)
            err_throw(VMERR_INT_VAL_REQD);
    }

    /* 
     *   Pop a numeric value; throws an error if the value is not numeric.
     *   (At the moment, this is equivalent to pop_int, since int is the
     *   only numeric type; however, we distinguish between numbers in
     *   general and integers in particular, in case additional numeric
     *   types [such as floating-point numbers] are added in the future.) 
     */
    void pop_num(VMG_ vm_val_t *val)
    {
        G_stk->pop(val);
        if (!val->is_numeric())
            err_throw(VMERR_NUM_VAL_REQD);
    }

    /* pop an object value */
    void pop_obj(VMG_ vm_val_t *val)
    {
        G_stk->pop(val);
        if (val->typ != VM_OBJ)
            err_throw(VMERR_OBJ_VAL_REQD);
    }

    /* pop a property pointer value */
    void pop_prop(VMG_ vm_val_t *val)
    {
        G_stk->pop(val);
        if (val->typ != VM_PROP)
            err_throw(VMERR_PROPPTR_VAL_REQD);
    }

    /* pop a function pointer value */
    void pop_funcptr(VMG_ vm_val_t *val)
    {
        G_stk->pop(val);
        if (val->typ != VM_FUNCPTR)
            err_throw(VMERR_FUNCPTR_VAL_REQD);
    }

    /* 
     *   Pop two values from the stack.  The values are popped in reverse
     *   order, so val2 has the value at the top of the stack.  If the
     *   left operand was pushed first, this results in placing the left
     *   operand in val1 and the right operand in val2.  
     */
    void popval_2(VMG_ vm_val_t *val1, vm_val_t *val2)
    {
        popval(vmg_ val2);
        popval(vmg_ val1);
    }

    /* 
     *   Pop two integers, throwing an error if either value is not an
     *   integer.  Pops the item at the top of the stack into val2, and
     *   the next value into val1. 
     */
    void pop_int_2(VMG_ vm_val_t *val1, vm_val_t *val2)
    {
        popval(vmg_ val2);
        popval(vmg_ val1);
        if (val1->typ != VM_INT || val2->typ != VM_INT)
            err_throw(VMERR_INT_VAL_REQD);
    }

    /*
     *   Pop two numbers, throwing an error if either value is not
     *   numeric. 
     */
    void pop_num_2(VMG_ vm_val_t *val1, vm_val_t *val2)
    {
        popval(vmg_ val2);
        popval(vmg_ val1);
        if (!val1->is_numeric() || !val2->is_numeric())
            err_throw(VMERR_NUM_VAL_REQD);
    }

    /* 
     *   get the active function's argument count - we read the value from
     *   the first item below the frame pointer in the current frame 
     */
    int get_cur_argc(VMG0_) const
    {
        return G_stk->get_from_frame(frame_ptr_, VMRUN_FPOFS_ARGC)
            ->val.intval;
    }
    
    /*
     *   Get a parameter value; 0 is the first parameter, 1 is the second,
     *   and so on.  
     */
    vm_val_t *get_param(VMG_ int idx) const
        { return get_param_from_frame(vmg_ frame_ptr_, idx); }

    /* get a parameter from a given frame */
    vm_val_t *get_param_from_frame(VMG_ vm_val_t *fp, int idx) const
        { return G_stk->get_from_frame(fp, VMRUN_FPOFS_ARG1 - idx); }

    /* get a parameter at the given stack level */
    vm_val_t *get_param_at_level(VMG_ int idx, int level) const
    {
        return get_param_from_frame(vmg_ get_fp_at_level(vmg_ level), idx);
    }

    /* 
     *   get a local variable's value; 0 is the first local variable, 1 is
     *   the second, and so on 
     */
    vm_val_t *get_local(VMG_ int idx) const
        { return get_local_from_frame(vmg_ frame_ptr_, idx); }

    /* get a local from a given frame */
    vm_val_t *get_local_from_frame(VMG_ vm_val_t *fp, int idx) const
        { return G_stk->get_from_frame(fp, VMRUN_FPOFS_LCL1 + idx); }

    /* get a local at the given stack level */
    vm_val_t *get_local_at_level(VMG_ int idx, int level) const
    {
        return get_local_from_frame(vmg_ get_fp_at_level(vmg_ level), idx);
    }


    /*
     *   Get the frame pointer at the given stack level.  Level 0 is the
     *   currently active frame, 1 is the first enclosing level, and so
     *   on.  Throws an error if the enclosing frame is invalid. 
     */
    vm_val_t *get_fp_at_level(VMG_ int level) const;

    /*
     *   Get the current frame depth.  This is the stack depth of the
     *   current frame pointer.  This can be used to compare two frame
     *   pointers to determine if one encloses the other - the pointer
     *   with the smaller depth value encloses the larger one.  
     */
    size_t get_frame_depth(VMG0_) const
        { return G_stk->ptr_to_index(frame_ptr_); }

    /* get the current frame pointer */
    vm_val_t *get_frame_ptr() const { return frame_ptr_; }

    /* given a frame pointer, get the enclosing frame pointer */
    static vm_val_t *get_enclosing_frame_ptr(VMG_ vm_val_t *fp)
    {
        return (vm_val_t *)
            G_stk->get_from_frame(fp, VMRUN_FPOFS_ENC_FP)->val.ptr;
    }

    /* get the number of arguments from a given frame */
    int get_argc_from_frame(VMG_ vm_val_t *fp) const
        { return G_stk->get_from_frame(fp, VMRUN_FPOFS_ARGC)->val.intval; }

    /* get the argument counter from a given stack level */
    int get_argc_at_level(VMG_ int level) const
        { return get_argc_from_frame(vmg_ get_fp_at_level(vmg_ level)); }

    /* given a frame pointer, get the 'self' object for the frame */
    static vm_obj_id_t get_self_from_frame(VMG_ vm_val_t *fp)
    {
        vm_val_t *self_val;

        /* get the 'self' slot on the stack */
        self_val = G_stk->get_from_frame(fp, VMRUN_FPOFS_SELF);

        /* return the appropriate value */
        return (self_val->typ == VM_NIL ? VM_INVALID_OBJ : self_val->val.obj);
    }

    /* get the 'self' object at a given stack level */
    vm_obj_id_t get_self_at_level(VMG_ int level) const
        { return get_self_from_frame(vmg_ get_fp_at_level(vmg_ level)); }

    /* 
     *   get the enclosing entry pointer from a given frame (it's the
     *   second item pushed before the enclosing frame ponter, hence it's
     *   at offset -2 in the frame) 
     */
    static pool_ofs_t get_enclosing_entry_ptr_from_frame(VMG_ vm_val_t *fp)
        { return G_stk->get_from_frame(fp, VMRUN_FPOFS_ENC_EP)->val.ofs; }

    /*
     *   Get the return offset from a given frame.  This is the offset of
     *   the return address from the start of the method header for the
     *   frame. 
     */
    static ulong get_return_ofs_from_frame(VMG_ vm_val_t *fp)
        { return G_stk->get_from_frame(fp, VMRUN_FPOFS_RET)->val.ofs; }

    /* 
     *   Get the return address from a given frame.  (The return address
     *   is the third item pushed before the enclosing frame pointer,
     *   hence it's at offset -3 from the frame pointer.)  Returns zero if
     *   we were called by recursive invocation of the VM - this is not
     *   ambiguous with an actual return address of zero, since zero is
     *   never a valid code address.
     */
    static pool_ofs_t get_return_addr_from_frame(VMG_ vm_val_t *fp)
    {
        pool_ofs_t ofs;

        /* get the return method offset from the stack */
        ofs = get_return_ofs_from_frame(vmg_ fp);

        /* 
         *   zero is never a valid method offset, so if the offset is zero
         *   it means that we're at a recursive invocation of the VM -
         *   indicate this by returning zero as the absolute pool address.
         */
        if (ofs == 0)
            return 0;

        /* 
         *   add the offset to the enclosing entry pointer to yield the
         *   absolute pool address of the return point 
         */
        return ofs + get_enclosing_entry_ptr_from_frame(vmg_ fp);
    }
                 
    /* 
     *   Given a frame pointer, set up a function pointer for the return
     *   address from the frame.  
     */
    static void set_return_funcptr_from_frame(VMG_ class CVmFuncPtr *func_ptr,
                                              vm_val_t *frame_ptr);

    /*
     *   Determine if we're in a recursive VM invocation.  Returns true if
     *   we're in a recursive invocation, false if we're in the top-level
     *   invocation.  (A recursive invocation occurs when native code
     *   called from byte code calls back into the VM to call byte code
     *   itself; because the invocation is recursive, we must return to
     *   the native code when the inner byte code returns rather than
     *   continuing with enclosing byte code directly.)
     */
    int is_recursive_invocation(VMG0_) const;

    /* write the interpreter registers to a saved state file */
    void write_to_file(VMG_ class CVmFile *fp);

    /* 
     *   read the interpreter registers from a saved state file; returns
     *   zero on success, or a non-zero VMERR_xxx error code on failure 
     */
    int read_from_file(VMG_ class CVmFile *fp);

    /* reset the machine registers to the initial conditions */
    void reset(VMG0_);
    
    /*
     *   Get the current "self" object.  The "self" object is always the
     *   implicit first parameter to any method.  
     */
    vm_obj_id_t get_self(VMG0_) const
    {
        return G_stk->get_from_frame(frame_ptr_, VMRUN_FPOFS_SELF)->val.obj;
    }

    /*
     *   Set the current "self" object
     */
    void set_self(VMG_ vm_obj_id_t new_self)
    {
        vm_val_t *valp;

        /* get the "self" slot in the stack */
        valp = G_stk->get_from_frame(frame_ptr_, VMRUN_FPOFS_SELF);

        /* set it */
        if (new_self == VM_INVALID_OBJ)
            valp->set_nil();
        else
            valp->set_obj(new_self);
    }

    /* push an object ID */
    static void push_obj(VMG_ vm_obj_id_t obj)
    {
        vm_val_t val;

        /* set the object reference */
        if (obj == VM_INVALID_OBJ)
        {
            /* set the value to nil */
            val.set_nil();

            /* 
             *   explicitly set the object part of the value to invalid,
             *   so we can easily tell that there is not a valid object
             *   here just by looking at the object part; this is
             *   necessary in some cases when we don't want to bother
             *   looking at the type because we know for a fact it has to
             *   be an object 
             */
            val.val.obj = VM_INVALID_OBJ;
        }
        else
            val.set_obj(obj);

        /* push it */
        pushval(vmg_ &val);
    }

    /* push a boolean value */
    static void push_bool(VMG_ int flag)
    {
        vm_val_t val;

        /* set the boolean */
        val.set_logical(flag);

        /* push it */
        pushval(vmg_ &val);
    }

    /* push nil */
    static void push_nil(VMG0_)
    {
        vm_val_t val;

        /* set up the 'nil' value */
        val.set_nil();

        /* push it */
        pushval(vmg_ &val);
    }

    /* push a code offset value */
    static void push_codeofs(VMG_ pool_ofs_t ofs)
    {
        vm_val_t val;

        /* set the code offset */
        val.set_codeofs(ofs);

        /* push it */
        pushval(vmg_ &val);
    }

    /* push a stack pointer value */
    static void push_stackptr(VMG_ vm_val_t *stack_ptr)
    {
        vm_val_t val;

        /* set the stack pointer value */
        val.set_stack((void *)stack_ptr);

        /* push it */
        pushval(vmg_ &val);
    }

    /* push an integer value */
    static void push_int(VMG_ int32 intval)
    {
        vm_val_t val;

        /* set the integer value */
        val.set_int(intval);

        /* push it */
        pushval(vmg_ &val);
    }

    /* push an enumerator value */
    static void push_enum(VMG_ uint32 intval)
    {
        vm_val_t val;

        /* set the enumerator value */
        val.set_enum(intval);

        /* push it */
        pushval(vmg_ &val);
    }

    /* restore a saved state file and resume execution in the saved state */
    void run_restore(VMG_ const char *saved_state_filename);

protected:
    /* 
     *   Execute byte code starting at a given address.  This function
     *   retains control until the byte code function invoked returns or
     *   throws an unhandled exception.
     *   
     *   If an exception occurs and is not handled by the byte code, we'll
     *   throw VMERR_UNHANDLED_EXC with the exception object as the first
     *   parameter.  
     */
    void run(VMG_ const uchar *p);

    /*
     *   Convert a pointer to the currently executing method into an
     *   offset from the start of the current method.  
     */
    ulong pcptr_to_method_ofs(VMG_ const uchar *p)
    {
        /* 
         *   get the memory address of the current entry pointer, and
         *   subtract that from the given memory pointer to get an offset
         *   from the start of the current method 
         */
        return p - (const uchar *)G_code_pool->get_ptr(entry_ptr_);
    }

    /*
     *   Convert a pointer to the currently executing method into a code
     *   pool offset. 
     */
    pool_ofs_t pcptr_to_code_ofs(VMG_ const uchar *p)
    {
        /* 
         *   get the offset from the start of the current method, and add
         *   it to the code pool offset of the start of the current method
         *   to yield the code pool offset of the pointer 
         */
        return (pool_ofs_t)pcptr_to_method_ofs(vmg_ p) + entry_ptr_;
    }

    /*
     *   Set up a function header pointer for the current function 
     */
    void set_current_func_ptr(VMG_ class CVmFuncPtr *func_ptr);

    /* 
     *   get the current entry pointer (the code pool offset of the start
     *   of the current method) 
     */
    pool_ofs_t get_entry_ptr() const { return entry_ptr_; }

    /* call a built-in function */
    void call_bif(VMG_ uint set_index, uint func_index, uint argc,
                  const uchar **pc_ptr);

    /* 
     *   Throw an exception.  Returns true if a handler was found, false
     *   if not.  If a handler was found, byte-code execution can proceed;
     *   if not, the byte-code execution loop must pass the exception up
     *   to its caller.  
     */
    int do_throw(VMG_ const uchar **pcptr, vm_obj_id_t exception_obj);
    
    /* 
     *   Inherit a property - this is essentially the same as get_prop,
     *   but rather than getting the property from the given object, this
     *   ignores any such property defined directly by the object and goes
     *   directly to the inherited definition.  However, the "self" object
     *   is still the same as the current "self" object, since we want to
     *   evaluate the inherited method in the context of the original
     *   target "self" object.  
     */
    void inh_prop(VMG_ const uchar **pcptr, vm_prop_id_t prop, uint argc);

    /*
     *   Evaluate a property value.  If the value contains code, we'll
     *   execute the code; if it contains a self-printing string, we'll
     *   display the string; otherwise, we'll push the value onto the
     *   stack.
     *   
     *   'found' indicates whether or not the property is defined by the
     *   object.  False indicates that the property is not defined, true
     *   that it is defined.  If the property isn't defined, we'll simply
     *   discard arguments and push nil.
     *   
     *   If 'pc_ptr' is null, we'll recursively invoke the interpreter
     *   loop if it's necessary to run a method; otherwise, we'll set up
     *   at the beginning of the method's code and let the caller proceed
     *   into the code.
     */
    void eval_prop_val(VMG_ int found, const uchar **pc_ptr,
                       const vm_val_t *val, vm_obj_id_t self, uint argc);

    /*
     *   Check a property for validity in a speculative evaluation.  If
     *   evaulating the property would cause any side effects, we'll throw
     *   an error (VMERR_BAD_SPEC_EXPR); otherwise, we won't do anything.
     *   Side effects include displaying a dstring or invoking a method. 
     */
    void check_prop_spec_eval(VMG_ vm_obj_id_t obj, vm_prop_id_t prop);

    /*
     *   Return from a function or method.  We'll fill in *pc_ptr with the
     *   new program counter at which to continue execution, and restore
     *   machine registers to the enclosing frame.
     *   
     *   Returns true if execution should proceed, false if we're
     *   returning from the outermost stack level.  When we return false,
     *   the caller must return control to the host environment, since the
     *   host environment called the function from which we're returning.  
     */
    int do_return(VMG_ const uchar **pc_ptr);

    /*
     *   Enter/leave a recursive call frame.  These functions are used
     *   only when a debugger is present. 
     */
    void enter_recursive_frame(VMG_ int argc);
    void leave_recursive_frame(VMG0_);

    /* push a value onto the stack */
    static void pushval(VMG_ const vm_val_t *val) { G_stk->push(val); }

    /* pop a value off the stack */
    void popval(VMG_ vm_val_t *val) { G_stk->pop(val); }
  
    /* add two values and push the result */
    void add_and_push(VMG_ vm_val_t *val1, vm_val_t *val2);

    /* add two values, leaving the result in *val1 */
    void compute_sum(VMG_ vm_val_t *val1, vm_val_t *val2);

    /* subtract one value from another and push the result */
    void sub_and_push(VMG_ vm_val_t *val1, vm_val_t *val2);

    /* subtract one value from another, leaving the result in *val1 */
    void compute_diff(VMG_ vm_val_t *val1, vm_val_t *val2);

    /* multiply and push the result */
    void mul_and_push(VMG_ vm_val_t *val1, vm_val_t *val2);

    /* divide and push the result */
    void div_and_push(VMG_ vm_val_t *val1, vm_val_t *val2);

    /* XOR two values and push the result */
    void xor_and_push(VMG_ vm_val_t *val1, vm_val_t *val2);

    /* 
     *   index container_val by index_val (i.e., compute
     *   container_val[index_val]) and push the result 
     */
    void index_and_push(VMG_ const vm_val_t *container_val,
                        const vm_val_t *index_val);

    /* 
     *   Set the element at index index_val in container_val to new_val,
     *   and push the new container value.  The container may be a new
     *   object, since some types (lists, for example) cannot have their
     *   values changed but instead create new objects when an indexed
     *   element is modified.  
     */
    void set_index_and_push(VMG_ vm_val_t *container_val,
                            const vm_val_t *index_val,
                            const vm_val_t *new_val);

    /* 
     *   create a new object of the given index into the metaclass
     *   dependency table for the load image file, using the given number
     *   of parameters; removes the parameters from the stack, and leaves
     *   the new object reference in register R0 
     */
    void new_and_store_r0(VMG_ const uchar **pc_ptr,
                          uint metaclass_idx, uint argc);

    /* 
     *   pop the two values at top of stack, and compare the two values
     *   for equality; returns true if the values are equal, false if not 
     */
    int pop2_equal(VMG0_)
    {
        vm_val_t val1, val2;

        /* pop the two values at top of stack */
        popval_2(vmg_ &val1, &val2);

        /* compare them and return the result */
        return val1.equals(vmg_ &val2);
    }

    /* 
     *   Pop the two values at top of stack, and compare the two values in
     *   magnitude; returns 1 if the value at (TOS-1) is greater than the
     *   value at TOS, -1 if (TOS-1) is less than (TOS), and 0 if the two
     *   value are equal. 
     */
    int pop2_compare(VMG0_)
    {
        vm_val_t val1, val2;

        /* pop the two values at top of stack */
        popval_2(vmg_ &val1, &val2);

        /* compare the values and return the result */
        return val1.compare_to(vmg_ &val2);
    }

    /* given a constant pool offset, get a pointer to the constant data */
    const char *get_const_ptr(VMG_ pool_ofs_t ofs) const
        { return G_const_pool->get_ptr(ofs); }

    /* 
     *   get a signed 16-bit byte-code operand, incrementing the
     *   instruction pointer past the operand 
     */
    int16 get_op_int16(const uchar **p)
    {
        int16 ret = (int16)osrp2s(*p);
        *p += 2;
        return ret;
    }

    /* get an unsigned 16-bit byte-code operand */
    uint16 get_op_uint16(const uchar **p)
    {
        uint16 ret = (uint16)osrp2(*p);
        *p += 2;
        return ret;
    }

    /* get a signed 32-bit byte-code operand */
    int32 get_op_int32(const uchar **p)
    {
        int32 ret = (int32)osrp4(*p);
        *p += 4;
        return ret;
    }

    /* get an unsigned 32-bit byte-code operand */
    uint32 get_op_uint32(const uchar **p)
    {
        uint32 ret = (uint32)osrp4(*p);
        *p += 4;
        return ret;
    }

    /* get a signed 8-bit byte-code operand */
    int get_op_int8(const uchar **p)
    {
        int ret = (int)(char)**p;
        ++(*p);
        return ret;
    }

    /* get an unsigned 8-bit byte-code operand */
    uint get_op_uint8(const uchar **p)
    {
        uint ret = (uint)**p;
        ++(*p);
        return ret;
    }

    /*
     *   Function header size - obtained from the image file upon loading 
     */
    size_t funchdr_size_;

    /*
     *   The function to invoke for the SAY and SAYVAL opcodes.  This is
     *   an offset into the code pool.  If say_func_inited_ is not set,
     *   the SAY opcode will throw an error.  
     */
    pool_ofs_t say_func_;
    int say_func_inited_;

    /*
     *   The method to invoke for the SAY and SAYVAL opcodes when a valid
     *   "self" object is available.  If no method is defined, this will
     *   be set to VM_INVALID_PROP.  
     */
    vm_prop_id_t say_method_;

    /*
     *   R0 - data register 0.  This register stores function return
     *   values. 
     */
    vm_val_t r0_;

    /* 
     *   EP - entrypoint register.  This register stores the code offset
     *   of the method header of the currently executing code.  
     */
    pool_ofs_t entry_ptr_;

    /*
     *   FP - frame pointer register.  This points to the base of the
     *   current stack activation frame.  Local variables and parameters
     *   are reachable relative to this register. 
     */
    vm_val_t *frame_ptr_;

    /*
     *   Reset pointer.  This is the stack pointer at entry to the
     *   function, which is just beyond the local variables. 
     */
    vm_val_t *reset_ptr_;

    /*
     *   Intrinsic function call program counter (PC) register pointer.
     *   
     *   We keep the PC register on the stack in the main execution loop
     *   for quick access, since accessing this variable is one of the
     *   primary bottlenecks in the interpreter; we keep a pointer to the
     *   stack variable here throughout execution, so that the register
     *   can be accessed outside of the local scope in which the stack
     *   variable is defined.  
     */
    const uchar **pc_ptr_;
};

#endif /* VMRUN_H */

