#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmsave.cpp - T3 save/restore state
Function
  
Notes
  
Modified
  08/02/99 MJRoberts  - Creation
*/

#include "t3std.h"
#include "os.h"
#include "vmglob.h"
#include "vmsave.h"
#include "vmfile.h"
#include "vmimage.h"
#include "vmobj.h"
#include "vmrun.h"
#include "vmstack.h"
#include "vmundo.h"


/* ------------------------------------------------------------------------ */
/*
 *   Saved state signature.  This signature is tied to a specific format
 *   version; it should be changed whenever the format version is modified
 *   so that it is incompatible with older versions.
 */
#define VMSAVEFILE_SIG "T3-state-v0002\015\012\032"


/* ------------------------------------------------------------------------ */
/*
 *   Save VM state to a file 
 */
void CVmSaveFile::save(VMG_ CVmFile *fp)
{
    const char *fname;
    size_t fname_len;
    
    /* 
     *   We cannot save the state from within a recursive VM invocation --
     *   the full machine state in a recursive invocation includes native
     *   machine state, which we are not equipped to save.  We could thus
     *   never restore such a state, so there's no point in saving it.  
     */
    if (G_interpreter->is_recursive_invocation(vmg0_))
        err_throw(VMERR_RECURSIVE_VM_CALL);

    /* write the header */
    fp->write_bytes(VMSAVEFILE_SIG, sizeof(VMSAVEFILE_SIG)-1);

    /* write the image file's timestamp */
    fp->write_bytes(G_image_loader->get_timestamp(), 24);

    /* get the image filename */
    fname = G_image_loader->get_filename();
    fname_len = strlen(fname);

    /* 
     *   write the image filename, so we can figure out what image file to
     *   load if we start the interpreter specifying only the saved state
     *   to be restored 
     */
    fp->write_int2(strlen(fname));
    fp->write_bytes(fname, fname_len);

    /* save the stack */
    G_stk->write_to_file(vmg_ fp);

    /* save the interpreter registers */
    G_interpreter->write_to_file(vmg_ fp);

    /* save all modified object state */
    G_obj_table->save(vmg_ fp);
}

/* ------------------------------------------------------------------------ */
/*
 *   Given a saved state file, get the name of the image file that was
 *   loaded when the state file was created. 
 */
int CVmSaveFile::restore_get_image(osfildef *fp,
                                   char *fname_buf, size_t fname_buf_len)
{
    char buf[128];
    size_t len;
    int err;

    /* read the signature and timestamp */
    if (osfrb(fp, buf, sizeof(VMSAVEFILE_SIG)-1 + 24))
        return VMERR_READ_FILE;

    /* check the signature */
    if (memcmp(buf, VMSAVEFILE_SIG, sizeof(VMSAVEFILE_SIG)-1) != 0)
        return VMERR_NOT_SAVED_STATE;

    /* read the length of the image file name */
    if (osfrb(fp, buf, 2))
        return VMERR_READ_FILE;

    /* get the length from the buffer */
    len = osrp2(buf);

    /* if it won't fit in the buffer, return an error */
    if (len + 1 > fname_buf_len)
        return VMERR_READ_FILE;

    /* read the name into the caller's buffer */
    if (osfrb(fp, fname_buf, len))
        return VMERR_READ_FILE;

    /* null-terminate the name */
    fname_buf[len] = '\0';

    /* success */
    return 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Restore VM state from a file.  Returns zero on success, non-zero on
 *   error.  
 */
int CVmSaveFile::restore(VMG_ CVmFile *fp)
{
    char buf[128];
    int err;

    /*
     *   We cannot restore state during a recursive VM invocation.  Part
     *   of our current machine state is in the native machine stack
     *   (i.e., the recursive call state); we are not equipped to modify
     *   the native state, hence we cannot restore in the current context.
     */
    if (G_interpreter->is_recursive_invocation(vmg0_))
        return VMERR_RECURSIVE_VM_CALL;

    /* read the signature and timestamp */
    fp->read_bytes(buf, sizeof(VMSAVEFILE_SIG)-1 + 24);

    /* check the signature */
    if (memcmp(buf, VMSAVEFILE_SIG, sizeof(VMSAVEFILE_SIG)-1) != 0)
        return VMERR_NOT_SAVED_STATE;

    /* check the timestamp */
    if (memcmp(buf + sizeof(VMSAVEFILE_SIG)-1,
               G_image_loader->get_timestamp(), 24) != 0)
        return VMERR_WRONG_SAVED_STATE;

    /* 
     *   skip the image filename - since we already have an image file
     *   loaded, this information is of no use to us here (it's only
     *   useful when we want to restore a saved state before we know what
     *   the image file is) 
     */
    fp->read_bytes(buf, 2);
    fp->set_pos_from_cur(osrp2(buf));

    /* 
     *   discard all undo information - any undo information we currently
     *   have obviously can't be applied to the restored state 
     */
    G_undo->drop_undo(vmg0_);

    /* reset all objects to the initial image file load state */
    G_obj_table->reset_to_image(vmg0_);

    /* load the stack */
    if ((err = G_stk->read_from_file(vmg_ fp)) != 0)
        return err;

    /* load the machine registers */
    if ((err = G_interpreter->read_from_file(vmg_ fp)) != 0)
        return err;

    /* load the object data from the file */
    if ((err = G_obj_table->restore(vmg_ fp)) != 0)
        return err;

    /* success */
    return 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Reset to initial image file state 
 */
void CVmSaveFile::reset(VMG0_)
{
    /* 
     *   We cannot reset the state from within a recursive VM invocation
     *   -- the full machine state in a recursive invocation includes
     *   native machine stack state that we cannot modify.
     */
    if (G_interpreter->is_recursive_invocation(vmg0_))
        err_throw(VMERR_RECURSIVE_VM_CALL);

    /* 
     *   discard undo information, since it applies only to the current VM
     *   state and obviously is no longer relevant after we reset to the
     *   initial state 
     */
    G_undo->drop_undo(vmg0_);

    /* reset all objects to initial image file load state */
    G_obj_table->reset_to_image(vmg0_);

    /* clear the stack - discard all elements */
    G_stk->discard(G_stk->get_depth());

    /* reset the machine registers */
    G_interpreter->reset(vmg0_);
}

