/* $Header: d:/cvsroot/tads/tads3/VMTOBJ.H,v 1.2 1999/05/17 02:52:29 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmtobj.h - VM TADS Object implementation
Function
  
Notes
  
Modified
  10/30/98 MJRoberts  - Creation
*/

#ifndef VMTOBJ_H
#define VMTOBJ_H

#include <stdlib.h>
#include <string.h>

#include "t3std.h"
#include "vmtype.h"
#include "vmglob.h"
#include "vmobj.h"
#include "vmundo.h"


/* ------------------------------------------------------------------------ */
/*
 *   TADS Object extension header for load image objects.  This structure
 *   appears at the very beginning of our object extension; it contains
 *   information that we keep in memory for the object but is not part of
 *   the persistent state of the object (hence this information is not
 *   loaded from an image file nor saved to or restored from a saved state
 *   file).
 *   
 *   The image file state is loaded into an image object data block; we
 *   keep a pointer to that data.  The image file data block is arranged
 *   as follows:
 *   
 *.  UINT2 superclass_count
 *.  UINT2 load_image_property_count
 *.  UINT2 flags
 *.  UINT4 superclass_1
 *.  ...
 *.  UINT4 superclass_N
 *.  UINT2 load_image_property_ID_1
 *.  DATAHOLDER load_image_property_value_1
 *.  ...
 *.  UINT2 load_image_property_ID_N
 *.  DATAHOLDER load_image_property_value_N 
 */  
/*   
 *   If we load the object from an image file, the above structure is in
 *   read-only memory and comes from the image file; we will keep a
 *   pointer directly to image file memory.
 *   
 *   Separately, we keep a set of persistent data (i.e., data that must be
 *   included when writing a saved state file).  If we load the object
 *   from an image file, the persistent data block is allocated
 *   separately; otherwise, the persistent data block is allocated as part
 *   of the same block that contains the above structure, and immediately
 *   follows the above structure in the memory block.
 *   
 *   The persistent state for a TADS object consists of an array of
 *   modified properties for the object, plus a set of "undo" bits.
 *   Immediately following the modified properties array, we keep a packed
 *   array of bits that we use to track for each property in the
 *   modifiable list whether we've created an undo record since the start
 *   of the current savepoint.
 *   
 *   PROP_SLOT_1
 *.  PROP_SLOT_2
 *.  ...
 *.  PROP_SLOT_N (number given by hdr->prop_alloc)
 *.  UNDO_BITS_BYTE_1
 *.  UNDO_BITS_BYTE_2
 *.  
 *.  UNDO_BITS_BYTE_N/8 (we pack 8 bits into each undo byte) 
 */
struct vm_tadsobj_hdr
{
    /* pointer to load image data */
    const char *image_data;

    /* number of allocated modifiable property slots */
    ushort prop_alloc;
};


/* ------------------------------------------------------------------------ */
/*
 *   Object Flag Values - these values are stored in the image file object
 *   header.  
 */

/* class - the object represents a class, not an instance */
#define VMTOBJ_OBJF_CLASS    0x0001


/* ------------------------------------------------------------------------ */
/*
 *   Size of a property slot.  Each property slot has a property ID and a
 *   data holder.
 */
const size_t VMTOBJ_PROP_SLOT_SIZE = VMB_PROP_ID + VMB_DATAHOLDER;

/*
 *   Property table expansion increment.  Each time we need to expand the
 *   number of available slots in a property table, we'll increase it by
 *   this amount. 
 */
const ushort VMTOBJ_PROP_INC = 16;

/*
 *   Initial empty property table size.  When we initially load an object,
 *   we'll allocate this many empty slots for modifiable properties. 
 */
const ushort VMTOBJ_PROP_INIT = 16;


/* ------------------------------------------------------------------------ */
/*
 *   TADS object interface.
 */
class CVmObjTads: public CVmObject
{
    friend class CVmMetaclassTads;
    
public:
    /* metaclass registration object */
    static class CVmMetaclass *metaclass_reg_;
    class CVmMetaclass *get_metaclass_reg() const { return metaclass_reg_; }

    /* create dynamically using stack arguments */
    static vm_obj_id_t create_from_stack(VMG_ const uchar **pc_ptr,
                                         uint argc);

    /* 
     *   call a static property - we don't have any of our own, so simply
     *   "inherit" the base class handling 
     */
    static int call_stat_prop(VMG_ vm_val_t *result,
                              const uchar **pc_ptr, uint *argc,
                              vm_prop_id_t prop)
        { return CVmObject::call_stat_prop(vmg_ result, pc_ptr, argc, prop); }

    /* create an object with no initial extension */
    static vm_obj_id_t create(VMG_ int in_root_set);

    /* 
     *   Create an object with a given number of superclasses, and a given
     *   number of property slots.  The property slots are all initially
     *   allocated to modified properties.  
     */
    static vm_obj_id_t create(VMG_ int in_root_set,
                              ushort superclass_count, ushort prop_slots);

    /* notify of deletion */
    void notify_delete(VMG_ int in_root_set);

    /* create an instance of this object */
    void create_instance(VMG_ vm_obj_id_t self,
                         const uchar **pc_ptr, uint argc);

    /* determine if the object has a finalizer method */
    virtual int has_finalizer(VMG_ vm_obj_id_t /*self*/);

    /* invoke the object's finalizer */
    virtual void invoke_finalizer(VMG_ vm_obj_id_t self);

    /* get the number of superclasses of this object */
    virtual int get_superclass_count(VMG_ vm_obj_id_t self) const
    {
        /* 
         *   if we have no superclass, inherit the default, since we
         *   inherit from the system TadsObject class; if we have our own
         *   superclasses, return them 
         */
        if (get_sc_count() == 0)
            return CVmObject::get_superclass_count(vmg_ self);
        else
            return get_sc_count();
    }

    /* get the nth superclass of this object */
    virtual vm_obj_id_t get_superclass(VMG_ vm_obj_id_t self, int idx) const
    {
        /* 
         *   if we have no superclass, inherit the default, since we
         *   inherit from the system TadsObject class; if we have our own
         *   superclasses, return them 
         */
        if (get_sc_count() == 0)
            return CVmObject::get_superclass(vmg_ self, idx);
        else if (idx >= get_sc_count())
            return VM_INVALID_OBJ;
        else
            return get_sc(idx);
    }

    /* determine if I"m a class object */
    virtual int is_class_object(VMG_ vm_obj_id_t /*self*/) const
        { return (get_obj_flags() & VMTOBJ_OBJF_CLASS) != 0; }

    /* determine if I'm an instance of the given object */
    int is_instance_of(VMG_ vm_obj_id_t obj);

    /* this object type provides properties */
    int provides_props(VMG0_) const { return TRUE; }

    /* enumerate properties */
    void enum_props(VMG_ vm_obj_id_t self,
                    void (*cb)(VMG_ void *ctx,
                               vm_obj_id_t self, vm_prop_id_t prop,
                               const vm_val_t *val),
                    void *cbctx);

    /* set a property */
    void set_prop(VMG_ class CVmUndo *undo,
                  vm_obj_id_t self, vm_prop_id_t prop, const vm_val_t *val);

    /* get a property */
    int get_prop(VMG_ vm_prop_id_t prop, vm_val_t *val,
                 vm_obj_id_t self, vm_obj_id_t *source_obj, uint *argc);

    /* inherit a property */
    int inh_prop(VMG_ vm_prop_id_t prop, vm_val_t *val,
                 vm_obj_id_t self, vm_obj_id_t *source_obj, uint *argc);

    /* 
     *   Receive notification of a new savepoint.  We keep track of
     *   whether or not we've saved undo information for each modifiable
     *   property in the current savepoint, so that we can avoid saving
     *   redundant undo information when repeatedly changing a property
     *   value (since only the first change in a given savepoint needs to
     *   be recorded).  When we start a new savepoint, we obviously
     *   haven't yet stored any undo information for the new savepoint, so
     *   we can simply clear all of the undo records.  
     */
    void notify_new_savept()
        { clear_undo_bits(); }

    /* apply undo */
    void apply_undo(VMG_ struct CVmUndoRecord *rec);

    /* mark a reference in an undo record */
    void mark_undo_ref(VMG_ struct CVmUndoRecord *undo);

    /* 
     *   remove stale weak references from an undo record -- we keep only
     *   normal strong references, so we don't need to do anything here 
     */
    void remove_stale_undo_weak_ref(VMG_ struct CVmUndoRecord *) { }

    /* mark references */
    void mark_refs(VMG_ uint state);

    /* 
     *   remove weak references - we keep only normal (strong) references,
     *   so this routine doesn't need to do anything 
     */
    void remove_stale_weak_refs(VMG0_) { }

    /* load from an image file */
    void load_from_image(VMG_ vm_obj_id_t self, const char *ptr, size_t siz);

    /* restore to image file state */
    void reset_to_image(VMG0_);

    /* determine if the object has been changed since it was loaded */
    int is_changed_since_load() const;

    /* save to a file */
    void save_to_file(VMG_ class CVmFile *fp);

    /* restore from a file */
    void restore_from_file(VMG_ vm_obj_id_t self,
                           class CVmFile *fp, CVmObject *obj);

    /* rebuild for image file */
    virtual ulong rebuild_image(VMG_ char *buf, ulong buflen);

    /* convert to constant data */
    virtual void convert_to_const_data(VMG_ class CVmConstMapper *mapper,
                                       vm_obj_id_t self);

    /* get the nth superclass */
    vm_obj_id_t get_sc(ushort n) const
        { return vmb_get_objid(get_sc_slot_ptr(n)); }

    /* set the nth superclass to the given object */
    void set_sc(ushort n, vm_obj_id_t obj)
        { vmb_put_objid((char *)get_sc_slot_ptr(n), obj); }

    /* get the object flags */
    uint get_obj_flags() const { return osrp2(get_flags_ptr()); }

    /* set the object flags */
    void set_obj_flags(uint flags) { oswp2(get_flags_ptr(), flags); }


protected:
    /* create an object with no initial extension */
    CVmObjTads() { ext_ = 0; }

    /* 
     *   Create an object with a given number of superclasses, and a given
     *   number of property slots.  All property slots are initially
     *   allocated to the modifiable property list.  
     */
    CVmObjTads(VMG_ ushort superclass_count, ushort prop_count);

    /* 
     *   Allocate memory - this replaces any existing extension, so the
     *   caller must take care to free the extension (if one has already
     *   been allocated) before calling this routine.
     *   
     *   If 'from_image' is true, we're allocating memory for use with an
     *   object loaded from an image file, so we'll ignore the superclass
     *   count and leave the image_data pointer in the header unchanged.
     *   If 'from_image' is false, we're allocating memory for a dynamic
     *   object that does not have a presence in the image file, so we'll
     *   allocate space for the superclass list as part of the extension
     *   and set the image_data pointer in the header to refer the extra
     *   space after the modifiable property array and undo bit array.  
     */
    void alloc_mem(VMG_ ushort sc_count, ushort mod_prop_count,
                   int from_image);

    /* calculate how much space we need for our extension */
    size_t calc_alloc_size(ushort superclass_count, ushort mod_prop_count,
                           int from_image) const;

    /* 
     *   Expand the property table, reallocating the object extension to
     *   make room for the larger list.  We expand the table so that it
     *   contains at least the given number of modifiable property slots;
     *   if we already have the requested number of slots available, we
     *   won't do anything.  
     */
    void expand_prop_table(VMG_ ushort new_mod_slots, class CVmObject *obj);

    /* 
     *   Get a property, considering only property lists (including our
     *   own property list and all of our superclass's property lists).
     *   Does not inherit from base metaclasses.  
     */
    int get_prop_from_list(VMG_ vm_prop_id_t prop, vm_val_t *val,
                           vm_obj_id_t self, vm_obj_id_t *source_obj,
                           uint *argc);

    /* 
     *   Inherit a property from a superclass properyt list, ignoring base
     *   metaclasses. 
     */
    int inh_prop_from_list(VMG_ vm_prop_id_t prop, vm_val_t *val,
                           vm_obj_id_t self, vm_obj_id_t *source_obj,
                           uint *argc);

    /* 
     *   Look up a property given the property ID, returning a pointer to
     *   the property's data in portable format; returns null if the
     *   property is not defined in this object.  This routine does not
     *   look at any superclasses. 
     */
    const char *look_up_prop(vm_prop_id_t prop) const;

    /*
     *   Look up a property given the property ID, looking only among the
     *   modifiable property slots.  Returns a pointer to the property's
     *   slot if found, null if not.
     *   
     *   If 'ret_empty_slot' is true, rather than returning null on
     *   failing to find the property, we'll return a pointer to the first
     *   empty slot.  The caller can determine whether the desired
     *   property was found by checking the property ID in the returned
     *   slot: if it's VM_INVALID_PROP, the property wasn't found, and we
     *   returned an empty slot instead.  This will still return null,
     *   even if ret_empty_slot is true, if there are no free slots and
     *   the property wasn't found.  
     */
    char *look_up_mod_prop(vm_prop_id_t prop, int ret_empty_slot) const;

    /* get/set the superclass count */
    ushort get_sc_count() const
        { return vmb_get_uint2(get_sc_cnt_ptr()); }
    void set_sc_count(ushort cnt)
        { vmb_put_uint2((char *)get_sc_cnt_ptr(), cnt); }

    /* get/set the number of load image properties */
    ushort get_li_prop_count() const
        { return vmb_get_uint2(get_li_prop_cnt_ptr()); }
    void set_li_prop_count(ushort cnt)
        { vmb_put_uint2((char *)get_li_prop_cnt_ptr(), cnt); }

    /* get/set the total number of property slots allocated */
    ushort get_mod_prop_alloc() const   { return get_hdr()->prop_alloc; }
    void set_mod_prop_alloc(ushort cnt) { get_hdr()->prop_alloc = cnt; }

    /* given a slot pointer, get the property ID from the slot */
    vm_prop_id_t get_slotptr_propid(const char *slotptr) const
        { return vmb_get_propid(slotptr); }

    /* set the property ID in a slot pointer */
    void set_slotptr_propid(char *slotptr, vm_prop_id_t prop)
        { vmb_put_propid(slotptr, prop); }

    /* given a slot pointer, get the value from the slot */
    void get_slotptr_val(const char *slotptr, vm_val_t *val) const
        { vmb_get_dh(get_slotptr_val_ptr_const(slotptr), val); }

    /* given a slot pointer, get the DATAHOLDER pointer for the slot */
    char *get_slotptr_val_ptr(char *slotptr) const
        { return slotptr + VMB_PROP_ID; }
    const char *get_slotptr_val_ptr_const(const char *slotptr) const
        { return get_slotptr_val_ptr((char *)slotptr); }

    /* given a slot pointer, set the value for the slot */
    void set_slotptr_val(char *slotptr, const vm_val_t *val)
        { vmb_put_dh(slotptr + VMB_PROP_ID, val); }

    /* set a slot's contents, given a slot pointer */
    void set_prop_slotptr(char *slotptr, vm_prop_id_t prop,
                          const vm_val_t *val)
    {
        /* set the property ID */
        vmb_put_propid(slotptr, prop);

        /* set the value */
        vmb_put_dh(slotptr + VMB_PROP_ID, val);
    }

    /*
     *   Get/set/clear the undo bit for the given property slot 
     */
    int get_undo_bit(size_t slot) const
        { return (((get_undo_bits()[slot >> 3]) & (1 << (slot & 7))) != 0); }
    void set_undo_bit(size_t slot)
        { get_undo_bits()[slot >> 3] |= (1 << (slot & 7)); }
    void clear_undo_bit(size_t slot)
        { get_undo_bits()[slot >> 3] &= ~(1 << (slot & 7)); }

    /*
     *   Get the undo slot number for a slot pointer.  Since we only store
     *   undo bits for the modifiable slots, the undo slot number is
     *   relative to the first modifiable slot. 
     */
    ushort get_undo_slotnum(char *slotptr)
        { return get_mod_slot_from_slot_ptr(slotptr); }

    /*
     *   Clear all undo bits 
     */
    void clear_undo_bits()
        { memset(get_undo_bits(), 0, ((get_mod_prop_alloc() + 7) >> 3)); }


    /* -------------------------------------------------------------------- */
    /*
     *   Low-level format management - these routines encapsulate the byte
     *   layout of the object in memory.  This is a bit nasty because we
     *   keep the object's contents in the portable image format.  
     */

    /* 
     *   get my header - the header is at the very beginning of our
     *   extension, so this simply casts the extension appropriately 
     */
    struct vm_tadsobj_hdr *get_hdr() const { return (vm_tadsobj_hdr *)ext_; }

    /* get a pointer to the image data */
    const char *get_image_data() const { return get_hdr()->image_data; }

    /* 
     *   Determine if this object was loaded from an image file.  Returns
     *   true if the object comes from an image file, false if it was
     *   created dynamically. 
     */
    int is_from_image() const
    {
        /*
         *   If the image data pointer in the header points to the part
         *   immediately after the property and undo bit tables, then we
         *   don't have image data but are storing the superclass
         *   information ourselves.
         *   
         *   To get the size of the mutable data, calculate our allocation
         *   size as though we were stored in the image file.  Add this
         *   size to our extension pointer.  If we're storing everything
         *   in our allocation block, then the "image data" pointer will
         *   actually point here.  Hence, if the "image data" pointer
         *   points to this point in our own block, we were dynamically
         *   created; otherwise, we were created from image file data.  
         */
        return (get_image_data()
                != ext_ + calc_alloc_size(0, get_mod_prop_alloc(), TRUE));
    }

    /* 
     *   get the start of the persistent data - this immediately follows
     *   the header 
     */
    char *get_pdata() const
        { return ext_ + sizeof(vm_tadsobj_hdr); }

    /*
     *   Calculate the size of the saved state image data needed to hold a
     *   given number of modified properties 
     */
    size_t calc_save_size(size_t prop_count)
    {
        /* we need space for each property slot in the modified list */
        return (prop_count * VMTOBJ_PROP_SLOT_SIZE);
    }

    /* get a pointer to the superclass count data */
    const char *get_sc_cnt_ptr() const { return get_image_data(); }

    /* get a pointer to the load image (fixed) property count data */
    const char *get_li_prop_cnt_ptr() const
        { return get_sc_cnt_ptr() + VMB_UINT2; }

    /* get a pointer to the object flags in the load image */
    const char *get_flags_ptr() const
        { return get_li_prop_cnt_ptr() + VMB_UINT2; }

    /* get a pointer to the first superclass slot */
    const char *get_sc_slot0_ptr() const
        { return get_flags_ptr() + VMB_UINT2; }

    /* get a pointer to the nth superclass slot */
    const char *get_sc_slot_ptr(size_t n) const
        { return get_sc_slot0_ptr() + (n * VMB_OBJECT_ID); }

    /* get a pointer to the first load image (fixed) property */
    const char *get_li_prop0_slot_ptr() const
        { return get_sc_slot_ptr(get_sc_count()); }

    /* get a pointer to the nth load image (fixed) property */
    const char *get_li_prop_slot_ptr(size_t n) const
        { return get_li_prop0_slot_ptr() + (n * VMTOBJ_PROP_SLOT_SIZE); }

    /* get a pointer to the first modifiable property slot */
    char *get_mod_prop_slot0_ptr() const
        { return (char *)(get_hdr() + 1); }

    /* get a pointer to a given property slot */
    char *get_mod_prop_slot_ptr(ushort slot) const
        { return (get_mod_prop_slot0_ptr()
                  + (slot * VMTOBJ_PROP_SLOT_SIZE)); }

    /* given a modifiable slot pointer, get the slot number */
    size_t get_mod_slot_from_slot_ptr(char *slotptr) const
        { return (slotptr -
                  get_mod_prop_slot0_ptr()) / VMTOBJ_PROP_SLOT_SIZE; }

    /* 
     *   given a pointer to a property slot, increment the pointer so that
     *   it points to the next slot in the table 
     */
    void inc_prop_slot_ptr(char **slotptr) const
        { *slotptr += (VMTOBJ_PROP_SLOT_SIZE); }

    /* get a pointer to the undo bit array */
    unsigned char *get_undo_bits() const
    {
        /* the undo bit array starts immediately after the property slots */
        return (unsigned char *)get_mod_prop_slot_ptr(get_mod_prop_alloc());
    }
};

/* ------------------------------------------------------------------------ */
/*
 *   Registration table object 
 */
class CVmMetaclassTads: public CVmMetaclass
{
public:
    /* get the global name */
    const char *get_meta_name() const { return "tads-object/030000"; }
    
    /* create from image file */
    void create_for_image_load(VMG_ vm_obj_id_t id)
        { new (vmg_ id) CVmObjTads(); }
    
    /* create from restoring from saved state */
    void create_for_restore(VMG_ vm_obj_id_t id)
        { new (vmg_ id) CVmObjTads(); }
    
    /* create dynamically using stack arguments */
    vm_obj_id_t create_from_stack(VMG_ const uchar **pc_ptr, uint argc)
        { return CVmObjTads::create_from_stack(vmg_ pc_ptr, argc); }
    
    /* call a static property */
    int call_stat_prop(VMG_ vm_val_t *result,
                       const uchar **pc_ptr, uint *argc,
                       vm_prop_id_t prop)
    {
        return CVmObjTads::call_stat_prop(vmg_ result, pc_ptr, argc, prop);
    }
};

#endif /* VMTOBJ_H */

/*
 *   Register the class 
 */
VM_REGISTER_METACLASS(CVmObjTads)
